import ij.*;
import ij.process.*;
import ij.gui.*;
import ij.util.Tools;
import java.awt.*;
import ij.plugin.*;

import ij.text.*;
import java.text.SimpleDateFormat;
import java.io.*;
import ij.io.*;

import ij.plugin.filter.PlugInFilter;
import java.awt.event.*;

import java.util.*;
import javax.swing.*;
import javax.swing.event.*;
import ij.plugin.filter.*;
import ij.plugin.filter.GaussianBlur;
import ij.measure.*;
import ij.measure.Calibration;
import ij.measure.ResultsTable;
import java.awt.Rectangle;

import java.awt.GraphicsDevice;
import java.awt.GraphicsEnvironment;

import java.awt.image.ColorModel;

//================================================================================
/**
	*  IO and VSD Signal Processor
	*  1.1.0.d00    (of  Jan 7, 2012)
	*
	*  Quantification of evoked intrinsic optical signal (IOS) or voltage sensitive dye (VSD) fluorescence signal
	*	in multiframe TIFF images.
	*
	*  Download the newest version of this plugin and (possibly) more detailed instructions
	*	at    http://tinyurl.com/vsdproc
	*	or    http://www.neuroscience.ubc.ca/faculty/murphy_software.html#VSDPROC
	*
	*  Evoked IO signals are determined as the fraction of light reflection change that is specific to stimulation.
	*  Evoked VSD signals are determined as the normalized fraction of fluorescence change specific to stimulation.
	*  The first set of frames in every stack (range from "baselinefirst"  to "baselinelast") 
	*	is considered as the signal of the unstimulated neuronal tissue,
	*  the second set of frames  (range from "responsefirst"  to "responselast") 
	*	is considered as the signal of the stimulated neuronal tissue.
	*  Optionally one can determine the ratio between those in the stacks in the selected folder
	*  	both for forelimb (file name "lo*.tif") and hindlimb ("hi*.tif") stimulation.
	*  Results are displayed and/or saved as new TIFF image files within the same folder:
	*  	lo_base.tif,  hi_base.tif		sum of multiple trials and then average baseline frames
	*  	lo_resp.tif,  hi_resp.tif		sum of multiple trials and then average response frames
	*  	lo_av.tif,  hi_av.tif			average stack from multiple trials 
	*  	lo_avn.tif,  hi_avn.tif		average stack from multiple trials (for normalized VSD map)
	*  	lo_diff.tif,  hi_diff.tif		difference of "lo_resp.tif" and "lo_base.tif" 
	*						( or "hi_resp.tif" and "hi_base.tif", respectively).
	*  	lo_deltaf_f.tif,  hi_deltaf_f.tif	[Delta]F / F for the averaged stack
	* Result files from previous runs won't be overwritten by this plugin.
	* Settings are saved as ini file to be reloaded with next run
	*
	* @author Albrecht Sigler / Shawn Zhang / Ian Winship / Timothy H. Murphy
	* @author University of British Columbia, Vancouver BC, Canada
	* @author Correspondence to asigler@interchange.ubc.ca
	* 
	*   see also:
	*	Harrison TC, Sigler A, Murphy TH.,
	*	Simple and cost-effective hardware and software for functional brain mapping using intrinsic optical signal imaging.
	*	J. Neurosci. Methods 182 (2009), 211-218;  doi:10.1016/j.jneumeth.2009.06.021.
	*
	*  Copyright (C) 2009, 2011 Albrecht Sigler
	* 
	*
	*  This plugin, described above and persented in this file (subsequently referred ''software'') is distributed in the hope 
	*  that it will be useful, but without any warranty; without even the implied warranty of merchantability or fitness for a particular purpose. 
	*  Along with any distribution of this software, rights and liabilities irrevocably stay in terms of copyright
	*  for any part of this software as generally either specified or, if specifications don't apply, as intended 
	*  by the most recent GNU General Public License (e.g. http://www.gnu.org/licenses/gpl-3.0.txt ). 
	*  In case you didn't receive an appropiate copy of the GNU General Public License along with this software and don't get it
	*  using the linked site, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
	* 
	*/
//================================================================================


public class Intrinsic_Signal_Processor_VSD_1 implements PlugIn {
	private static final String PluginTitle="IO and VSD Signal Processor";
	private static final String ShortPluginTitle="IO / VSD";
	private static final String PluginVersion="1.1.0.d00";
	
	private GenericDialog gd;
	private Button operationButton, filterSetButton, optionButton;
	private Button NButton, EButton, SButton, WButton;
	private JButton inputFolderButton;
	private JComboBox inputFolderChoice, conSelCBox;
	private JLabel inputFolderLabel;
	private Panel inputPanel;
	private Button helpButton0, helpButton1, helpButton2, helpButton3;
	private Button ApplyButton, ResetButton;
	
	private static final String BSL=System.getProperty("file.separator");
	private static final String DST="\u0394";  // "[Delta]" string (Greek letter)
	private static final String SZS=".";   // "\u2080"; subscript "0" string
	
	private String DirIn="";
	private String DirIn0 = "";
	private String [] PDir;
	private String [] HDir;
	private int NbH = 10;
	private String IniFileName="vsdiomap.ini";
	private String Pth="";
	
	// boolean array.  - Value is "true" if the stack with the same index is to be ignored or not available.
	// 	iob[0] [][] is true if file does NOT exist
	//	iob[1] [][] is true if file does not exist OR is to be ignored
	// 	iob[] [0] [j] no1, no2, ..., no(j)
	// 	iob[] [1] [j] hi1, hi2, ..., hi(j)
	// 	iob[] [2] [j] lo1, lo2, ..., lo(j)
	boolean [][][] iob;
	
	// int array.  - counter to keep statistics of what is done and what to do 
	int [][] counter;   // [0][i]  # of files found to analyze; [1][i] # of  files that do have been analyzed;  [2][i] normalization factor;  [3][i] # of ignored/interpolated pixels.
	
	// int array counting invalid values (NaN, infinity) in raw data 
	int [][] failureIdxList = null;
	//	failureIdxList[0][i]  x
	//	failureIdxList[1][i]  y
	//	failureIdxList[2][i]  z
	//	failureIdxList[3][i]  error count
	
	// checkboxes that represent the value of boolean [][] iob
	private JCheckBox [][] IgnoreBox;
	private int IboxStartRow=0, IboxStartColumn=0;
	private Checkbox NormCB, ZFilterAvBox, ZFilterAvBox2, FilterXYBox, doBlankBox, GaussianFBox, MeanFBox, NoFilterBox, FilterAvBox, FilterBRBox, FilterDeltaBox, hasConBox;
	private Checkbox  aBlankBox, verboseBox, normmBox, silentBox, cBaseBox, ShowPStackBox, SavePStackBox;
	private Checkbox  dontKeepBox, localKeepBox, personalKeepBox, globalKeepBox, customKeepBox, ignoreNBox, interpolateBox;
	private Checkbox  ShowAvBox, SaveAvBox, AverageBox, MedianBox, StdevBox, SterrBox;
	private Label XFilterLabel, FilterTypeLabel;
	private JLabel SigmaLabel, FramesLabel, SigmaLabel2, FramesLabel2;
	private JTextField BaselineFirstField, BaselineLastField, ResponseFirstField, ResponseLastField;
	private JTextField blankFromField, blankToField, XYFilterField, XYFilterField2, ZFilterField, ZFilterField2, NCField, NHLField;
	private String winDir = "", homeDrive = "", publicDir = "", progrDir = "";
	
	//  file name trunks
	private String nhl [] = {"no","hi","lo"};
	private boolean hasControl=true;   // true if the first element in 'nhl' array is for a control file series
	private int truncateFileNameColectionAt = 30;  // -- maximal length of file name in dialog; 
								// longer file names will be truncated.	
	private static final int [] screenSize = getScreenSize();
	
	// number of Checkboxes per name trunk to display
	private int NbChkbxs=24;	// e.g. 10, 20, ...
	// maximal possible number (depending on screen height); 
	//	will be set by calling getMaxNbChkbxs(boolean extend);
	private int MaxNbChkbxs = 0;
	// size of graphic elements;
	//	needed to calculate MaxNbChkbxs etc.
	private int lineHeight = getLineHeight();  // = IJ.isJava14()&&IJ.isWindows()?19:17;  // until Java 1.5: 17 px
	private int checkboxHeight = 25;  // Java 1.6 up: 24 only.
	
	//  options and settings for the optional frame interpolations (which "blank" artifacts)
	private boolean blankArtifact = false;  // if true, frames are interpolated
	private boolean allowBlankArtifact = false;  // if false, intrepolation option are invisible
	private int blankFrom = 0;  // first frame that is interpolated;  must be larger than baselinelast
	private int blankTo = 0;  // last fraem that is interpolated;  must be small than stack that responsefirst
	
	private long PgrTime = 0;  // timer for progress bar
	
	//  calculation modus presets for private void "Icalculate" and "IcalculateSlow" (which both replace "CalculatorPlus"):
	private static final int ADD=0, SUBTRACT=1, MULTIPLY=2, DIVIDE1=3, DIVIDE2=4, NDIFF1=5, NDIFF2=6, NDIV1=7, NDIV2=8;
	//  calculation modus presets for private void "IcalculateSingle" and "IcalculateSingleSlow":
	private static final int SCALE=0, SCALEREV=1, SCALEINV=2, ABS=3, BYPASS=4;
	
	//  switches for the presettings keeper
	private static final int GET=0, SET=1, CHECK=2;
	
	// operation modi, tell waht to do.  At least one of the following must be true or there's nothing to do.
	private boolean ShowRaw=false;
	private boolean ShowBase=false;
	private boolean SaveBase=false;
	private boolean ShowResp=false;
	private boolean SaveResp=false;
	private boolean ShowDiff=false;
	private boolean SaveDiff=false;
	private boolean ShowPerc=true;
	private boolean SavePerc=true;
	private boolean ShowDiv=false;  // not used in this version,
	private boolean SaveDiv=false;  //   both should be kept false
	private boolean ShowNo=false;
	private boolean SaveNo=true;
	private boolean ShowAv=true;
	private boolean SaveAv=true;
	private boolean ShowDelta=false;
	private boolean SaveDelta=false;
	private boolean ShowPStack=false;
	private boolean SavePStack=false;
	private boolean ShowRatio=false;
	private boolean SaveRatio=false;
	private boolean ShowQual=false;
	private boolean SaveQual=false;
	private boolean DoSaveLog=true;
	private boolean DoShowLog=true;

	private boolean commonBase = true;
	
	// default range of frames that will be considered as images recorded before stimulation
	private int baselinefirst = 1;
	private int baselinelast =15;
	private int baselinefirstIO=1;  // - default settings for IO signal processing
	private int baselinelastIO=15;
	private int baselinefirstVSD=1;  // - default settings for VSD signal processing
	private int baselinelastVSD=10;
	
	// default range of frames that will be considered as images for during and after stimulation
	private int responsefirst = 16;
	private int responselast = 30;
	private int responsefirstIO= 16;  // - IO default
	private int responselastIO= 30;
	private int responsefirstVSD= 11;  // - VSD default
	private int responselastVSD= 17;
	
	// z Filter,  +/- frames to average
	private int FilterZ=1;
	private int FilterZIO=0;  // - IO default
	private int FilterZVSD=1;  // - VSD default
	
	// x,y (Gaussian or Mean) filter, kernel radius
	private double FilterXy=2.0;
	private double FilterXyIO=4.0;  // - IO default
	private double FilterXyVSD=0.0;  // - VSD default
	
	private int tps= 11;  // number of presettings
	private int framecnt= 0;
	private long filesize= 0;
	private long filesize0= 0;
	private int src = 0;
	
	private String fileDate="";
	
	//  pre-settings for differrent file types
	private long filesize1= 12199376;
	private int framecnt1= 31;
	private int baselinefirst1= 1;
	private int baselinelast1=15;
	private int responsefirst1= 16;
	private int responselast1= 30;
	private int FilterZ1=0;
	private double FilterXy1=2.0;
	
	private long filesize2= 29771664;
	private int framecnt2= 31;
	private int baselinefirst2= 1;
	private int baselinelast2=15;
	private int responsefirst2= 16;
	private int responselast2= 30;
	private int FilterZ2=0;
	private double FilterXy2=1.0;
	
	private long filesize3= 7448936;
	private int framecnt3= 31;
	private int baselinefirst3= 1;
	private int baselinelast3=15;
	private int responsefirst3= 16;
	private int responselast3= 30;
	private int FilterZ3=0;
	private double FilterXy3=1.0;
	
	private long filesize4= 3600648;  // lo17
	private int framecnt4= 145;
	private int baselinefirst4= 33;
	private int baselinelast4=42;
	private int responsefirst4= 44;
	private int responselast4= 53;
	private int FilterZ4=0;
	private double FilterXy4=0.0;
	
	private long filesize5= 3055864; // lo5
	private int framecnt5= 31;
	private int baselinefirst5= 1;
	private int baselinelast5=15;
	private int responsefirst5= 16;
	private int responselast5= 30;
	private int FilterZ5=0;
	private double FilterXy5=1.5;
	
	private long filesize6= 2957288; // lo9
	private int framecnt6= 30;
	private int baselinefirst6= 1;
	private int baselinelast6=10;
	private int responsefirst6= 11;
	private int responselast6= 17;
	private int FilterZ6=0;
	private double FilterXy6=2.0;
	
	private long filesize7= 2607368;  //  lo18  xfast
	private int framecnt7= 105;
	private int baselinefirst7= 33;
	private int baselinelast7=42;
	private int responsefirst7= 44;
	private int responselast7= 53;
	private int FilterZ7=0;
	private double FilterXy7=2.0;
	
	private long filesize8= 1986568; // lo4  -- 150.5 Hz
	private int framecnt8= 80;
	private int baselinefirst8= 2;
	private int baselinelast8=29;
	private int responsefirst8= 30;
	private int responselast8= 57;
	private int FilterZ8=0;
	private double FilterXy8=1.0;
	
	private long filesize9=1688584; // lo2f  -- 150 Hz
	private int framecnt9= 68;
	private int baselinefirst9= 2;
	private int baselinelast9=29;
	private int responsefirst9= 30;
	private int responselast9= 57;
	private int FilterZ9=0;
	private double FilterXy9=1.0;
	
	private long filesize10=48772928;
	private int framecnt10=31;
	private int baselinefirst10=1;
	private int baselinelast10=15;
	private int responsefirst10=16;
	private int responselast10=30;
	private int FilterZ10=0;
	private double FilterXy10=1.5;
	
	private long filesize11=2681864;
	private int framecnt11=108;
	private int baselinefirst11=15;
	private int baselinelast11=31;
	private int responsefirst11=32;
	private int responselast11=46;
	private int FilterZ11=1;
	private double FilterXy11=1.0;
	
	private long [] filesizeA;
	private int [] framecntA;
	private int [] baselinefirstA;
	private int [] baselinelastA;
	private int [] responsefirstA;
	private int [] responselastA;
	private int [] FilterZA;
	private double [] FilterXyA;
	
	// ----- hidden parameters ---------------------------------------------------
	// scaling saturation of displayed images
	private static final double ScalingD=0.5;
	
	// interpolate NaN and Infinity values of pixels for each frame in input files
	//     (otherwise, these pixels will be ignored).
	private boolean interpolateXY=false;
	
	//  -- empiric params for optimal frame size reduction
	private static final double dborder=0.1;
	private static final double QRfactor=2.3;
	
	//  -- number of digits after separator in Qual table
	private int precision = 4;
	
	// -- check / ignore names of windows to generate unique names
	private static final boolean ignoreOpenWindows=false;
	
	// -- list seperator in log file
	private static final String LogSpt="  ";  
	//  private static final String LogSpt="\t";  // -- tabstop:
	
	//  -- options of  "data data deviation plot"
	private boolean rankQual=true;
	private boolean normQual=false;
	
	// recursicve search for active directory if dir of first plugin run has ot been found
	// 	(set minimal subdirlivel from root)
	private static final int minSubDirLevel=1;  // DONT CHANGE
	
	// -- hours of being idle until plugin re-asks initially for setting options
	//	(set 0 for always asking)
	private static final long hToSetOpt=8;  // ######
	
	// keep comments in ini file (may allow up-compatiblity to later versions if needed)
	private static final boolean keepComInIni=false;  //  ######
	
	// ----- responsiveness of the algorithm -------------------------------------------------
	// - return messages to user at incident or minor concern
	private boolean verboseMode=false;
	// - don't return any error messages to user at error or major concern
	private boolean silentMode=false;
	// - debug code (for code debugging purpose only; nomally to be set false) 
	private static final boolean testMode=false;
	
	//  --- (kind of) obsolete params...	## - keep for future version!
	// set whether to normalize vsd images to the average of no# (if applicable)
	private boolean Normalizeb=true;
	
	// --  NEW OPTIONS, to be fully implemented -------------------------------
	// scale [Delta]F/F0-stacks or [Delta]I/I0-stacks
	private static final boolean DeltaAutoScale=true;
	private static final double DeltaScalingD=0.5;
	// -- fixed values to scale [Delta]F/F0-stacks if they are not autoscaling
	private static final double DeltaMinScale=-0.05, DeltaMaxScale=0.15;
	// -- mutiply [Delta]F/F0-stacks by 100 to display percentage
	private static final boolean DoDelta100=true;
	
	private boolean DoFilterXY=false;
	private boolean DoFilterXY0=false;
	private boolean DoFilterAv=false;
	private boolean DoFilterBR=true;
	private boolean DoFilterDelta=false;
	private boolean DoZFilterAv=false;
	private boolean DoZFilterAv0=false;
	private boolean FilterMean = false;  // uses Gaussian if false
	private final int MinStack = 6;
	private final int PBwait = 31;
	
	// --- the following constants are the same as in ZProjector (except 'SE_METHOD'):
	private static final int AVG_METHOD=0, MAX_METHOD=1, MIN_METHOD=2, SUM_METHOD=3, SD_METHOD=4, MEDIAN_METHOD=5, SE_METHOD=6;
	private int AMethod = AVG_METHOD;
	private static final String [] AVS={"_av","","","","_sd","_md","_se"};
	private static final String [] AVL={" averaged ","","",""," stand. dev. "," median "," stand. err. "};
	private long lastRun=0;
	private String iniComments="";
	private boolean Canceled=false;
	private boolean LayoutChanged=false;
	private boolean DidSave=false;
	private String errString="";
	
	private String storedSettings = "";
	private boolean gotSet=false;
	private boolean keepSet=false;  //  true if settings are to be saved even if Canceld = true (i.e. after error)
	private boolean newSettings=false;  // true if user loaded settings and updated the current settings
	//  ---  constants and variables to preset the method how the plug should store and retrieve its settings
	private static final int DONT_KEEP=0, KEEP_LOCAL=1, KEEP_PERSONAL=2, KEEP_GLOBAL=3, KEEP_CUSTOM=4;
	private int storeModus=KEEP_GLOBAL;
	
	// -- Set prefered size of an image window (screen size = 1)
	private static final double WinSizeOnScreen=0.4;
	//  uses 0.4 if  WinSizeOnScreen is <= 0 or > 1)
	//  -- Zoom Level, ratios to increase/decrease window size
	private double zoomLevel=0;
	//  -- ImageJ default presets:
	private static final double[] zoomLevels = { 1/72.0, 1/48.0, 1/32.0, 1/24.0, 1/16.0, 1/12.0, 
		1/8.0, 1/6.0, 1/4.0, 1/3.0, 1/2.0, 0.75, 1.0, 2.0, 3.0, 4.0, 6.0, 8.0, 12.0, 16.0, 24.0, 32.0 }; 
	
	// -- Set size of the log text window same as the image window size
	private static final boolean DoFitLogWinSizeToScreen=true;
	// -- otherwise, use explicitly the following values
	private int LogWinHeight=250;
	private int LogWinWidth=350;
	private static final boolean DoAdjQmapWinSizeToScreen=true;
	private int QmapWinHeight=400;  // #####  also could become an user set option.
	private int QmapWinWidth=600;   //  #####  make sure it's smaller than screen size...!
	
public void run(String arg) {
	
	if (verboseMode) silentMode=false;
	IJ.showStatus(PluginTitle);
	
	//  ----------  get default presettings   -----------
	filesizeA = new long [tps];
	framecntA = new int [tps];
	baselinefirstA = new int [tps];
	baselinelastA = new int [tps];
	responsefirstA = new int [tps];
	responselastA = new int [tps];
	FilterZA = new int [tps];
	FilterXyA = new double [tps];
	
	filesizeA[0] = filesize1;
	framecntA[0] = framecnt1;
	baselinefirstA[0] = baselinefirst1;
	baselinelastA[0] = baselinelast1;
	responsefirstA[0] = responsefirst1;
	responselastA[0] = responselast1;
	FilterZA[0] = FilterZ1;
	FilterXyA[0] = FilterXy1;
	
	filesizeA[1] = filesize2;
	framecntA[1] = framecnt2;
	baselinefirstA[1] = baselinefirst2;
	baselinelastA[1] = baselinelast2;
	responsefirstA[1] = responsefirst2;
	responselastA[1] = responselast2;
	FilterZA[1] = FilterZ2;
	FilterXyA[1] = FilterXy2;
	
	filesizeA[2] = filesize3;
	framecntA[2] = framecnt3;
	baselinefirstA[2] = baselinefirst3;
	baselinelastA[2] = baselinelast3;
	responsefirstA[2] = responsefirst3;
	responselastA[2] = responselast3;
	FilterZA[2] = FilterZ3;
	FilterXyA[2] = FilterXy3;
	
	filesizeA[3] = filesize4;
	framecntA[3] = framecnt4;
	baselinefirstA[3] = baselinefirst4;
	baselinelastA[3] = baselinelast4;
	responsefirstA[3] = responsefirst4;
	responselastA[3] = responselast4;
	FilterZA[3] = FilterZ4;
	FilterXyA[3] = FilterXy4;
	
	filesizeA[4] = filesize5;
	framecntA[4] = framecnt5;
	baselinefirstA[4] = baselinefirst5;
	baselinelastA[4] = baselinelast5;
	responsefirstA[4] = responsefirst5;
	responselastA[4] = responselast5;
	FilterZA[4] = FilterZ5;
	FilterXyA[4] = FilterXy5;
	
	filesizeA[5] = filesize6;
	framecntA[5] = framecnt6;
	baselinefirstA[5] = baselinefirst6;
	baselinelastA[5] = baselinelast6;
	responsefirstA[5] = responsefirst6;
	responselastA[5] = responselast6;
	FilterZA[5] = FilterZ6;
	FilterXyA[5] = FilterXy6;
	
	filesizeA[6] = filesize7;
	framecntA[6] = framecnt7;
	baselinefirstA[6] = baselinefirst7;
	baselinelastA[6] = baselinelast7;
	responsefirstA[6] = responsefirst7;
	responselastA[6] = responselast7;
	FilterZA[6] = FilterZ7;
	FilterXyA[6] = FilterXy7;
	
	filesizeA[7] = filesize8;
	framecntA[7] = framecnt8;
	baselinefirstA[7] = baselinefirst8;
	baselinelastA[7] = baselinelast8;
	responsefirstA[7] = responsefirst8;
	responselastA[7] = responselast8;
	FilterZA[7] = FilterZ8;
	FilterXyA[7] = FilterXy8;
	
	filesizeA[8] = filesize9;
	framecntA[8] = framecnt9;
	baselinefirstA[8] = baselinefirst9;
	baselinelastA[8] = baselinelast9;
	responsefirstA[8] = responsefirst9;
	responselastA[8] = responselast9;
	FilterZA[8] = FilterZ9;
	FilterXyA[8] = FilterXy9;
	
	filesizeA[9] = filesize10;
	framecntA[9] = framecnt10;
	baselinefirstA[9] = baselinefirst10;
	baselinelastA[9] = baselinelast10;
	responsefirstA[9] = responsefirst10;
	responselastA[9] = responselast10;
	FilterZA[9] = FilterZ10;
	FilterXyA[9] = FilterXy10;
	
	filesizeA[10] = filesize11;
	framecntA[10] = framecnt11;
	baselinefirstA[10] = baselinefirst11;
	baselinelastA[10] = baselinelast11;
	responsefirstA[10] = responsefirst11;
	responselastA[10] = responselast11;
	FilterZA[10] = FilterZ11;
	FilterXyA[10] = FilterXy11;
	
	//  ----------  Get settings of previous run from ini file   -----------
	HDir = new String [NbH+1];
	for (int hc=0; hc<=NbH; hc++) HDir[hc]="";
	
	//  ----------  get presettings   -----------
	try {
		checkboxHeight = (int)Prefs.get("checkboxHeight.int", (double) checkboxHeight);
		storeModus = (int)Prefs.get("VSDproc_store.int", (double) storeModus);
		Pth = Prefs.get("VSDproc_pth.string", "");
		lastRun = (long)Prefs.get("VSDproc_lastRun.long", 0.0);
	} catch (NoSuchMethodError e) {}
	gotSet = findPathAndDoSettings (Pth, GET, !silentMode && 0 < Pth.length());
	
	// ----- find optional folders to read ----------
	String TDir="";
	File directory;
	String [] FDir = new String[HDir.length+2];
	int  pc=0;
	for (int hc=0; hc<HDir.length; hc++)  {
		TDir=HDir[hc];
		if (IJ.isWindows()) {for (int c=0; c<pc; c++) if (0<TDir.length()) if (0==FDir[c].compareToIgnoreCase(TDir)) TDir="";}
		else {for (int c=0; c<pc; c++) if (0<TDir.length()) if (0==FDir[c].compareTo(TDir)) TDir="";}
		if (0<TDir.length()) {
			try {
				directory = new File(TDir);
				if (directory.isDirectory()) FDir[pc++]=""+directory;
			}
			catch (NullPointerException e) {}
			catch (SecurityException e) {}
		}
	}
	if (IJ.isWindows()) {
		if (0==homeDrive.length()) homeDrive = GetEnvironS("HOMEDRIVE");
		TDir = homeDrive;
		if (0<TDir.length()) {
	 		directory = new File(TDir +BSL+"XCAP"+BSL+"data");
			try {
				if (directory.isDirectory()) {
					TDir=""+directory;
					for (int c=0; c<pc; c++) if (0<TDir.length()) if (0==FDir[c].compareTo(TDir)) TDir="";
					if (0<TDir.length()) FDir[pc++] = TDir;
				}
			}
			catch (NullPointerException e) {}
			catch (SecurityException e) {}
		}
	}
	if (pc<FDir.length-2 && IJ.isWindows()) {
		if (0==publicDir.length()) {
			publicDir = GetEnvironS("PUBLIC");
			if (0==publicDir.length() && 0<homeDrive.length()) publicDir = ".";
		}
		if (1< publicDir.length()) {
			TDir = publicDir;
			directory = new File(TDir+BSL+"Documents"+BSL+"EPIX"+BSL+"XCAP"+BSL+"data");
			try {
				if (directory.isDirectory()) {
					TDir=""+directory;
					for (int c=0; c<pc; c++) if (0<TDir.length()) if (0==FDir[c].compareTo(TDir)) TDir="";
					if (0<TDir.length()) FDir[pc++] = TDir;
				}
			}
			catch (NullPointerException e) {}
			catch (SecurityException e) {}
		}
	}
	if (pc<FDir.length-2 && IJ.isWindows()) {
		if (0==progrDir.length()) progrDir = GetEnvironS("ProgramFiles");
		TDir = progrDir;
		if (0<TDir.length()) {
	 		directory = new File(TDir +BSL+"XCAP"+BSL+"data");
			try {
				if (directory.isDirectory()) {
					TDir=""+directory;
					for (int c=0; c<pc; c++) if (0<TDir.length()) if (0==FDir[c].compareTo(TDir)) TDir="";
					if (0<TDir.length()) FDir[pc++] = TDir;
				}
			}
			catch (NullPointerException e) {}
			catch (SecurityException e) {}
		}
	}
	if (pc<3 && IJ.isWindows()) {
		if (0==homeDrive.length()) homeDrive = GetEnvironS("HOMEDRIVE");
		TDir = homeDrive;
		if (0<TDir.length()) {
			directory = new File(TDir);
			try {
				if (directory.isDirectory()) {
					TDir=""+directory;
					for (int c=0; c<pc; c++) if (0<TDir.length()) if (0==FDir[c].compareTo(TDir)) TDir="";
					if (0<TDir.length()) FDir[pc++] = TDir;
				} 
			}
			catch (NullPointerException e) {}
			catch (SecurityException e) {}
		}
	}
	if (pc<FDir.length-2) {
		try {
			directory = new File(IJ.getDirectory("image"));
			if (directory.isDirectory()) {
				TDir=""+directory;
				for (int c=0; c<pc; c++) if (0<TDir.length()) if (0==FDir[c].compareTo(TDir)) TDir="";
				if (0<TDir.length()) FDir[pc++] = TDir;
			}
		}
		catch (NoSuchMethodError e) {}
		catch (NullPointerException e) {}
		catch (SecurityException e) {}
	}
	if (pc<3) {
		TDir="";
		try {
		 	if (IJ.getVersion().compareTo("1.34s") < 0)
				TDir = System.getProperty("user.home");
			else TDir = IJ.getDirectory("home");
		}
		catch(NoSuchMethodError e) {
			TDir = System.getProperty("user.home");
			if (0==TDir.length() && IJ.isWindows()) TDir = GetEnvironS("HOMEPATH");
		}
		try {
			directory = new File(TDir);
			if (directory.isDirectory()) {
				TDir=""+directory;
				for (int c=0; c<pc; c++) if (0<TDir.length()) if (0==FDir[c].compareTo(TDir)) TDir="";
				if (0<TDir.length()) FDir[pc++] = TDir;
			}
		}
		catch (NullPointerException e) {}
		catch (SecurityException e) {}
	}
	try {
		directory = new File(Prefs.getString(Prefs.DIR_IMAGE));
		if (directory.isDirectory()) {
			TDir=""+directory;
			for (int c=0; c<pc; c++) if (0<TDir.length()) if (0==FDir[c].compareTo(TDir)) TDir="";
			if (0<TDir.length()) FDir[pc++] = TDir;
		}
	} 
	catch (NoSuchMethodError e) {}
	catch (NullPointerException e) {}
	catch (SecurityException e) {}
	
	PDir = new String [pc];
	for (int c=0; c<pc; c++) PDir[c]=FDir[c];
	
	//  ----------  ask user for settings if plugin hasn't been run for some time ----------  
	long runstart=System.currentTimeMillis();
	if (runstart-lastRun>hToSetOpt*3600000) {
		boolean wasSilent = silentMode;
		silentMode=false;
		if (!showOperationDialog()) {
			IJ.showStatus("");
			return;
		}
		silentMode=wasSilent;
	}
	
	//  ----------  check some values for consitency  ----------
	DoFilterXY0 = DoFilterXY;  // keep initial setting for later undos
	if (!DoFilterAv && !DoFilterBR && !DoFilterDelta) DoFilterXY = false;
	if (0 >= FilterXy) DoFilterXY = false;
	DoZFilterAv0 = DoZFilterAv;  // keep initial setting for later undos, too
	if (0 >= FilterZ) DoZFilterAv = false;
	boolean FilterMean0 = FilterMean;
	// -- in very old versions of ImageJ Gaussian filtering doesn't seem to work properly
	boolean mustBeMean = true;
	try {mustBeMean = IJ.getVersion().compareTo("1.33c") < 0;}  // false in all newer versions
	catch (NoSuchMethodError ex) {}
	if (mustBeMean) FilterMean=true;  // filter Mean if Gaussian isn't available
	
	//  ----------  ask user  ----------
	if (!showDialog()) return;
	runstart=System.currentTimeMillis();  // keeping the current time (for statistics purpose only)
	
	// -------------- local index bounds  ---------------
	int xi=nhl.length;
	int xj=0;
	for (int i=0; i<xi; i++) if (iob[1][i].length>xj) xj=iob[1][i].length;
	
	// -------------- prevent file overwriting  -------------------------------------------
	directory=new File(DirIn);
	String [] s=null;
	try {if (directory.exists()) s=directory.list();}
	catch (SecurityException e) {
		if (!silentMode && !Canceled) IJ.error("Failed access to ''"+DirIn+"''!"); 
		if (!Canceled) errString+=", failed access to ''"+DirIn+"''";
		return;
	}
	catch (NullPointerException e) {s=null;}
	if (null==s) {
		if (!silentMode && !Canceled) IJ.error("No files found in ''"+DirIn+"''!"); 
		if (!Canceled) errString+=", No files found in ''"+DirIn+"''";
		return;
	}
	int Fidx =0;
	boolean FileFound;
	String FTag="";
	String [] w=null;
	if (!ignoreOpenWindows) w=winList();
	do {
		FileFound=checkNames(s,xj, FTag,false);
		if (!FileFound && null!=w) {
			//  FileFound=checkNames(w,xj, FTag,ShowRaw);
			FileFound=checkNames(w,xj, FTag, true);   // checks for raw files also by default
			if (!FileFound) FileFound = null!=WindowManager.getFrame("Log"+FTag+".txt");
			if (!FileFound) FileFound = null!=WindowManager.getFrame("Log"+FTag);
			if (!FileFound) FileFound = null!=WindowManager.getFrame("Data Quality Map"+FTag+".txt");
			if (!FileFound) FileFound = null!=WindowManager.getFrame("Data Quality Map"+FTag);
		}
		if (FileFound) {
			Fidx++;
			FTag="["+Fidx+"]";
		}
	} while (FileFound);
	
	//  ----------  get image data from files ---------------------------
	ImagePlus impNo=null;
	ImagePlus [] impL = new ImagePlus [xi+1];
	int width  = 0;
	int height = 0;
	int Sz=0;  //  -- stack size (number of frames)
	double [][][] qv=null;

	for(int i=0; i<xi; i++) if(0<counter[0][i] && !Canceled && (0<i || (0==i && (Normalizeb || SaveNo || ShowNo || !hasControl || 1==nhl.length)))) {
		xj=0;
		for (int j=1; j<iob[1][i].length; j++) if (!iob[1][i][j]) xj=iob[1][i].length;
		ImagePlus imp0;
		ImagePlus imp1=null, impS=null, impMB=null, impMR=null;
		String [] tmpFileNames = null;
		ImagePlus [] tmpImp = null; // new ImagePlus[0];
		boolean was32bit = false;  // true if one of the loaded images is initially 32 bit
		int imagesToSpool = 0;  
		for(int j=1; j<xj; j++) if (!iob[1][i][j]){
			imp0=null;
			try {
				if (IJ.escapePressed()) {
					IJ.showStatus("canceled.");
					// for(int jj=j; jj<xj; jj++)iob[1][i][jj]=true;
					// for(int ii=i+1; ii<xi; ii++) for(int jj=1; jj<xj; jj++) iob[1][ii][jj]=true;
					IJ.resetEscape();
					if (!Canceled) errString+=", Escape pressed";
					Canceled=true;
					keepSet = true;
				}
			} catch (NoSuchMethodError ex) {}
			IJ.showStatus("Processsing: "+nhl[i]+j+".tif");
			imp0=openImage(""+directory,nhl[i]+j+".tif");
			// int  allCount = 0;
			// for (int ii=0; ii<=i; ii++) allCount+=counter[1][ii];
			if (null==imp0) {
				if (verboseMode && !Canceled) IJ.error("Failed to read '"+nhl[i]+j+".tif'!");
				if (!Canceled) errString+=", Failed to read '"+nhl[i]+j+".tif'";
				iob[1][i][j]=true;
			} else if (0==Sz) {
				if (MinStack > imp0.getStackSize()) {
				// if (MinStack > Sz) {
					if (verboseMode && !Canceled) IJ.error("File '"+nhl[i]+j+".tif' is not a stack or has too small stack size!");
					if (!Canceled) errString+=(", File '"+nhl[i]+j+".ti	f' is not a stack or has too small stack size");
					imp0 = CloseImageWindow(imp0);
					iob[1][i][j]=true;
				} else {
					// width  = imp0.getWidth();
					// height = imp0.getHeight();
					framecnt = imp0.getStackSize();
					// Sz = framecnt;
					for (int ni=0;ni<tps;ni++) if (filesize==filesizeA[ni]) framecntA[ni]=framecnt;
					if(responsefirst>framecnt) {
						if (!silentMode) IJ.error("Error: Stacks are too small\nLower frame number of response is higher than stack size !"); 
						imp0 = CloseImageWindow(imp0);
						//  ## save settings here?
						IJ.showStatus("restarting the " + PluginTitle);
						run(null);  // --  run plugin again
						return;
					}
					if(responselast > framecnt) {
						if (!silentMode) {
							// -- stack to small; ask user what to do
							String Msg="Stacks are smaller than the highest frame\n";
							Msg+="number of the 'response' range (" + responselast+").\n";
							Msg+="Do you agree with Averaging until frame number " + framecnt + " only?";;
							switch (JOptionPane.showConfirmDialog(null, Msg, PluginTitle,JOptionPane.YES_NO_CANCEL_OPTION, JOptionPane.WARNING_MESSAGE)) {
								case JOptionPane.YES_OPTION:
									break;
								case JOptionPane.NO_OPTION:
									responselast = framecnt;
									if (responsefirst>responselast) responsefirst=responselast;
									if (baselinelast>responsefirst-1) baselinelast=responsefirst-1;
									if (baselinefirst>baselinelast) baselinefirst=baselinelast;
									imp0 = CloseImageWindow(imp0);
									// -- save settings for next run
									if (DONT_KEEP==storeModus) {
										try {
											Prefs.set("VSDproc_store.int", (double)storeModus);
											Prefs.set("VSDproc_lastRun.long", ""+System.currentTimeMillis());
										} catch (NoSuchMethodError e) {}
									} else gotSet=findPathAndDoSettings (Pth, SET, false);
									IJ.showStatus("restarting the " + PluginTitle);
									run(null);  // --  run plugin again
									return;
								default:
									imp0 = CloseImageWindow(imp0);
									return;
							}
						} // if (!silentMode)
						responselast = framecnt;
						if (responsefirst > responselast) {
							if (!silentMode) {
								String Msg="End of 'response' range is lower than its beginning.\n";
								Msg+="Do you agree with taking frame "+ framecnt + " only for 'response'?";;
								switch (JOptionPane.showConfirmDialog(null, Msg, PluginTitle,JOptionPane.YES_NO_CANCEL_OPTION, JOptionPane.WARNING_MESSAGE)) {
									case JOptionPane.YES_OPTION:
										break;
									case JOptionPane.NO_OPTION:
										responsefirst=responselast;
										if (baselinelast>responsefirst-1) baselinelast=responsefirst-1;
										if (baselinefirst>baselinelast) baselinefirst=baselinelast;
										imp0 = CloseImageWindow(imp0);
										// -- save settings for next run
										if (DONT_KEEP==storeModus) {
											try {
												Prefs.set("VSDproc_store.int", (double)storeModus);
												Prefs.set("VSDproc_lastRun.long", ""+System.currentTimeMillis());
											} catch (NoSuchMethodError e) {}
										} else gotSet=findPathAndDoSettings (Pth, SET, false);
										IJ.showStatus("restarting the " + PluginTitle);
										run(null);  // -- run plugin again
										return;
									default:
										imp0 = CloseImageWindow(imp0);
										return;
								}
							}
							responsefirst=responselast;
						}
						if (baselinelast>responsefirst-1) {
							if (!silentMode) {
								String Msg="The 'baseline' range overlaps with the 'response' range.\n";
								Msg+="Do you agree with limiting 'baseline' range to "+ (responsefirst-1) + "?";;
								switch (JOptionPane.showConfirmDialog(null, Msg, PluginTitle,JOptionPane.YES_NO_CANCEL_OPTION, JOptionPane.WARNING_MESSAGE)) {
									case JOptionPane.YES_OPTION:
										break;
									case JOptionPane.NO_OPTION:
										baselinelast=responsefirst-1;
										if (baselinefirst>baselinelast) baselinefirst=baselinelast;
										imp0 = CloseImageWindow(imp0);
										if (DONT_KEEP==storeModus) {
											try {
												Prefs.set("VSDproc_store.int", (double)storeModus);
												Prefs.set("VSDproc_lastRun.long", ""+System.currentTimeMillis());
											} catch (NoSuchMethodError e) {}
										} else gotSet=findPathAndDoSettings (Pth, SET, false);
										IJ.showStatus("restarting the " + PluginTitle);
										run(null);  // run plugin again
										return;
									default:
										imp0 = CloseImageWindow(imp0);
										return;
								}
							}
							baselinelast = responsefirst-1;
						}
						if (baselinefirst>baselinelast) {
							if (!silentMode) {
								String Msg="End of 'baseline' range is lower than its beginning.\n";
								Msg+="Do you agree with taking frame "+ baselinelast + " only for 'baseline'?";;
								switch (JOptionPane.showConfirmDialog(null, Msg, PluginTitle,JOptionPane.YES_NO_CANCEL_OPTION, JOptionPane.WARNING_MESSAGE)) {
									case JOptionPane.YES_OPTION:
										break;
									case JOptionPane.NO_OPTION:
										baselinefirst=baselinelast;
										imp0 = CloseImageWindow(imp0);
										if (DONT_KEEP==storeModus) {
											try {
												Prefs.set("VSDproc_store.int", (double)storeModus);
												Prefs.set("VSDproc_lastRun.long", ""+System.currentTimeMillis());
											} catch (NoSuchMethodError e) {}
										} else gotSet=findPathAndDoSettings (Pth, SET, false);
										IJ.showStatus("restarting the " + PluginTitle);
										run(null);  // run plugin again
										return;
									default:
										imp0 = CloseImageWindow(imp0);
										return;
								}
							}
							baselinefirst = baselinelast;
						}
					} // if(responselast > framecnt)
					width  = imp0.getWidth();
					height = imp0.getHeight();
					Sz = framecnt;
					fileDate=getFileDate(new File(directory,nhl[i]+j+".tif"));
				} // if (MinStack > imp0.getStackSize()) {} else
			} else if (Sz!=imp0.getStackSize()) {
				if (!silentMode && !Canceled) IJ.error("Stack size of '"+nhl[i]+j+".tif' does not match the one of loaded images!");
				if (!Canceled) errString+=(", Stack size of '"+nhl[i]+j+".tif' does not match the one of loaded images");
				imp0 = CloseImageWindow(imp0);
				iob[1][i][j]=true;
			}
			if (!iob[1][i][j]) if (width != imp0.getWidth() || height != imp0.getHeight()) {
				if (!silentMode && !Canceled) IJ.error("Size missmatch in '"+nhl[i]+j+".tif' with loaded images!");
				if (!Canceled) errString+=(", Size missmatch in '"+nhl[i]+j+".tif' with loaded images");
				imp0 = CloseImageWindow(imp0);
				iob[1][i][j]=true;
			}
			try {
				if (IJ.escapePressed()) {
					IJ.showStatus("canceled.");
					IJ.resetEscape();
					Canceled=true;
					keepSet = true;
					iob[1][i][j]=true;
				}
			} catch (NoSuchMethodError ex) {}
			if (!iob[1][i][j]) {
				if (i==0) showProgress((double)j/(double)(xj*xi));
				else  showProgress((double)j+xj*(i-1)/(double)(xj*xi));
				IJ.showStatus("Processsing: "+nhl[i]+j+".tif");
				if (imp0.getType()==ImagePlus.GRAY32) was32bit = true;
				
				// Estimate the memory limitations and calculate the requirements
				try {
					double freeMemory = (double) (IJ.maxMemory() - IJ.currentMemory());
					double memPerFrame = (double) 8 * imp0.getWidth() * imp0.getHeight();
					// -- consider memory required for the average stack
					freeMemory -= memPerFrame*(double)Sz;
					// -- consider memory required for the SD stack
					if (AMethod==SD_METHOD || AMethod==SE_METHOD) freeMemory -= memPerFrame*(double)Sz;
					if (ShowPStack || SavePStack) freeMemory -= memPerFrame * 2 * (double) counter[0][i];
					double memPerTempImage = memPerFrame * (double) counter[0][i];
					imagesToSpool = (int) Math.ceil (((double) Sz+1) - 0.95 * freeMemory / memPerTempImage) ;
					// -- in case of AVG method, no temp images are needed; 
					// 	next equation corrects the above one for this case. 
					if (AMethod==AVG_METHOD) imagesToSpool -= Sz;
					if (0 > imagesToSpool || 1== counter[0][i]) imagesToSpool = 0;
					if (Sz < imagesToSpool) {
						if (!silentMode) IJ.error("Error: not enough memory!");
						errString+=(", Not enough memory  for processing "+nhl[i]+j);
						Canceled=true;
						keepSet = true;
						iob[1][i][j]=true;
					}
				}
				catch  (NoSuchMethodError ex) {
					//  -- workaround for old ImageJ versions (< 1.33u)
					imagesToSpool = Sz;
				}
				if (ShowRaw && !Canceled) {
					ImagePlus imp0R =null;
					if (was32bit && (!blankArtifact || !allowBlankArtifact)) imp0R = imp0;
					else imp0R = duplicateStack(imp0);
					imp0R.setTitle(nhl[i]+j+FTag+".tif");
					imp0R.show();
					ScaleTheMiddle(imp0R,(responsefirst+responselast)/2, ScalingD);
					imp0R.setSlice((responsefirst+responselast)/2);
					ZoomGood(imp0R);
					imp0R.updateAndRepaintWindow();
				}
				if (!was32bit && !Canceled) iob[1][i][j]=(!CovertTo32bitStack(imp0));
			}
			if (!iob[1][i][j]) iob[1][i][j]=null==imp0;
			if (!iob[1][i][j] && blankArtifact && allowBlankArtifact) {
				//private void blankArtifactInStack(ImagePlus img1, int fromZ, int toZ) {
				blankArtifactInStack(imp0, blankFrom, blankTo);
				if (null==imp0) {
					// ######  Error msg....
					iob[1][i][j]=true;
				}
			}
			if (!iob[1][i][j]) {
				if (!was32bit && (ShowQual||SaveQual)) {
					// -- 32-bit inputs to be analyzed after interpolation (below)
					if (null==qv) qv=new double [xi][iob[1][0].length][Sz+1];
					double [] qq=getZProfile(imp0);
					if (null!=qq) if (qq.length==Sz+1) for (int k=1; k<qq.length; k++) qv[i][j][k]=qq[k]; 
				}
				if (((ShowPStack && (counter[0][i]>1 || !ShowPerc)) || (SavePStack && (counter[0][i]>1 || !SavePerc))) && (i>0 || !hasControl || 1==nhl.length) && !Canceled) {
					IJ.showStatus("Projection for trial overview ("+nhl[i]+")");
					ImagePlus impBp, impRp;
					if (AMethod==SD_METHOD || AMethod==SE_METHOD) {
						impBp = ZProjStack(imp0, baselinefirst, baselinelast, AVG_METHOD, was32bit);
						impRp = ZProjStack(imp0, responsefirst, responselast, AVG_METHOD, was32bit);
					} else {
						impBp = ZProjStack(imp0, baselinefirst, baselinelast, AMethod, was32bit);
						impRp = ZProjStack(imp0, responsefirst, responselast, AMethod, was32bit);
					}
					if (!Canceled && null != impBp && null != impRp) {
						String Ns = (null!=impNo && Normalizeb && hasControl && 1<nhl.length)?"n":"";
						if (counter[1][i]==0) {
							impMB = impBp;
							impMR = impRp;
							impMB.setTitle(nhl[i]+"_mbase"+Ns+FTag+".tif");
							impMR.setTitle(nhl[i]+"_mresp"+Ns+FTag+".tif");
							// -- set slice name (base)
							ImageStack bstack = impMB.createEmptyStack();
							bstack.addSlice(nhl[i]+j+"_base"+Ns, impMB.getProcessor());
							impMB.setStack(null, bstack);
							// -- set slice name (resp.); will become %-change later
 							ImageStack rstack = impMR.createEmptyStack();
							rstack.addSlice(nhl[i]+j+"_div"+Ns+"_%change", impMR.getProcessor());
							impMR.setStack(null, rstack);
						} else if (null != impMB && null != impMR) {
							try {
								ImageStack stackMb = impMB.getStack();
								ImageStack stackMr = impMR.getStack();
								stackMb.addSlice(nhl[i]+j+"_base"+Ns, impBp.getProcessor());
								stackMr.addSlice(nhl[i]+j+"_div"+Ns+"_%change", impRp.getProcessor());
								impMB.setStack(null, stackMb);
								impMR.setStack(null, stackMr);
							}
							catch(OutOfMemoryError e) {
								if (!silentMode && !Canceled) IJ.error("Failed to extend m-image stack " + 
									"with ''"+imp0.getTitle()+"'':\nNot enough memory!");
								if (!Canceled) errString+=(", Failed to extend m-image stack " + 
									"with ''"+imp0.getTitle()+"'': not enough memory!");
								impMB=CloseImageWindow(impMB);
								impMR=CloseImageWindow(impMR);
							}
							catch(NullPointerException e) {
								if (!silentMode && !Canceled) IJ.error("Failed to extend m-image stack " + 
									"with ''"+imp0.getTitle()+"'':\nImage got lost somehow!");
								if (!Canceled) errString+=(", Failed to extend m-image stack " + 
									"with ''"+imp0.getTitle()+"'': image got lost somehow!");
								impMB=CloseImageWindow(impMB);
								impMR=CloseImageWindow(impMR);
							}
						} else {
							impMB=CloseImageWindow(impMB);
							impMR=CloseImageWindow(impMR);
						}
					} else {
						impMB=CloseImageWindow(impMB);
						impMR=CloseImageWindow(impMR);
					}
					if (counter[1][i] > 0) {
						impBp=CloseImageWindow(impBp);
						impRp=CloseImageWindow(impRp);
					}
				}  // if (((ShowPStack && (counter[0][i]>1 || !ShowPerc)) || (SavePStack && (counter[0][i]>1 || !SavePerc))) && (i>0 || !hasControl || 1==nhl.length) && !Canceled)
				
				// -- average stacks or rearange them for z-projection
				if (counter[1][i]==0 && AMethod == AVG_METHOD) {
					if (ShowRaw && was32bit) imp1=duplicateStack(imp0);
					else imp1= imp0;
					if (null!=imp1) {
						// -- check whether single pixeles have value NaN or infinity; 
						//	optionally, interpolate from X,Y-neighborhood if pixel values are NaN or Infinity.
						if (was32bit) {
							// ######  IJ status ....
							if (interpolateXY) {
								// -- interpolate NaN and Infinity values
								for (int fi=1; fi<=Sz; fi++) preFilter (imp1.getStack().getProcessor(fi), FilterMean, FilterXy);
								// -- use interpolated image for Qual calculation
								if (ShowQual||SaveQual) {
									if (null==qv) qv=new double [xi][iob[1][0].length][Sz+1];
									double [] qq=getZProfile(imp1);
									if (null!=qq) if (qq.length==Sz+1) for (int k=1; k<qq.length; k++) qv[i][j][k]=qq[k];
								}
							} else if (ShowQual||SaveQual) {
								ImagePlus imp11 = duplicateStack(imp1);
								// -- interpolate duplicate
								for (int fi=1; fi<=Sz; fi++) preFilter (imp11.getStack().getProcessor(fi), FilterMean, FilterXy);
								// -- use interpolated duplicate for Qual calculation
								if (null==qv) qv=new double [xi][iob[1][0].length][Sz+1];
								double [] qq=getZProfile(imp11);
								if (null!=qq) if (qq.length==Sz+1) for (int k=1; k<qq.length; k++) qv[i][j][k]=qq[k]; 
								imp11 = CloseImageWindow(imp11);
							}
						}
						if (AMethod==SD_METHOD || AMethod==SE_METHOD) {
							String Ns="";						
							if (0 < i && null!=impNo && Normalizeb && hasControl && 1<nhl.length) {
								Ns="n";
								Icalculate(impNo,imp1,1,0,DIVIDE1);
							}
							if (null != imp1) imp1.setTitle(nhl[i]+AVS[ZProjector.AVG_METHOD]+Ns+FTag+".tif");
						} else imp1.setTitle(nhl[i]+AVS[AMethod]+FTag+".tif");
						imp1.changes = true;
						if (AMethod==AVG_METHOD) {
							if ((hasControl && 1<nhl.length && (0==i)?ShowNo:ShowAv)) {
								imp1.show();
								ScaleTheMiddle(imp1,(responsefirst+responselast)/2, ScalingD);
								imp1.setSlice((responsefirst+responselast)/2);
								ZoomGood(imp1);
								imp1.updateAndRepaintWindow();
							}
						}
					} else {
						if (verboseMode &&  !Canceled) IJ.error("Unexpected Error (imp0 -> imp1): "+nhl[i]+j);
						if (!Canceled) errString+=( "Unexpected Error (imp0 -> imp1): "+nhl[i]+j);
					}
				} else if ((null != imp1 || AMethod  != AVG_METHOD) && null != imp0 && !Canceled) {
					if (was32bit) {
						if (interpolateXY) {
							// -- interpolate NaN and Infinity values
							for (int fi=1; fi<=Sz; fi++) preFilter (imp0.getStack().getProcessor(fi), FilterMean, FilterXy);
							if (ShowQual||SaveQual) {
								if (null==qv) qv=new double [xi][iob[1][0].length][Sz+1];
								double [] qq=getZProfile(imp0);
								if (null!=qq) if (qq.length==Sz+1) for (int k=1; k<qq.length; k++) qv[i][j][k]=qq[k];
							}
						} else if (ShowQual||SaveQual) {
							ImagePlus imp01 = duplicateStack(imp0);
							// -- interpolate NaN and Infinity values in copy
							for (int fi=1; fi<=Sz; fi++) preFilter (imp01.getStack().getProcessor(fi), FilterMean, FilterXy);
							if (null==qv) qv=new double [xi][iob[1][0].length][Sz+1];
							double [] qq=getZProfile(imp01);
							if (null!=qq) if (qq.length==Sz+1) for (int k=1; k<qq.length; k++) qv[i][j][k]=qq[k]; 
							imp01 = CloseImageWindow(imp01);
						}
					}
										
					if (AMethod == AVG_METHOD) {
						if (was32bit && !interpolateXY) {
							// -- set value to 0.0 and register x,y,z position if pixel values are NaN or Infinity.  
							// 	-- in first stack if we have a second one
							if (1==counter[1][i]) IcalculateSingleSlow(imp1, 1.0, 0.0, BYPASS);
							// 	-- at adding calcluation
							IcalculateSlow (imp0,imp1,1,0,ADD);
						} else {
							// -- normalize here if it is the last image of the series to add for sure 
							// (otherwise counter[2] will be larger than 1, causing later normalization.
							if (j==iob[1][0].length-1) {
								//  --  dviding by the number of summed images to obtain the average among them
								Icalculate(imp0,imp1,1/(double)(counter[2][i]+1),0,ADD);
								counter[2][i]=0;
							} else Icalculate (imp0,imp1,1,0,ADD);
						}
						if (!Canceled) if ((hasControl && 1<nhl.length && (0==i))?ShowNo:ShowAv) ScaleTheMiddle(imp1,(responsefirst+responselast)/2, ScalingD);
					} else {
						if ((AMethod==SD_METHOD || AMethod==SE_METHOD) && 0 < i && null!=impNo && Normalizeb && hasControl && 1<nhl.length) {
							ImagePlus imp00=null;
							if (ShowRaw && was32bit) imp00=duplicateStack(imp0);
							else imp00= imp0;
							if (null != imp00) Icalculate(impNo,imp00,1,0,DIVIDE1);
							if (null != imp00) {
			 					if (null == tmpFileNames) {
									tmpImp = new ImagePlus[Sz-imagesToSpool];
									tmpFileNames = getTmpFileNames(imagesToSpool);
								}
								if (!distributeStackToTmpFile(imp00, tmpImp, tmpFileNames)) imp00 = CloseImageWindow(imp00);
							} else {
								if (verboseMode &&  !Canceled) IJ.error("Unexpected Error SDvSE (imp0 -> imp1): "+nhl[i]+j);
								if (!Canceled) errString+=( "Unexpected Error SDvSE (imp0 -> imp1): "+nhl[i]+j);
								Canceled=true;
							}
							if (ShowRaw && was32bit) imp00 = CloseImageWindow(imp00);
						} else {
				 			if (null == tmpFileNames) {
								tmpImp = new ImagePlus[Sz-imagesToSpool];
								tmpFileNames = getTmpFileNames(imagesToSpool);
							}
							if (!distributeStackToTmpFile(imp0, tmpImp, tmpFileNames)) imp0 = CloseImageWindow(imp0);
						}
					}
					if (!ShowRaw || !was32bit || interpolateXY) imp0 = CloseImageWindow(imp0);
				}
				if (!Canceled && (null != imp1 || AMethod != AVG_METHOD)) {
					counter[1][i]++; 
					counter[2][i]++;
				}
			} // if (!iob[1][i][j])
			if (Canceled) {
				if (ShowRaw) {
					imp0.setTitle(nhl[i]+j+FTag+".tif");
					ScaleTheMiddle(imp0,(responsefirst+responselast)/2, ScalingD);
					imp0.setSlice((responsefirst+responselast)/2);
					ZoomGood(imp0);
					imp0.updateAndRepaintWindow();
				} else imp0=CloseImageWindow(imp0);
				for(int jj=j; jj<xj; jj++)iob[1][i][jj]=true;
				for(int ii=i+1; ii<xi; ii++) for(int jj=1; jj<xj; jj++) iob[1][ii][jj]=true;
			}
		} // for(int j=1; j<xj; j++) if (!iob[1][i][j]) {
		
		// --------------- average and normalize image data; start calculations ------------------------------
		if (0<counter[1][i] && !Canceled) {
			if (AMethod == AVG_METHOD) {
				if (1<counter[2][i]) {
					if (null == failureIdxList) IcalculateSingle(imp1,(double)counter[2][i],0,SCALEINV);
					else {
						// int width  = imp1.getWidth();
						// int height = imp1.getHeight();
						double v1=(double)counter[2][i];
						double v2= 0.0;
						ImageStack stK = new ImageStack(width, height);
						FloatProcessor ipK = null;
						for (int z=1; z<=Sz; z++) {
							ipK = new FloatProcessor (width, height);   // ### use createProcessor() method?
							for (int x=0; x<width; x++) for (int y=0; y<height; y++) ipK.putPixelValue(x, y, v1);
							for (int fi = 0; fi < failureIdxList[0].length; fi++) if (z==failureIdxList[2][fi]) {
								if (failureIdxList[3][fi] >= counter[2][i]) v2=Double.NaN;
								else  v2 = v1 - (double)failureIdxList[3][fi];
								ipK.putPixelValue(failureIdxList[0][fi], failureIdxList[1][fi], v2);
							}
							stK.addSlice("", ipK);
						}
						ImagePlus impK = new ImagePlus("normalization image", stK);
						Icalculate (impK, imp1,1,0, DIVIDE1);
						impK = CloseImageWindow(impK);
					}
					if (Canceled) {for(int ii=i+1; ii<xi; ii++) for(int jj=1; jj<xj; jj++) iob[1][ii][jj]=true;}
					else if ((hasControl && 1<nhl.length && (0==i))?ShowNo:ShowAv) ScaleTheMiddle(imp1,(responsefirst+responselast)/2, ScalingD);
				}
			} else {
				String Ns="";						
				if (0 < i && null!=impNo && Normalizeb && hasControl && 1<nhl.length) Ns="n";
				if (AMethod==SD_METHOD || AMethod==SE_METHOD) {
					if (ShowDelta || SaveDelta || ((hasControl && 1<nhl.length && (0==i))? (ShowNo || SaveNo) : (ShowAv || SaveAv))) {
						impS = getProjectionFromTmpImg(tmpImp, tmpFileNames, SD_METHOD, was32bit && !interpolateXY);
						if (null != impS) {
							impS.setTitle(nhl[i]+AVS[AMethod]+Ns+FTag+".tif");
							if (AMethod==SE_METHOD && 1 < counter[1][i]) IcalculateSingle(impS,Math.sqrt((double)counter[1][i]),0,SCALEINV);
						}
						if (null != impS) {
							if ((hasControl && 1<nhl.length && (0==i)?ShowNo:ShowAv)) {
								impS.show();
								ScaleTheMiddle(impS,(responsefirst+responselast)/2, ScalingD);
								impS.setSlice((responsefirst+responselast)/2);
								ZoomGood(impS);
								impS.updateAndRepaintWindow();
							}
						} else {
							if (!silentMode && !Canceled) IJ.error("Failed to calculate "+AVS[AMethod]+" of ''"+nhl[i]+"''!"); 
							if (!Canceled) errString+=(", Failed to calculate "+AVS[AMethod]+" of: "+nhl[i]);
							// Canceled = true;
						}
					}
					
					imp1 = getProjectionFromTmpImg(tmpImp, tmpFileNames, AVG_METHOD, was32bit && !interpolateXY);
					deleteTmpFiles(tmpFileNames);
					for (int tt=0; tt<tmpImp.length; tt++) CloseImageWindow(tmpImp[tt]);
					tmpImp = null;
					
					if (null != imp1) {
						imp1.setTitle(nhl[i]+AVS[AVG_METHOD]+Ns+FTag+".tif");
						if ((hasControl && 1<nhl.length && (0==i)?ShowNo:ShowAv)) {
							imp1.show();
							ScaleTheMiddle(imp1,(responsefirst+responselast)/2, ScalingD);
							imp1.setSlice((responsefirst+responselast)/2);
							ZoomGood(imp1);
							imp1.updateAndRepaintWindow();
						}
					} else {
						if (!silentMode && !Canceled) IJ.error("Failed to calculate average of ''"+nhl[i]+"''!"); 
						if (!Canceled) errString+=(", Failed to calculate average of : "+nhl[i]);
						// Canceled = true;
					}
				} else  {
					imp1 = getProjectionFromTmpImg(tmpImp, tmpFileNames, AMethod, was32bit && !interpolateXY);
					deleteTmpFiles(tmpFileNames);
					for (int tt=0; tt<tmpImp.length; tt++) CloseImageWindow(tmpImp[tt]);
					tmpImp = null;
					if (null != imp1) {
						imp1.setTitle(nhl[i]+AVS[AMethod]+Ns+FTag+".tif");
						if ((hasControl && 1<nhl.length && (0==i)?ShowNo:ShowAv)) {
							imp1.show();
							ScaleTheMiddle(imp1,(responsefirst+responselast)/2, ScalingD);
							imp1.setSlice((responsefirst+responselast)/2);
							ZoomGood(imp1);
							imp1.updateAndRepaintWindow();
						}
					} else {
						if (!silentMode && !Canceled) IJ.error("Failed to calculate "+AVL[AMethod]+" projection for ''"+nhl[i]+"''!"); 
						if (!Canceled) errString+=(", Failed to calculate "+AVL[AMethod]+" projection for : "+nhl[i]);
						// Canceled = true;
					}
				}
				if (Canceled) {
					for(int ii=i+1; ii<xi; ii++) for(int jj=1; jj<xj; jj++) iob[1][ii][jj]=true;
				} else {
					if ((hasControl && 1<nhl.length && (0==i))?ShowNo:ShowAv) ScaleTheMiddle(imp1,(responsefirst+responselast)/2, ScalingD);
					if (null != impS) if ((hasControl && 1<nhl.length && (0==i))?ShowNo:ShowAv) ScaleTheMiddle(impS,(responsefirst+responselast)/2, ScalingD);
				}
			}
			if (null!=imp1) {
				// ---------------  show process status  -----------------
				if (!Canceled) IJ.showStatus("Processsing: "+nhl[i]+AVS[AMethod]+FTag+".tif");
				// ---------------  optionally, apply Gaussian / Mean filter -----------------
				if (0<FilterXy && DoFilterXY && DoFilterAv) {
					ImageStack imp1stack = imp1.getStack();
					if (was32bit && !interpolateXY) for (int fi=1; fi<=Sz; fi++) preFilter (imp1stack.getProcessor(fi), FilterMean, FilterXy);
					if (!FilterMean) {for (int fi=1; fi<=Sz; fi++) new GaussianBlur().blur(imp1stack.getProcessor(fi),FilterXy);}
					else {
						RankFilters rankFilter=new RankFilters();
						for (int fi=1; fi<=Sz; fi++) rankFilter.rank(imp1stack.getProcessor(fi), Math.rint(FilterXy), RankFilters.MEAN);
					}
					imp1.setStack(null, imp1stack);
					if (hasControl && 1<nhl.length && (0==i)) {
						if (ShowNo || SaveNo) ScaleTheMiddle(imp1,(responsefirst+responselast)/2,ScalingD);
					} else {
						if (ShowAv || SaveAv) ScaleTheMiddle(imp1,(responsefirst+responselast)/2,ScalingD);
					}
					if (AMethod==SD_METHOD || AMethod==SE_METHOD) if (null!=impS) {
						ImageStack impSstack = impS.getStack();
						if (was32bit && !interpolateXY) for (int fi=1; fi<=Sz; fi++) preFilter (impSstack.getProcessor(fi), FilterMean, FilterXy);
						if (!FilterMean) {for (int fi=1; fi<=Sz; fi++) new GaussianBlur().blur(impSstack.getProcessor(fi),FilterXy);}
						else {
							RankFilters rankFilter=new RankFilters();
							for (int fi=1; fi<=Sz; fi++) rankFilter.rank(impSstack.getProcessor(fi), Math.rint(FilterXy), RankFilters.MEAN);
						}
						impS.setStack(null, impSstack);
						if (hasControl && 1<nhl.length && (0==i)) {
							if (ShowNo || SaveNo) ScaleTheMiddle(impS,(responsefirst+responselast)/2,ScalingD);
						} else {
							if (ShowAv || SaveAv) ScaleTheMiddle(impS,(responsefirst+responselast)/2,ScalingD);
						}
					}
				}
				// ---------------  do the calculation(s)  -----------------
				if (0==i && 1<nhl.length && hasControl) {
					if (0<FilterZ && DoZFilterAv) impNo = ZFilterT(imp1,FilterZ, was32bit && !interpolateXY);
					else impNo = imp1;
					if (null==impNo) {
						IJ.showStatus("Failed to process: "+nhl[i]);
						if (!silentMode && !Canceled) IJ.error("Failed to process: ''"+nhl[i]+"''!"); 
						if (!Canceled) errString+=(", Failed to process: "+nhl[i]);
						return;
					}
				} else {
					if (!Canceled) impL[i] =CalculateStim(imp1, impNo, impS, directory, nhl[i], FTag, was32bit && !interpolateXY);
				}
				try {
					if (IJ.escapePressed()) {
						IJ.showStatus("canceled.");
						IJ.resetEscape();
						Canceled=true;
						keepSet = true;
					}
				} catch (NoSuchMethodError ex) {}
				// ---------------  scale, save, and display  -----------------
				if (0<i || 1==nhl.length || !hasControl) {
					String Ns="";
					int sMethod=(AMethod==SD_METHOD || AMethod==SE_METHOD)?AVG_METHOD:AMethod;
					if (null!=impNo && Normalizeb && hasControl && 1<nhl.length) Ns="n";
					if (ShowAv || SaveAv) ScaleTheMiddle(imp1,(responsefirst+responselast)/2,ScalingD);
					String AvName = nhl[i]+AVS[sMethod]+Ns+FTag+".tif";
					if (1== counter[1][i]) {
						int jj=0;
						for(int j=1; j<xj; j++) if (!iob[1][i][j]) jj=j;
						if (0<jj) AvName = nhl[i]+jj+Ns+FTag+".tif";
					}
					if (SaveAv) {
						new FileSaver(imp1).saveAsTiffStack(directory+BSL+AvName);
						DidSave=true;
					}
					IJ.showStatus("Processsing: "+AvName);
					if (ShowAv) {
						imp1.setTitle(AvName);
						imp1.changes=!SaveAv;
					} else if (!ShowDelta && !SaveDelta && i!=0) imp1 = CloseImageWindow(imp1);
					if (AMethod==SD_METHOD || AMethod==SE_METHOD) if (null!=impS) {
						if (ShowAv || SaveAv) ScaleTheMiddle(impS,(responsefirst+responselast)/2,ScalingD);
						if (SaveAv) {
							new FileSaver(impS).saveAsTiffStack(directory+BSL+nhl[i]+AVS[AMethod]+Ns+FTag+".tif");
							DidSave=true;
						}
						IJ.showStatus("Processsing: "+nhl[i]+AVS[AMethod]+Ns+FTag+".tif");
						if (ShowAv) {
							impS.setTitle(nhl[i]+AVS[AMethod]+Ns+FTag+".tif");
							impS.changes=!SaveAv;
						} else if (!ShowDelta && !SaveDelta && i!=0) impS = CloseImageWindow(impS);
					}
				}
				if (0==i && 1<nhl.length && hasControl &&  (AMethod==SD_METHOD || AMethod==SE_METHOD) && null !=impS) {
					if (SaveNo) {
						new FileSaver(impS).saveAsTiffStack(directory+BSL+nhl[0]+AVS[AMethod]+FTag+".tif");
						DidSave=true;
					}
					IJ.showStatus("Processsing: "+nhl[0]+AVS[AMethod]+FTag+".tif");
					if (ShowNo) {
						impS.changes=!SaveNo;
						impS.setTitle(nhl[0]+AVS[AMethod]+FTag+".tif");
						ScaleTheMiddle(impS,(responsefirst+responselast)/2,ScalingD);
					} else impS = CloseImageWindow(impS);
				}
			}  else {
				for (int j=1; j<xj; j++) iob[1][i][j]=true;
				counter[1][i] = 0;
				counter[2][i] = 0;
			} // if (null!=imp1) {} else {
			
			// -- process the Multiple %-Stack (trial overview)
			if (((ShowPStack && (counter[0][i]>1 || !ShowPerc)) || (SavePStack && (counter[0][i]>1 || !SavePerc))) && !Canceled && (i>0 || !hasControl || 1==nhl.length)) {
				if (!Canceled && null !=impMB && null != impMR) {
					IJ.showStatus("Projection for trial overview: "+nhl[i]);
					int mb = impMB.getStackSize();
					if (commonBase && 1 < mb) {
						ImagePlus impBp = ZProjStack(impMB, 1, mb, AVG_METHOD, was32bit);
						impMB = CloseImageWindow(impMB);
						impMB = impBp;
						// ---------------  optionally, apply Gaussian / Mean filter -----------------
						if (0<FilterXy && DoFilterXY && DoFilterAv) {
							preFilter (impMB.getProcessor(), FilterMean, FilterXy);
							if (!FilterMean) new GaussianBlur().blur(impMB.getProcessor(),FilterXy);
							else {
								RankFilters rankFilter=new RankFilters();
								rankFilter.rank(impMB.getProcessor(), Math.rint(FilterXy), RankFilters.MEAN);
							}
						}
					}  else {
						if (0<FilterXy && DoFilterXY && DoFilterAv) {
							ImageStack MBstack = impMB.getStack();
							if (was32bit && !interpolateXY) for (int fi=1; fi<=mb; fi++) {
								preFilter (MBstack.getProcessor(fi), FilterMean, FilterXy);
							}
							if (!FilterMean) {
								for (int fi=1; fi<=mb; fi++) {
									new GaussianBlur().blur(MBstack.getProcessor(fi),FilterXy);
								}
							} else {
								RankFilters rankFilter=new RankFilters();
								for (int fi=1; fi<=mb; fi++) {
									rankFilter.rank(MBstack.getProcessor(fi), Math.rint(FilterXy), RankFilters.MEAN);
								}
							}
							impMB.setStack(null, MBstack);
						}
					}
					if (!Canceled && 0<FilterXy && DoFilterXY && DoFilterAv) {
						ImageStack MRstack = impMR.getStack();
						if (mb != MRstack.getSize()) {
							IJ.showStatus("Stack size mismatch at projection for trial overview ("+nhl[i]+")!");
							if (!silentMode && !Canceled) IJ.error("Stack size mismatch at projection for trial overview ("+nhl[i]+")!"); 
							if (!Canceled) errString+=(", Stack size mismatch at projection for trial overview ("+nhl[i]+")");
							mb = MRstack.getSize();
						}
						if (was32bit && !interpolateXY) for (int fi=1; fi<=mb; fi++) {
							preFilter (MRstack.getProcessor(fi), FilterMean, FilterXy);
						}
						if (!FilterMean) {
							for (int fi=1; fi<=mb; fi++) {
								new GaussianBlur().blur(MRstack.getProcessor(fi),FilterXy);
							}
						} else {
							RankFilters rankFilter=new RankFilters();
							for (int fi=1; fi<=mb; fi++) {
								rankFilter.rank(MRstack.getProcessor(fi), Math.rint(FilterXy), RankFilters.MEAN);
							}
						}
						impMR.setStack(null, MRstack);
					}
					String Ns="";
					if (!Canceled && Normalizeb && hasControl && 1<nhl.length && null!=impNo) {
						ImagePlus impBp, impRp;
						if (AMethod==SD_METHOD || AMethod==SE_METHOD) {
							impBp = ZProjStack(impNo, baselinefirst, baselinelast, AVG_METHOD, was32bit);
							impRp = ZProjStack(impNo, responsefirst, responselast, AVG_METHOD, was32bit);
						} else {
							impBp = ZProjStack(impNo, baselinefirst, baselinelast, AMethod, was32bit);
							impRp = ZProjStack(impNo, responsefirst, responselast, AMethod, was32bit);
						}
						if (null==impBp || null==impRp) {		
							IJ.showStatus("Projection for trial overview ("+nhl[i]+") failed: lost image!");
							if (!silentMode && !Canceled) IJ.error("Projection for trial overview ("+nhl[i]+") failed: lost image!"); 
							if (!Canceled) errString+=(", Projection for trial overview ("+nhl[i]+") failed: lost image");
						} else {
							Icalculate(impBp,impMB,1,0,DIVIDE1);
							Icalculate(impRp,impMR,1,0,DIVIDE1);
							Ns="n";
						}
						impBp = CloseImageWindow(impBp);
						impRp = CloseImageWindow(impRp);
					}
					if (!Canceled && null != impMB && null != impMR) {
						impMR.setTitle(nhl[i]+"_div"+Ns+"_%Stack"+FTag+".tif");
						Icalculate(impMB, impMR, 1, 0, DIVIDE1);
						if (!Canceled && null != impMR) {
							IcalculateSingle(impMR,100,-1,SCALEREV);
							ScaleTheMiddle(impMR,1,ScalingD);
							if (SavePStack) {
								if (1 == impMR.getStackSize()) new FileSaver(impMR).saveAsTiff(directory+BSL+nhl[i]+"_div"+Ns+"_%Stack"+FTag+".tif");
								new FileSaver(impMR).saveAsTiffStack(directory+BSL+nhl[i]+"_div"+Ns+"_%Stack"+FTag+".tif");
							}
							if (ShowPStack) {
								impMR.changes=!SavePStack;
								impMR.show();
								ZoomGood(impMR);
							} else impMR = CloseImageWindow(impMR);
						} else impMR=CloseImageWindow(impMR);
					} else impMR=CloseImageWindow(impMR);
				} else impMR=CloseImageWindow(impMR);
				impMB = CloseImageWindow(impMB);
			} // if ((ShowPStack || SavePStack) && (i>0 || !hasControl || 1==nhl.length) && (counter[0][i]>1 || (!ShowPerc && !SavePerc)))		
		} // -- if (0<counter[1][i] && !Canceled)
		if (null != failureIdxList && null !=imp1) {
			for (int fi = 0; fi < failureIdxList[3].length; fi++) counter[3][i] += failureIdxList[3][fi];
			failureIdxList=null;
		}
	} //  -- for(int i=0; i<xi; i++) if(0<counter[0][i] && (0<i || (0==i && (Normalizeb || SaveNo || ShowNo || !hasControl || 1==nhl.length)))
	
	//  ----------  show status at status bar  ----------
	int  allCount = 0;
	for (int i=0; i<counter[1].length; i++) allCount+=counter[1][i];
	if (0<allCount) IJ.showStatus("All files loaded.");
	else IJ.showStatus("No files loaded.");
	
	//  ----------  save control average (i.e. "no") to file ----------
	if (0<counter[1][0] && hasControl && 1<nhl.length) {
		IJ.showStatus("Processsing: "+nhl[0]+AVS[AMethod]+FTag+".tif");
		if (SaveNo) {
			if (AMethod==SD_METHOD || AMethod==SE_METHOD) new FileSaver(impNo).saveAsTiffStack(directory+BSL+nhl[0]+AVS[AVG_METHOD]+FTag+".tif");
			else new FileSaver(impNo).saveAsTiffStack(directory+BSL+nhl[0]+AVS[AMethod]+FTag+".tif");
			DidSave=true;
		}
		IJ.showStatus("Processsing: "+nhl[0]+AVS[AMethod]+FTag+".tif");
		if (ShowNo) {
			impNo.changes=!SaveNo;
			if (AMethod==SD_METHOD || AMethod==SE_METHOD) impNo.setTitle(nhl[0]+AVS[AVG_METHOD]+FTag+".tif");
			else impNo.setTitle(nhl[0]+AVS[AMethod]+FTag+".tif");
			ScaleTheMiddle(impNo,(responsefirst+responselast)/2,ScalingD);
		} else impNo = CloseImageWindow(impNo);
	}
	
	//  ------  calculation of ratios (i.e.  HI over LO) ----------------
	if ((ShowRatio||SaveRatio) && (2<nhl.length || (!hasControl && 1<nhl.length))) {
		int c1 = (hasControl && 1<nhl.length)?1:0;
		if (null !=impL[c1])  {
			for (int i=c1+1; i < impL.length; i++) if(null !=impL[i]) {
				IJ.showStatus("Processsing: ratio ("+i+" of "+(impL.length-1)+")");
				ImagePlus impQ=null;
				if (ShowResp) impQ=duplicateStack(impL[i]);
				else impQ=impL[i];
				Icalculate(impL[c1],impL[i],1,0,DIVIDE2);
				String Ns="";
				if (0<counter[1][0] && Normalizeb && hasControl && 1<nhl.length) Ns="n";
				if (!Canceled) {
					if (SaveRatio||ShowRatio) ScaleTheMiddle(impQ, 1, ScalingD);
					if (SaveRatio) {
						new FileSaver(impQ).saveAsTiff(directory+BSL+nhl[i]+"_div"+Ns+"_by_"+nhl[c1]+"_32bit"+FTag+".tif");
						DidSave=true;
					}
				}
				if (!Canceled && ShowRatio) {
					impQ.setTitle(nhl[i]+"_div"+Ns+"_by_"+nhl[c1]+"_32bit"+FTag+".tif");
					impQ.changes=!SaveRatio;
					impQ.show();
					ZoomGood(impQ);
				} else impQ = CloseImageWindow(impQ);
			}
			if (!SaveResp&& !ShowResp) impL[c1] = CloseImageWindow(impL[c1]);
		} // if (null !=impL[c1])
	} // if ((ShowRatio||SaveRatio)...
	
	//  ----------  display and/or save "data quallity table"  --------------------
	if (ShowQual||(DidSave && SaveQual)) {
		int cint=0;
		for (int i=0; i<counter[1].length; i++) cint+=counter[1][i];
		if (0<cint) {
			IJ.showStatus("Processsing: result table");
			for(int i=0; i<xi; i++) {
				// calculate mean across j stacks for each frame
				for (int k=1; k<=Sz; k++) {
					double cnj=0;
					for(int j=1; j<xj; j++) if(!iob[1][i][j]) {  // if(0<qv[i][j][k]) {
						//  if(qv[i][0][k]>0) qv[i][0][k]=0;
						qv[i][0][k]+=qv[i][j][k];
						cnj++;
					}
					// ###  use counter[i] instead!! ########  (test!)
					if(cnj>0) qv[i][0][k]/=cnj;
				}
				// calculate stdv across stacks across frames
				for(int j=1; j<xj; j++) if (!iob[1][i][j]) {  //  if (0<qv[i][j][1]) {
					//  qv[i][j][0]=0;
					double cnk=0;
					for(int k=1; k<=Sz; k++) { // if(0<qv[i][j][k]) {
						qv[i][j][0]+=Math.pow(qv[i][j][k]/qv[i][0][k]-1,2.0);
						cnk++;
					}
					// ### use Sz instead!! ########  (test!)
					if(cnk>0) qv[i][j][0]=Math.sqrt(qv[i][j][0]/cnk)*100;
				}
			}
			double [] minq = new double[xi];
			double [] maxq = new double[xi];
			double [][] rankqv = new double[qv.length][qv[0].length];
			xj=qv[0].length;
			if (normQual) for (int i=0; i<xi; i++) if (0<counter[1][i]) {
				minq[i]=Double.MAX_VALUE;
				maxq[i]=Double.MIN_VALUE;
				double cntd=0.0;
				for (int j=1; j<xj; j++) if (!iob[1][i][j]) {  // if (0.0<=qv[i][j][0]) {
					if (minq[i]>qv[i][j][0]) minq[i]=qv[i][j][0];
					if (maxq[i]<qv[i][j][0]) maxq[i]=qv[i][j][0];
					cntd++;
				}
				// ###  use counter[i] instead of cntd !! ########  (test!)
				if (minq[i]!=maxq[i]) maxq[i]-=minq[i];
			}
			if (rankQual) {
				for (int i=0; i<xi; i++) if (0<counter[1][i]) {
					double [] tqv = new double[xj];
					for (int j=1; j<xj; j++) tqv[j] = qv[i][j][0];
					for (int k=0; k<counter[1][i]; k++) {
						double mv=-1.0;
						for (int j=1; j<xj; j++) if (mv < tqv[j]) mv = tqv[j];
						if (mv>-1.0) for (int j=1; j<xj; j++) if (mv == tqv[j]) {
							rankqv [i][j] = k+1;
							tqv[j]=-1;
						}
					}
				}
			}
			TextPanel tp=null;
			boolean QualSaved=false;
			
			String ret = "";
			if (IJ.isWindows()) ret+=(char)13;
			ret +="\n";
			// String ret = "\n";
			// if (IJ.isWindows()) ret =""+(char)13+(char)10;
			
			if (ShowQual) {
				ResultsTable rt =ResultsTable.getResultsTable();
				rt.reset();
				rt.setPrecision(precision);
				for (int j=1; j<xj; j++) {
					boolean vFound =false;
					for (int i=0; i<xi; i++) if (!iob[1][i][j]) vFound =true;
					if (vFound) {
						rt.incrementCounter();
						rt.addValue("index", j);
						for (int i=0; i<xi; i++) if (0<counter[1][i]) rt.addValue(nhl[i]+",%",iob[1][i][j]?-1:qv[i][j][0]);
						if(rankQual) for(int i=0; i<xi; i++) if (0<counter[1][i]) rt.addValue(nhl[i]+",rank", iob[1][i][j]?0:rankqv[i][j]);
						if(normQual) for(int i=0; i<xi; i++) if (0<counter[1][i]) rt.addValue(nhl[i]+",n.",iob[1][i][j]?-1:((qv[i][j][0]-minq[i])/maxq[i]));
					}
				}
				// - show results table. Code implemented to avoid false error messages thrown by the rt class.
				String tableHeadings = rt.getColumnHeadings();
				if (DoAdjQmapWinSizeToScreen && 0>=zoomLevel) FitLogWinSizeToScreen();
				tp = new TextWindow("Data Quality Map"+FTag, "", QmapWinWidth, QmapWinHeight).getTextPanel();
				tp.setColumnHeadings(tableHeadings);
				
				int n = rt.getCounter();
				if (n>0) {
					StringBuffer sb = new StringBuffer(n*tableHeadings.length());
					for (int j=0; j<n; j++)  sb.append(rt.getRowAsString(j)+ret);
					tp.append(new String(sb));
					if (SaveQual && DidSave) QualSaved=SaveToTxtFile(DirIn+BSL+"Data Quality Map"+FTag+".txt", sb.toString()+ret);
				}
			}
			if (!ShowQual || (DidSave && SaveQual && !QualSaved)) {
				//  circumvent the results table if not required or saving failure.
				String Ostr="index";
				for (int i=0; i<xi; i++) if (0<counter[1][i]) Ostr=Ostr+"\t"+nhl[i]+",%";
				if(rankQual) for(int i=0; i<xi; i++) if (0<counter[1][i]) Ostr=Ostr+"\t"+nhl[i]+",rank";
				if(normQual) for(int i=0; i<xi; i++) if (0<counter[1][i]) Ostr=Ostr+"\t"+nhl[i]+",n.";
				for (int j=1; j<xj; j++) {
					boolean vFound =false;
					for (int i=0; i<xi; i++) if (!iob[1][i][j]) vFound =true;
					if (vFound) {
						Ostr=Ostr+ret+j;
						for (int i=0; i<xi; i++) if (0<counter[1][i]) Ostr=Ostr+"\t"+(iob[1][i][j]?"-1":""+Double.valueOf(IJ.d2s(qv[i][j][0],precision)));
						if(rankQual) for(int i=0; i<xi; i++) if (0<counter[1][i]) Ostr=Ostr+"\t"+(iob[1][i][j]?"0":""+Double.valueOf(IJ.d2s(rankqv[i][j],0)));
						if(normQual) for(int i=0; i<xi; i++) if (0<counter[1][i]) Ostr=Ostr+"\t"+(iob[1][i][j]?"-1":""+Double.valueOf(IJ.d2s((qv[i][j][0]-minq[i])/maxq[i],precision)));
					}
				}
				QualSaved=SaveToTxtFile(DirIn+BSL+"Data Quality Map"+FTag+".txt", Ostr+ret);
			}
		}
	}
	
	//  ----------  write log to "Results" file  -----------------
	if ((DidSave&& DoSaveLog) || DoShowLog)  if (!SaveLog(FTag,nhl,runstart)) {
		IJ.showStatus("Log results...");
		if (!silentMode && !Canceled) IJ.error("Failed to save or display log!");
		if (!Canceled) errString+=(", Failed to save or display log");
	}
	
	//  ----------  keep settings for next run in an ini file -----------
	if (!Canceled || keepSet) {
		boolean FilterMeanN=FilterMean;
		if (mustBeMean) FilterMean=FilterMean0;
		if (DONT_KEEP==storeModus) {
			try {
				Prefs.set("VSDproc_store.int", (double)storeModus);
				Prefs.set("VSDproc_lastRun.long", ""+System.currentTimeMillis());
			} catch (NoSuchMethodError e) {}
		} else gotSet=findPathAndDoSettings (Pth, SET, !silentMode);
		FilterMean=FilterMeanN;
	}
	
	//  ----------  repaint images and finish  -----------
	IJ.showStatus("Repaint...");
	if (0<PBwait) IJ.wait(PBwait+1);
	try {WindowManager.repaintImageWindows();}  // 1.34 required
	catch (NoSuchMethodError ex) { 
		int[] list = WindowManager.getIDList();
		if (list!=null) for (int i=0; i<list.length; i++) {
			ImagePlus imp2 = WindowManager.getImage(list[i]);
			if (imp2!=null) {
				ImageWindow win = imp2.getWindow();
				if (win!=null) win.repaint();
			}
		}
	}
	catch (NullPointerException ex) {} 
	showProgress(2.0);
	IJ.showStatus(Canceled?"canceled.":"done.");
}

private boolean SaveLog(String ftag, String[] nnhl, long runstart) {
	String ret = "\n";
	if (IJ.isWindows()) ret =""+(char)13+(char)10;
	String LogStr=PluginTitle;
	LogStr+=", Vers."+LogSpt+PluginVersion+":r"+src+", IJ ";
	try {LogStr+=IJ.getVersion();}
	catch (NoSuchMethodError ex) {LogStr+="<1.32";}
	LogStr+=" - "+DateTimeString(new Date())+ret;
	LogStr=LogStr+"Run time:"+LogSpt+((double)(System.currentTimeMillis()-runstart))/1000+" s"+ret;
	LogStr+="Input directory:"+LogSpt+DirIn+ret;
	if (0<ftag.length()) LogStr+="File Tag:  "+LogSpt+ftag+ret;
	else LogStr+="no File Tag added."+LogSpt+ret;
	
	// ---- report calculation methods used (for projections)
	LogStr+="Calculation method:"+LogSpt;
	switch (AMethod) {
		case AVG_METHOD: LogStr+=" Average stacks"+ret; break;
		case MAX_METHOD: LogStr+=" Maximum"+ret; break;
		case MIN_METHOD: LogStr+=" Minimum"+ret; break;
		case SUM_METHOD: LogStr+=" Sum stacks"+ret; break;
		case SD_METHOD: LogStr+=" Stdev in stacks"+ret; break;
		case MEDIAN_METHOD: LogStr+=" Median of stacks"+ret; break;
		case SE_METHOD: LogStr+=" Sterr in stacks"+ret;
	}
	int fc = 0;
	for(int i=0; i<counter[3].length; i++) fc += counter[3][i];
	if (0 < fc) {
		LogStr+="Pixels ignored:"+LogSpt+ fc;
		if (1<counter[3].length) {
			LogStr +=" (";
			for(int i=0; i<counter[3].length-1; i++) LogStr +=counter[3][i]+", ";
			LogStr +=counter[3][counter[3].length-1];
			LogStr +=")";
		}
		LogStr +=ret;
	}
	if (Normalizeb && hasControl && 1 < nhl.length && 0<counter[1][0]) {
		LogStr+="Subtracted & normalized to"+LogSpt+"'"+nhl[0];
		if(AMethod==SD_METHOD || AMethod==SE_METHOD) LogStr=LogStr+AVS[AVG_METHOD];
		else LogStr=LogStr+AVS[AMethod];
		LogStr+=ftag+".tif'";
		if (!SaveNo) LogStr+=(ShowNo?" (displayed only)":" (not saved)");
		LogStr+=ret;
	}
	
	// ---- report projection parameters (frame number from/to)
	LogStr=LogStr+"Number of processed frames per stack:"+LogSpt+framecnt+ret;
	LogStr=LogStr+"Baseline from:  "+LogSpt+baselinefirst+ret;
	LogStr=LogStr+"Baseline to:       "+LogSpt+baselinelast+ret;
	LogStr=LogStr+"Response from:"+LogSpt+responsefirst+ret;
	LogStr=LogStr+"Response to:    "+LogSpt+responselast+ret;
	
	// ---- report general statistics (hom many files of each...)
	int nni=(!Normalizeb && !SaveNo && !ShowNo && hasControl && 1<nhl.length)?1:0;
	for (int i=nni; i<nnhl.length; i++) if(0<counter[0][i]) LogStr=LogStr+"Processed '"+nnhl[i]+"' files:"+LogSpt+counter[1][i]+" of "+counter[0][i]+ret;
	LogStr=LogStr+"File size:"+LogSpt+filesize+" bytes."+ret;
	LogStr=LogStr+"File date:"+LogSpt+fileDate+ret;
	int ncnt=0;
	nni=1==nni?0:1;
	for (int i=nni; i<nnhl.length; i++) if(0<counter[1][i]) ncnt++;
	
	// ---- report grouped frame projection (filtering in Z dimension)
	if (0<FilterZ && DoZFilterAv) {
		LogStr=LogStr+"Filter in z-direction, +/- :"+LogSpt+FilterZ+ret;
		LogStr=LogStr+"Filtered in z-direction:"+LogSpt;
		if (0<ncnt && DoZFilterAv) {
			int nzcnt=ncnt;
			for (int i=nni; i<nnhl.length; i++) {
				if(0<counter[1][i]) LogStr+="'"+nnhl[i]+"'";
				nzcnt--;
				if (0<nzcnt) LogStr+=", ";
			}
			LogStr+=" averaged stack";
			if (1<ncnt) LogStr+="s";
			if (DoZFilterAv &&0<counter[1][0]) LogStr+="; ";
		}
		if (DoZFilterAv &&0<counter[1][0]) LogStr+="'"+nnhl[0]+"' averaged stack";
		else if (0==ncnt) LogStr+="N/A";
		LogStr+=(" "+ret);
	}
	
	// ---- report filtering (X, Y)
	if (0<FilterXy&&DoFilterXY&&(DoFilterAv||DoFilterBR||(DoFilterDelta&&(ShowDelta||SaveDelta)))) {
		LogStr=LogStr+"Filter in x,y-direction :"+LogSpt;
		if (!FilterMean) LogStr=LogStr+"Filter type:"+LogSpt+"Gaussian, sigma = "+FilterXy+ret;
		else LogStr=LogStr+"Filter type:"+LogSpt+"Mean, radius = "+Math.round(FilterXy)+ret;
		LogStr+="Filtered in x,y-direction:"+LogSpt;
		if(0<counter[1][0]) ncnt++;
		if (0<ncnt ) {
			if (DoFilterAv) {
				int nxcnt=ncnt;
				for (int i=0; i<nnhl.length; i++) {
					if(0<counter[1][i]) LogStr+="'"+nnhl[i]+"'";
					nxcnt--;
					if (0<nxcnt) LogStr+=", ";
				}
				LogStr+=" averaged stack";
				if (1<ncnt) LogStr+="s";
				if (DoFilterBR||(DoFilterDelta&&(ShowDelta||SaveDelta))) LogStr+="; ";
			}
			if (DoFilterBR) {
				LogStr+="baseline and response projections";
				if (DoFilterDelta&&(ShowDelta||SaveDelta)) LogStr+="; ";
			}
			if (DoFilterDelta&&(ShowDelta||SaveDelta)) {
				String fLabel=(0<counter[0][0]&&1<nhl.length&&Normalizeb && hasControl)?"F":" I ";
				LogStr+="'[Delta]"+fLabel;
				LogStr+="/"+fLabel+"(0)' stacks";
			} else if (!DoFilterAv && !DoFilterBR) LogStr+="nothing";
			LogStr+=(".");
		} else LogStr+="nothing!";
		LogStr+=(" "+ret);
	}
	
	// ---- report Multiple %-Ratio Stack (trial overviev) calculation detail
	if (ShowPStack || SavePStack) {
		boolean hasP = false;
		if (hasControl && 1<nhl.length) {
			for (int i=1; i<counter[1].length; i++) if (1 < counter[1][i]) hasP = true;
		} else {
			for (int i=0; i<counter[1].length; i++) if (1 < counter[1][i]) hasP = true;
		}
		if (hasP) {
			LogStr+="Multiple %-Ratio Stack (trial overview):"+LogSpt;
			if (commonBase) LogStr+="calculated to common";
			else LogStr+="calculated to individual";
			LogStr+=(" baseline."+ret);
		}
	}
	
	// ---- report quality map return
	if (ShowQual||SaveQual) {
		LogStr+="Data deviation map:"+LogSpt;
		LogStr+="percentage";
		if (normQual) LogStr+="; normalized values";
		if (rankQual) LogStr+="; value rank";
		if (!SaveQual) LogStr+=ShowQual?" (displayed only)":" (not saved)";
		LogStr+=("."+ret);
	}
	
	// ---- report interpolated frames (artifact removal)
	if (blankArtifact && allowBlankArtifact) {
		if (blankTo > blankFrom) {
			LogStr+="Putative artifact frames interpolated:"+LogSpt;
			LogStr+="from "+blankFrom+" to "+blankTo+"."+ret;
		} else {
			LogStr+="Putative artifact frame interpolated:"+LogSpt;
			LogStr+="frame "+blankFrom+"."+ret;
		}
	}
	
	// ---- report names of included and excluded files
	int excludedFilesCount = 0, includedFilesCount = 0;
	String excludeString="", includeString="";
	int nci=(!Normalizeb && !SaveNo && !ShowNo && hasControl && 1<nhl.length)?1:0;
	for (int i=nci; i<nhl.length; i++) {
		int firstIncluded=0, firstExcluded=0, lastIncluded=0,  lastExcluded=0;
		for (int j=1; j<iob[1][i].length; j++) {
			if (!iob[0][i][j]) {   // file exists
				if (iob[1][i][j]) {   // file ignored or failed to load (="excluded")
					if (0==firstExcluded) {
						firstExcluded = j;
						lastExcluded = j;
					} else {
						if (1 < j-lastExcluded) {
							if (0==excludedFilesCount) excludeString+="Excluded file(s): ";
							else excludeString+=", ";
							if (lastExcluded == firstExcluded) excludeString+=(nhl[i]+firstExcluded);
							else excludeString+=(nhl[i]+firstExcluded+"-"+nhl[i]+lastExcluded);
							excludedFilesCount+=(lastExcluded-firstExcluded+1);
							firstExcluded = j;
							lastExcluded = j;
						} else lastExcluded=j;
					}
				} else {
					if (0==firstIncluded) {
						firstIncluded = j;
						lastIncluded = j;
					} else {
						if (1 < j-lastIncluded) {
							if (0==includedFilesCount) includeString+="Included file(s): ";
							else includeString+=", ";
							if (lastIncluded == firstIncluded) includeString+=(nhl[i]+firstIncluded);
							else includeString+=(nhl[i]+firstIncluded+"-"+nhl[i]+lastIncluded);
							includedFilesCount+=(lastIncluded-firstIncluded+1);
							firstIncluded = j;
							lastIncluded = j;
						} else lastIncluded=j;
					}
				}
			}
		}
		if (0<firstExcluded) {
			if (0==excludedFilesCount) excludeString+="Excluded file(s): ";
			else excludeString+=", ";
			if (lastExcluded == firstExcluded) excludeString+=(nhl[i]+lastExcluded);
			else excludeString+=(nhl[i]+firstExcluded+"-"+nhl[i]+lastExcluded);
			excludedFilesCount+=(lastExcluded-firstExcluded+1);
		}
		if (0<firstIncluded) {
			if (0==includedFilesCount) includeString+="Included file(s): ";
			else includeString+=", ";
			if (lastIncluded == firstIncluded) includeString+=(nhl[i]+lastIncluded);
			else includeString+=(nhl[i]+firstIncluded+"-"+nhl[i]+lastIncluded);
			includedFilesCount+=(lastIncluded-firstIncluded+1);
		}				
	}
	if (0<excludedFilesCount) {
		LogStr+=excludeString;
		if (1<excludedFilesCount) LogStr+=(" ("+excludedFilesCount+").");
		LogStr+=ret;
	}
	if (0<includedFilesCount) {
		LogStr+=includeString;
		if (1<includedFilesCount) LogStr+=(" ("+includedFilesCount+").");
		LogStr+=ret;
	}
	
	// ---- include error reports if there are any
	if (Canceled) {
		LogStr+="Not completed";
		if (0 == errString.length()) LogStr+=(" (probably canceled)."+ret);
		else LogStr+=(" with incident(s)"+errString+"."+ret);
	} else if (0< errString.length()) LogStr+=("Error(s)"+errString+"!"+ret);
	
	// ---- display the log
	if (DoShowLog) {
		if (DoFitLogWinSizeToScreen && 0>=zoomLevel) FitLogWinSizeToScreen();
		TextWindow logWindow = new TextWindow ("Log"+ftag+".txt", "", LogWinWidth, LogWinHeight);
		TextPanel logPanel = logWindow.getTextPanel();
		if (logPanel!=null) logPanel.append(LogStr);
		else {
			if (!silentMode && !Canceled)  IJ.error("Failed to Display Log!");
			if (!Canceled) errString+=(", Failed to Display Log");
		}
	}
	
	// ---- save log to file
	if (DoSaveLog && DidSave) return SaveToTxtFile(DirIn+BSL+"Log"+ftag+".txt", LogStr);
	return true;
}

private boolean SaveToTxtFile(String fileName, String st) {
	FileWriter output =null;
	boolean SuccessSoFar=true;
	try {
		File aFile = new File(fileName);
		if (aFile.isFile()) {
			if (!aFile.canWrite()) SuccessSoFar=false;
			else aFile.delete();
		}
		if (SuccessSoFar) if (aFile.isFile()) SuccessSoFar=false;
		if (SuccessSoFar) {if (!aFile.createNewFile()) SuccessSoFar=false;}
		if (SuccessSoFar) {
			output =new FileWriter(aFile);
			output.write(st);
		}
	}
	catch (IOException ex) {SuccessSoFar=false;}
	catch (SecurityException ex) {SuccessSoFar=false;}
	catch (NullPointerException ex) {SuccessSoFar=false;}
	finally {
		try {if (output != null) output.close();}
		catch (IOException ex) {}
	}
	return SuccessSoFar;
}

private String getTmpPath() {
	String tmpPath =  "";
	try {tmpPath = IJ.getDirectory("temp");}
	catch(NoSuchMethodError e) {
		// --- workaround for ImageJ version < 1.32 ---
		tmpPath = System.getProperty("java.io.tmpdir");
		if (tmpPath == null) tmpPath = "";
		if (0==tmpPath.length() && IJ.isWindows()) {
			tmpPath = GetEnvironS("TEMP");
			if (0==tmpPath.length()) tmpPath = GetEnvironS("TMP");
		}
		if (!tmpPath.endsWith(BSL)) tmpPath += BSL;
	}
	return tmpPath;
}

private String [] getTmpFileNames(int n) {
	String tmpPath = getTmpPath();
	String tmpFileNames[] = new String[n];
	if (0==n) return tmpFileNames;  // nothing more to do
	int numberOfLoops = 50;   // limit to avoid that it searches forever for a possible file name...
	int numberOfLoopsAtFail = 3;	
	int looped = 0;  // count up
	for (int z=0; z<tmpFileNames.length; z++)  if (numberOfLoops >= looped) {
		looped = 0;
		do {
			tmpFileNames[z] = "" + Math.round(100000000 * Math.random());
			while (8 > tmpFileNames[z].length()) tmpFileNames[z] +="0"; 
			tmpFileNames[z] = tmpFileNames[z] + ".tif";
			File tFile = new File (tmpPath + tmpFileNames[z]);
			try {
				// check wether file is there,
				//     if not, try to make a testing one (for the first file name only)
				if (tFile.isFile()) tmpFileNames[z] = "";
				else if (0==z) {
					if (!tFile.createNewFile()) {
						tmpFileNames[z] = "";
						// --  push up counter
						looped+=(numberOfLoops / numberOfLoopsAtFail);
					}
				}
			}
			catch (IOException ex) {tmpFileNames[z] = "";}
			catch (SecurityException ex) {tmpFileNames[z] = "";}
			catch (NullPointerException ex) {tmpFileNames[z] = "";}
			//   -- try to delete the first testing file
			if (0==z && 0 < tmpFileNames[z].length()) {
				try {
					tFile.delete();
					if (tFile.isFile()) tmpFileNames[z] = "";
				}
				catch (SecurityException e) {tmpFileNames[z] = "";}
				catch (NullPointerException ex) {tmpFileNames[z] = "";}
			}
			
			// -- check whether user had pressed the Cancel button meanwhile
			if (0 == tmpFileNames[z].length()) {
				try {
					if (IJ.escapePressed()) {
						IJ.showStatus("canceled.");
						IJ.resetEscape();
						if (!Canceled) errString+=", Escape pressed";
						Canceled=true;
						keepSet = true;
						return null;
					}
				} catch (NoSuchMethodError e) {}
			}
		} while (0 == tmpFileNames[z].length() && numberOfLoops > looped++);
	}
	if (numberOfLoops < looped) {
		if (!silentMode && !Canceled) IJ.error("Fail to generate temporary files!");
		if (!Canceled) errString+=(", Fail to generate temporary files");
		return null;
	}
	return tmpFileNames;
}

private ImagePlus getProjectionFromTmpImg(ImagePlus[] tmpImp, String []  tmpFileNames, int method, boolean was32bit) {
	String tmpPath = getTmpPath();
	ImagePlus impIn, impProj;
	ImageStack stack = null;
	ImageProcessor ip;
	int Pz = 0;
	int Tz = tmpImp.length;
	int Sz = tmpFileNames.length + Tz;
	
	for (int z = 1; z <= Sz; z++) {
		// -- get data back form temporary images
		impProj = null;
		if (z <= Tz) {
			IJ.showStatus("Rearranging frames from temp. images ("+z+")");
			impIn = tmpImp[z-1];
		} else {
			IJ.showStatus("Loading frames to temp. files ("+z+")");
			impIn = openImage(tmpPath, tmpFileNames[z-Tz-1]);
		}
		if (null != impIn) {
			if (1==z) Pz = impIn.getStackSize();
			else if (Pz != impIn.getStackSize()) {
				// -- missmatch error, non-essential but indicating that somthing else went wrong before
				if (verboseMode && !Canceled) IJ.error("Missmatch with temp. images/files at reload ("+z+").");
				if (!Canceled) errString+=(", Missmatch with temp. images/files at reload ("+z+").");
				Pz = impIn.getStackSize();
			}
			impProj = ZProjStack(impIn, 1, Pz, method, was32bit);
		} else {
			if (!silentMode && !Canceled) IJ.error("Failed to get frames form temp. images/files at reload ("+z+")!");
			if (!Canceled) errString+=(", Failed to get frames form temp. images/files at reload ("+z+")");
			return null;
		}
		
		// -- get rid of imput image (that is projected)
		if (z <= Tz) impIn = null;
		else impIn = CloseImageWindow(impIn);

		// -- rebuild stack
		try {
			ip = impProj.getProcessor();
			if (1== z) {
				stack = new ImageStack(ip.getWidth(), ip.getHeight());
			}
			stack.addSlice(null, ip);
		}
		catch(OutOfMemoryError e) {
			if (!silentMode && !Canceled) IJ.error("Projection of Tmp failed:\nNot enough memory at "+z+"!");
			if (!Canceled) errString+=(", Projection of Tmp failed: Not enough memory at "+z+"!");
			return null;
		}
		catch(NullPointerException e) {
			if (!silentMode && !Canceled) IJ.error("Projection of Tmp failed:\nProjection "+z+" failed!");
			if (!Canceled) errString+=(", Projection of Tmp failed: Projection "+z+" failed!");
			return null;
		}
		catch (IllegalArgumentException e) {
			if (!silentMode && !Canceled) IJ.error("Projection of Tmp failed:\nPossibly canceled.");
			if (!Canceled) errString+=(", Projection of Tmp failed: Possibly canceled.");
			Canceled=true;
			return null;
		}

		// -- get rid of projection image
		impProj = CloseImageWindow(impProj);
		
		// -- check if Cancel is pressed
		try {
			if (IJ.escapePressed()) {
				IJ.showStatus("canceled.");
				IJ.resetEscape();
				if (!Canceled) errString+=", Escape pressed";
				Canceled=true;
				keepSet = true;
				return null;
			}
		} catch (NoSuchMethodError e) {}
	} // for (int z = 1; z <= nStacks; z++)
	if (null == stack) return null;
	ImagePlus impOut = new ImagePlus("temp", stack);
	return impOut;
}

private void deleteTmpFiles(String [] tmpFileNames) {
	if (null == tmpFileNames) return;
	int nStacks = tmpFileNames.length;
	String tmpPath = getTmpPath();
	for (int z = 1; z <= nStacks; z++) {  
		IJ.showStatus("Deleting temp. file ("+z+" of "+nStacks+")");
		try {
			File tFile = new File (tmpPath + tmpFileNames[z-1]);
			if (tFile.isFile()) tFile.delete();
		}
		catch (SecurityException ex) {}
		catch (NullPointerException ex) {}
	}
	IJ.showStatus("");
}

private boolean distributeStackToTmpFile(ImagePlus img1, ImagePlus[] tmpImp, String []  tmpFileNames) {
	String ImageName = img1.getTitle();
	String tmpPath = getTmpPath();
	ImageStack stack2 = null;
	ImagePlus impTrans = null;
	ImageStack stack1 = img1.getStack();
	int Sz = stack1.getSize();
	int Tz = tmpImp.length;
	if (Sz != tmpFileNames.length + Tz) {
		if (!silentMode && !Canceled) IJ.error("Missmatch of '"+ImageName+"' with temp. images/files!");
		if (!Canceled) errString+=(", Missmatch of '"+ImageName+"' with temp. images/files!");
		return false;
	}
	int fframe=0;
	for (int z = 1; z <= Sz; z++) {
		try {
			fframe=z;
			ImageProcessor ip1 = stack1.getProcessor(z);
			if (z <= Tz) {
				IJ.showStatus("Rearranging frames to temp. images ("+z+")");
				if (null == tmpImp[z-1]) {
					tmpImp[z-1] = new ImagePlus("temp"+z, ip1);
				} else {
					stack2 = tmpImp[z-1].getStack();
					stack2.addSlice(null, ip1);
					tmpImp[z-1].setStack(null, stack2);
				}
			} else {
				IJ.showStatus("Saving frames to temp. files ("+z+")");
				File tFile = new File (tmpPath + tmpFileNames[z-Tz-1]);
				if (tFile.isFile()) {
					impTrans=openImage(tmpPath, tmpFileNames[z-Tz-1]);
					stack2 = impTrans.getStack();
					stack2.addSlice(null, ip1);
					impTrans.setStack(null, stack2);
					new FileSaver(impTrans).saveAsTiffStack(tmpPath + tmpFileNames[z-Tz-1]);
					IJ.showStatus("Saving frames to temp. files ("+z+" of "+Sz+")");
				} else {
					impTrans = new ImagePlus("temp"+z, ip1);
					new FileSaver(impTrans).saveAsTiff(tmpPath + tmpFileNames[z-Tz-1]);
				}
				impTrans = CloseImageWindow (impTrans);
			}
		}
		catch(OutOfMemoryError e) {
			if (!silentMode && !Canceled) IJ.error("Failed to rearrange/save frame "+fframe+ " of '"+ImageName+"':\nNot enough memory!");
			if (!Canceled) errString+=(", Failed to rearrange/save frame "+fframe+ " of '"+ImageName+": Not enough memory!");
			return false;
		}
		catch(NullPointerException e) {
			if (!silentMode && !Canceled) IJ.error("Failed to rearrange/save frame "+fframe+ " of '"+ImageName+"':\nNot found!");
			if (!Canceled) errString+=(", Failed to rearrange/save frame "+fframe+ " of '"+ImageName+": Not fond!");
			return false;
		}
		catch (IllegalArgumentException e) {
			if (!silentMode && !Canceled) IJ.error("Failed to rearrange/save frame "+fframe+ " of '"+ImageName+"':\nprobably canceled.");
			if (!Canceled) errString+=(", Failed to rearrange/save frame "+fframe+ " of '"+ImageName+": probably canceled.");
			Canceled=true;
			return false;
		}
		try {
			if (IJ.escapePressed()) {
				IJ.showStatus("canceled.");
				IJ.resetEscape();
				if (!Canceled) errString+=", Escape pressed";
				Canceled=true;
				keepSet = true;
				return false;
			}
		} catch (NoSuchMethodError e) {}
	}
	return true;
}


//  ----------  //  ----------  //  ----------  //  ----------  //  ----------  CALCULATIONS  ----------  //  ----------  //  ----------  //  ----------  //  ----------  
private ImagePlus CalculateStim(ImagePlus imp1, ImagePlus impNo, ImagePlus impS, File directory, String fileN, String FTag, boolean was32bit) {
	String Ns="";
	// ------------------  dviding    Image1   by the control stack (without stimmulation) if applicable  ------------
	if ( Normalizeb && hasControl && 1<nhl.length && null!=impNo && null!=imp1) {
		//  if (null==impNo) {if (!silentMode && !Canceled) IJ.error("Failed to normalize stack!");}
		if (AMethod != SD_METHOD && AMethod !=SE_METHOD) {
			Icalculate(impNo,imp1,1,0,DIVIDE1);
			if (Canceled) return null;
			if (ShowAv || SaveAv) ScaleTheMiddle(imp1,(responsefirst+responselast)/2,ScalingD);
		}
		Ns="n";
	}
	ImagePlus impB=null;
	ImagePlus impR=null;
	//  ----------  BASELINE POJECTION  ----------  
	if (null==imp1) {
		if (verboseMode && !Canceled) IJ.error("Can't calculate results for '"+fileN+"'!");
		if (!Canceled) errString+=(", Can't calculate results for '"+fileN+"'");
		return null;
	}
	if (ShowBase||SaveBase||ShowDiff||SaveDiff||ShowDiv||SaveDiv||ShowPerc||SavePerc||ShowDelta||SaveDelta) {
		IJ.showStatus("Calculating: " + fileN+"_base"+Ns+FTag+".tif");
		impB=ZProjStack(imp1, baselinefirst, baselinelast, AVG_METHOD, was32bit);
		// ---------------  optionally, apply Gaussian / Mean filter to base projection -------------------
		if (!Canceled && 0<FilterXy && DoFilterXY && DoFilterBR) {
			if (was32bit) preFilter (impB.getProcessor(), FilterMean, FilterXy);
			if (!FilterMean) new GaussianBlur().blur(impB.getProcessor(),FilterXy);
			else new RankFilters().rank(impB.getProcessor(), Math.rint(FilterXy), RankFilters.MEAN);
		}
		if (!Canceled && ShowBase||SaveBase) {
			ScaleTheMiddle(impB,1,ScalingD);
			if (SaveBase) {
				new FileSaver(impB).saveAsTiff(directory+BSL+fileN+"_base"+Ns+FTag+".tif");
				DidSave=true;
			}
			if (ShowBase) {
				impB.setTitle(fileN+"_base"+Ns+FTag+".tif");
				impB.changes=!SaveBase;
				impB.show();
				ZoomGood(impB);
			}
		}
	}
	
	//  ----------  RESPONSE PROJECTION  ----------  
	if (!Canceled && (ShowResp||SaveResp||ShowDiff||SaveDiff||ShowDiv||SaveDiv|| ((ShowRatio||SaveRatio) && (2<nhl.length || (!hasControl && 1<nhl.length))) || ShowPerc||SavePerc)) {
		IJ.showStatus("Calculating: " + fileN+"_resp"+Ns+FTag+".tif");
		impR=ZProjStack(imp1, responsefirst, responselast, AVG_METHOD, was32bit);
		// ---------------  optionally, apply Gaussian or Mean filter to response projection -----------------
		if (!Canceled && 0<FilterXy && DoFilterXY && DoFilterBR) {
			if (was32bit) preFilter (impR.getProcessor(), FilterMean, FilterXy);
			if (!FilterMean) new GaussianBlur().blur(impR.getProcessor(),FilterXy);
			else new RankFilters().rank(impR.getProcessor(), Math.rint(FilterXy), RankFilters.MEAN);
		}
		if (!Canceled && ShowResp||SaveResp) {
			ScaleTheMiddle(impR,1,ScalingD);
			if (SaveResp) {
			new FileSaver(impR).saveAsTiff(directory+BSL+fileN+"_resp"+Ns+FTag+".tif");
				DidSave=true;
			}
			if (ShowResp) {
				impR.setTitle(fileN+"_resp"+Ns+FTag+".tif");
				impR.changes=!SaveResp;
				impR.show();
				ZoomGood(impR);
			}
		}
	}
	
	//  ----------  DIFFERENCE BETWEEN RESPONSE AND BASELINE ----------  
	if (!Canceled && (ShowDiff||SaveDiff)) {
		IJ.showStatus("Calculating: " + fileN+"_diff"+Ns+FTag+".tif");
		ImagePlus impF=null;
		if (ShowResp||ShowDiv||SaveDiv|| ((ShowRatio||SaveRatio) && (2<nhl.length || (!hasControl && 1<nhl.length)))  ||ShowPerc||SavePerc) impF=duplicateStack(impR);
		else impF=impR;
		if (null!=impF) {
			Icalculate(impB,impF,1,0,SUBTRACT);
			if (!Canceled) {
				ScaleTheMiddle(impF,1,ScalingD);
				if (SaveDiff) {
					new FileSaver(impF).saveAsTiff(directory+BSL+fileN+"_diff"+Ns+FTag+".tif");
					DidSave=true;
				}
			}
			if (!Canceled && ShowDiff) {
				impF.setTitle(fileN+"_diff"+Ns+FTag+".tif");
				impF.changes=!SaveDiff;
				impF.show();
				ZoomGood(impF);
			} else impF = CloseImageWindow(impF);
		} else {
			if (verboseMode && !Canceled) IJ.error("Can't calculate difference for '"+fileN+"'!");
			if (!Canceled) errString+=(", Can't calculate difference for '"+fileN+"'");
		}
		if (!ShowBase && !ShowDiv && !SaveDiv && !ShowDelta && !SaveDelta && !ShowPerc && !SavePerc) impB = CloseImageWindow(impB);
	}
	
	//  ----------  RATIO OF RESPONSE AND BASELINE ----------  
	if (!Canceled && (ShowDiv||SaveDiv)) {
		IJ.showStatus("Calculating: " + fileN+"_div"+Ns+"_32bit"+FTag+".tif");
		ImagePlus impV=null;
		if (ShowResp||ShowDiff|  ((ShowRatio||SaveRatio) && (2<nhl.length || (!hasControl && 1<nhl.length)))  ||ShowPerc||SavePerc) impV=duplicateStack(impR);
		else impV=impR;
		if (null!=impV) {
			Icalculate(impB,impV,1,0,DIVIDE1);
			if (!Canceled) {
				ScaleTheMiddle(impV,1,ScalingD);
				if (SaveDiv) {
					new FileSaver(impV).saveAsTiff(directory+BSL+fileN+"_div"+Ns+"_32bit"+FTag+".tif");
					DidSave=true;
				}
			}
			if (!Canceled && ShowDiv) {
				impV.setTitle(fileN+"_div"+Ns+"_32bit"+FTag+".tif");
				impV.changes=!SaveDiv;
				impV.show();
				ZoomGood(impV);
			} else impV=CloseImageWindow(impV);
		} else {
			if (verboseMode && !Canceled) IJ.error("Can't calculate 'Div' for '"+fileN+"'!");
			if (!Canceled) errString+=(", Can't calculate 'Div' for '"+fileN+"'");
		}
		if (!ShowBase && !ShowDelta && !SaveDelta && !ShowPerc && !SavePerc) impB = CloseImageWindow(impB);
	}
	
	//  ----------  PERCENT RATIO OF RESPONSE AND BASELINE ----------  
	if (!Canceled && (ShowPerc||SavePerc)) {
		IJ.showStatus("Calculating: " + fileN+"_div"+Ns+"_%change"+FTag+".tif");
		ImagePlus impT=null;
		if (ShowResp||ShowDiff||ShowDiv|| ((ShowRatio||SaveRatio) && (2<nhl.length || (!hasControl && 1<nhl.length))) ) impT=duplicateStack(impR);
		else impT=impR;
		if (null!=impT) {
			Icalculate(impB,impT,1,0,DIVIDE1);
			if (!Canceled) IcalculateSingle(impT,100,-1,SCALEREV);
			if (!Canceled) {
				ScaleTheMiddle(impT,1,ScalingD);
				if (SavePerc) {
					new FileSaver(impT).saveAsTiff(directory+BSL+fileN+"_div"+Ns+"_%change"+FTag+".tif");
					DidSave=true;
				}
			}
			if (!Canceled && ShowPerc) {
				impT.setTitle(fileN+"_div"+Ns+"_%change"+FTag+".tif");
				impT.changes=!SavePerc;
				impT.show();
				ZoomGood(impT);
			} else impT = CloseImageWindow(impT);
		} else {
			if (verboseMode && !Canceled) IJ.error("Can't calculate percentage for '"+fileN+"'!");
			if (!Canceled) errString+=(", Can't calculate percentage for '"+fileN+"'");
		}
		if (!ShowBase && !ShowDelta && !SaveDelta) impB = CloseImageWindow(impB);
	}
	
	//  ----------  RATIO [DELTA]-F-OVER-F(0)  ----------  
	if (!Canceled && (ShowDelta||SaveDelta)) {
		ImagePlus impD=null;
		ImagePlus impDS=null;
		String MsgTxt = (0<counter[0][0]&&1<nhl.length&&Normalizeb && hasControl)?"Generating "+DST+"F/F"+SZS:"Generating "+DST+"I/I"+SZS;
		try {
			if (IJ.escapePressed()) {
				IJ.showStatus("canceled.");
				IJ.resetEscape();
				if (!Canceled) errString+=", Escape pressed";
				Canceled=true;
				keepSet = true;
				if (!SaveAv && !ShowAv) imp1 = CloseImageWindow(imp1);
				if (!ShowBase) impB = CloseImageWindow(impB);
			}
		} catch (NoSuchMethodError ex) {}
		if (!Canceled) {
	 		MsgTxt = MsgTxt + " stack: " + fileN + FTag;
			IJ.showStatus(MsgTxt );
			if (SaveAv || ShowAv || (0<FilterXy && DoFilterXY && DoFilterBR)) {
				impD=duplicateStack(imp1);
				// ---------------  apply Gaussian / Mean filter if projection was filtererd -----------------
				if (0<FilterXy && DoFilterXY && DoFilterBR) {
					try {
						ImageStack impDstack = impD.getStack();
						int Sz=impDstack.getSize();
						if (was32bit) for (int fi=1; fi<=Sz; fi++) preFilter (impDstack.getProcessor(fi), FilterMean, FilterXy);
						if (!FilterMean) {
							GaussianBlur gaussian=new GaussianBlur();
							for (int fi=1; fi<=Sz; fi++) gaussian.blur(impDstack.getProcessor(fi),FilterXy);
						} else {
							RankFilters rankFilter=new RankFilters();
							for (int fi=1; fi<=Sz; fi++) rankFilter.rank(impDstack.getProcessor(fi), Math.rint(FilterXy), RankFilters.MEAN);
						}
						impD.setStack(null, impDstack);
					}  catch (NullPointerException e) {
						impD = CloseImageWindow(impD);
						if (!silentMode && !Canceled) IJ.error("Failed to filter by "+MsgTxt+"! (Error dg)");
						if (!Canceled) errString+=(", Failed to filter by "+MsgTxt+" (Error dg)");
					}
					if (!SaveAv && !ShowAv) imp1 = CloseImageWindow(imp1);
				}
			} else {
				impD=imp1;
				imp1=null;
			}
		}
		try {
			if (IJ.escapePressed()) {
				IJ.showStatus("canceled.");
				IJ.resetEscape();
				if (!Canceled) errString+=", Escape pressed";
				Canceled=true;
				keepSet = true;
				impD = CloseImageWindow(impD);
				if (!ShowBase) impB = CloseImageWindow(impB);
				impD=null;
			}
		} catch (NoSuchMethodError ex) {}
		
		// -- calculate Delta SD - over - baseline
		if (!Canceled && (AMethod==SD_METHOD || AMethod==SE_METHOD) && null !=impS) {
	 		if (SaveAv || ShowAv || (0<FilterXy && DoFilterXY && DoFilterBR)) {
				impDS=duplicateStack(impS);
				// ---------------  apply Gaussian / Mean filter if projection was filtererd -----------------
				if (0<FilterXy && DoFilterXY && DoFilterBR) {
					try {
						ImageStack impDSstack = impDS.getStack();
						int Sz=impDSstack.getSize();
						if (was32bit) for (int fi=1; fi<=Sz; fi++) preFilter (impDSstack.getProcessor(fi), FilterMean, FilterXy);
						if (!FilterMean) {
							GaussianBlur gaussian=new GaussianBlur();
							for (int fi=1; fi<=Sz; fi++) gaussian.blur(impDSstack.getProcessor(fi),FilterXy);
						} else {
							RankFilters rankFilter=new RankFilters();
							for (int fi=1; fi<=Sz; fi++) rankFilter.rank(impDSstack.getProcessor(fi), Math.rint(FilterXy), RankFilters.MEAN);
						}
						impDS.setStack(null, impDSstack);
					}  catch (NullPointerException e) {
						impDS = CloseImageWindow(impDS);
						if (!silentMode && !Canceled) IJ.error("Failed to filter by "+MsgTxt+"! (Error dgS)");
						if (!Canceled) errString+=(", Failed to filter by "+MsgTxt+" (Error dgS)");
					}
					if (!SaveAv && !ShowAv) impS = CloseImageWindow(impS);
				}
			} else {
				impDS=impS;
				impS=null;
			}
		}
		try {
			if (IJ.escapePressed()) {
				IJ.showStatus("canceled.");
				IJ.resetEscape();
				if (!Canceled) errString+=", Escape pressed";
				Canceled=true;
				keepSet = true;
				impD = CloseImageWindow(impD);
				if (!ShowBase) impB = CloseImageWindow(impB);
				impD=null;
			}
		} catch (NoSuchMethodError ex) {}
		if (null!=impD && null!=impB) {
			ImagePlus impE=null;
			String ps=DoDelta100?"%":"";
			String FI=(0<counter[0][0]&&1<nhl.length&&Normalizeb && hasControl)?"F":"I";
			String DN="_delta"+FI;
			// ---------------  normalize to baseline (i.e. IOS data) unless already normalized -----------------
			if (0==Ns.length()) {
				Icalculate(impB,impD,1,0,DIVIDE1);
				if (Canceled || (!ShowBase &&  AMethod !=SD_METHOD && AMethod != SE_METHOD))
	 				impB = CloseImageWindow(impB);
				if (!Canceled) {
					impD.setTitle(fileN+DN+Ns+"_"+FI+Ns+ps+FTag+".tif");
					impE=ZProjStack(impD, baselinefirst, baselinelast, AVG_METHOD, was32bit);
				}
			} else {
				if (ShowBase || AMethod==SD_METHOD || AMethod==SE_METHOD) impE=duplicateStack(impB);
				else impE=impB;
			}
			// ---------------  optionally, apply (another) Gaussian / Mean filter -----------------
			if (!Canceled && 0<FilterXy && DoFilterXY && DoFilterDelta) {
				try {
					ImageStack impDstack = impD.getStack();
					int Sz=impDstack.getSize();
					if (was32bit) for (int fi=1; fi<=Sz; fi++) preFilter (impDstack.getProcessor(fi), FilterMean, FilterXy);
					if (!FilterMean) {
						GaussianBlur gaussian=new GaussianBlur();
						gaussian.blur(impE.getProcessor(),FilterXy);
						for (int fi=1; fi<=Sz; fi++) gaussian.blur(impDstack.getProcessor(fi),FilterXy);
					} else {
						RankFilters rankFilter=new RankFilters();
						rankFilter.rank(impE.getProcessor(), Math.rint(FilterXy), RankFilters.MEAN);
						for (int fi=1; fi<=Sz; fi++) rankFilter.rank(impDstack.getProcessor(fi), Math.rint(FilterXy), RankFilters.MEAN);
					}
					impD.setStack(null, impDstack);
				} 
				catch (NullPointerException e) {
					impD = CloseImageWindow(impD);
					if (!silentMode && !Canceled) IJ.error("Failed to filter by "+MsgTxt+"! (Error df)");
					if (!Canceled) errString+=(", Failed to filter by "+MsgTxt+" (Error df)");
				}
			}
			try {
				if (IJ.escapePressed()) {
					IJ.showStatus("canceled.");
					IJ.resetEscape();
					if (!Canceled) errString+=", Escape pressed";
					Canceled=true;
					keepSet = true;
					impD = CloseImageWindow(impD);
				}
			} catch (NoSuchMethodError ex) {}
			if  (!Canceled) {
				try {
					IJ.showStatus(MsgTxt );
					Icalculate(impE,impD,DoDelta100?100.0:1.0, 0, NDIV1);
					if (DeltaAutoScale) ScaleTheMiddle(impD, (responsefirst+responselast)/2,DeltaScalingD);
					else impD.getProcessor().setMinAndMax(DeltaMinScale, DeltaMaxScale);
					if (SaveDelta) {
						new FileSaver(impD).saveAsTiffStack(directory+BSL+fileN+DN+Ns+"_"+FI+Ns+ps+FTag+".tif");
						DidSave=true;
						impD = CloseImageWindow(impD);
						if (ShowDelta) {
							IJ.showStatus("reload '"+fileN+DN+Ns+"_"+FI+Ns+ps+FTag+".tif'");
							if (ShowDelta) impD=openImage(""+directory,fileN+DN+Ns+"_"+FI+Ns+ps+FTag+".tif");
							IJ.showStatus("'"+fileN+DN+Ns+"_"+FI+Ns+ps+FTag+".tif' reloaded.");
						}
						if (null==impD ){
							if (!silentMode && null==impD && !Canceled) IJ.error("Failed to display "+DST+FI+"/"+FI+SZS+" stack! (Error ddsp0)!");
							if (!Canceled) errString+=(", Failed to display Delta"+FI+"/"+FI+"0 stack (Error ddsp0)");
						}
					} else {
						impD.setTitle(fileN+DN+Ns+"_"+FI+Ns+ps+FTag+".tif");
						impD.changes=true;
					}
					if (ShowDelta && null!=impD) {
						impD.show();
						ZoomGood(impD);
						impD.setSlice((responsefirst+responselast)/2);
						impD.updateAndRepaintWindow();
					}
				}
				catch (NullPointerException e) {
					impD = CloseImageWindow(impD);
					if (!silentMode && !Canceled) IJ.error("Failed with  "+MsgTxt+" (Error dc2)!");
					if (!Canceled) errString+=(", Failed with  "+MsgTxt+" (Error dc2)");
				}
			}
			impE = CloseImageWindow(impE);
		} else {
			if (!Canceled && !silentMode) IJ.error("Failed with  "+MsgTxt+" (Error dc1)!");
			if (!Canceled) errString+=(", Failed with  "+MsgTxt+" (Error dc1)");
		}
		if (null!=impDS && null!=impB) {
			ImagePlus impES=null;
			String ps=DoDelta100?"%":"";
			String FI=(0<counter[0][0]&&1<nhl.length&&Normalizeb && hasControl)?"F":"I";
			String DN="_deltaSD"+FI;
			// ---------------  normalize to baseline (i.e. IOS data) unless already normalized -----------------
			if (0==Ns.length()) {
				Icalculate(impB,impDS,1,0,DIVIDE1);
				if (Canceled || !ShowBase) impB = CloseImageWindow(impB);
				if (!Canceled) {
					impDS.setTitle(fileN+DN+Ns+"_"+FI+Ns+ps+FTag+".tif");
					impES=ZProjStack(impDS, baselinefirst, baselinelast, AVG_METHOD, was32bit);
				}
			} else {
				if (ShowBase) impES=duplicateStack(impB);
				else impES=impB;
			}
			// ---------------  optionally, apply (another) Gaussian / Mean filter -----------------
			if (!Canceled && 0<FilterXy && DoFilterXY && DoFilterDelta) {
				try {
					ImageStack impDSstack = impDS.getStack();
					int Sz=impDSstack.getSize();
					if (was32bit) for (int fi=1; fi<=Sz; fi++) preFilter (impDSstack.getProcessor(fi), FilterMean, FilterXy);
					if (!FilterMean) {
						GaussianBlur gaussian=new GaussianBlur();
						gaussian.blur(impES.getProcessor(),FilterXy);
						for (int fi=1; fi<=Sz; fi++) gaussian.blur(impDSstack.getProcessor(fi),FilterXy);
					} else {
						RankFilters rankFilter=new RankFilters();
						rankFilter.rank(impES.getProcessor(), Math.rint(FilterXy), RankFilters.MEAN);
						for (int fi=1; fi<=Sz; fi++) rankFilter.rank(impDSstack.getProcessor(fi), Math.rint(FilterXy), RankFilters.MEAN);
					}
					impDS.setStack(null, impDSstack);
				} 
				catch (NullPointerException e) {
					impDS = CloseImageWindow(impDS);
					if (!silentMode && !Canceled) IJ.error("Failed to filter by "+MsgTxt+"! (Error dfS)");
					if (!Canceled) errString+=(", Failed to filter by "+MsgTxt+" (Error dfS)");
				}
			}
			try {
				if (IJ.escapePressed()) {
					IJ.showStatus("canceled.");
					IJ.resetEscape();
					if (!Canceled) errString+=", Escape pressed";
					Canceled=true;
					keepSet = true;
					impDS = CloseImageWindow(impDS);
				}
			} catch (NoSuchMethodError ex) {}
			if  (!Canceled) {
				try {
					IJ.showStatus(MsgTxt );
					Icalculate(impES,impDS,DoDelta100?100.0:1.0, 0, DIVIDE1);
					if (DeltaAutoScale) ScaleTheMiddle(impDS, (responsefirst+responselast)/2,DeltaScalingD);
					else impDS.getProcessor().setMinAndMax(DeltaMinScale, DeltaMaxScale);
					if (SaveDelta) {
						new FileSaver(impDS).saveAsTiffStack(directory+BSL+fileN+DN+Ns+"_"+FI+Ns+ps+FTag+".tif");
						DidSave=true;
						impDS = CloseImageWindow(impDS);
						if (ShowDelta) impDS=openImage(""+directory,fileN+DN+Ns+"_"+FI+Ns+ps+FTag+".tif");
						if (null==impDS ){
							if (!silentMode && null==impDS && !Canceled) IJ.error("Failed to display "+DST+FI+"/"+FI+SZS+" stack! (Error ddsp1)!");
							if (!Canceled) errString+=(", Failed to display Delta"+FI+"/"+FI+"0 stack (Error ddsp1)");
						}
					} else {
						impDS.setTitle(fileN+DN+Ns+"_"+FI+Ns+ps+FTag+".tif");
						impDS.changes=true;
					}
					if (ShowDelta && null!=impDS) {
						impDS.show();
						ZoomGood(impDS);
						impDS.setSlice((responsefirst+responselast)/2);
						impDS.updateAndRepaintWindow();
					}
				}
				catch (NullPointerException e) {
					impDS = CloseImageWindow(impDS);
					if (!silentMode && !Canceled) IJ.error("Failed with  "+MsgTxt+" (Error dc2S)!");
					if (!Canceled) errString+=(", Failed with  "+MsgTxt+" (Error dc2S)");
				}
			}
			impES = CloseImageWindow(impES);
		} else if (AMethod==SD_METHOD || AMethod==SE_METHOD) {
			if (!Canceled && !silentMode) IJ.error("Failed with  "+MsgTxt+" (Error dc1)!");
			if (!Canceled) errString+=(", Failed with  "+MsgTxt+" (Error dc1)");
		}
	}
	return impR;
}
//  ----------  //  ----------  //  ----------  //  ----------  //  ----------  //  ----------  //  ----------  //  ----------  //  ----------  //  ----------  //  ----------  //  ----------  //  ----------  

private boolean showDialog() {
	gd = new GenericDialog(PluginTitle, IJ.getInstance());
	inputFolderLabel = new JLabel("Input Folder:    ");
	
	int pi=-1;
	if (0<DirIn.length()) {
		if (!IJ.isWindows()) {for (int pc=0;pc<PDir.length;pc++) if(0>pi) if (0==DirIn.compareTo(PDir[pc])) pi=pc;}
		if(0>pi) {for (int pc=0;pc<PDir.length;pc++) if(0>pi) if (0==DirIn.compareToIgnoreCase(PDir[pc])) pi=pc;}
		if(0>pi) {
			String [] TDir = new String [PDir.length];
			for (int tc=0;tc<TDir.length;tc++) TDir[tc]=PDir[tc];
			PDir= new String [PDir.length+1];
			PDir[0]=DirIn;
			for (int tc=0;tc<TDir.length;tc++) PDir[tc+1]=TDir[tc];
			pi=0;
		} else PDir[pi]=DirIn;
	} else {
		pi=0;
		DirIn=PDir[0];
	}
	IJ.showStatus("Select files in " + DirIn);
	
	inputFolderChoice = new JComboBox(PDir);
	inputFolderChoice.setEditable(true);
	
	// --- set Combobox size; if necessary with workaround allowing run with IJ 1.36bnd, "setPrototypeDisplayValue()"
	try {inputFolderChoice.setPrototypeDisplayValue("XXXXXXXXXXXXX.XXXXXXXXXXXXX");}
	catch (NoSuchMethodError ex) {((JTextField)inputFolderChoice.getEditor().getEditorComponent()).setColumns(19);}
			//  ############ find better workaround (i.e. add blanks and trim afterwards!)  
	if(0<pi) inputFolderChoice.setSelectedIndex(pi);
	
	inputPanel = new Panel();
	inputFolderButton = new JButton( "browse" );
	inputPanel.add(inputFolderLabel);
	inputPanel.add(inputFolderChoice);
	
	inputPanel.add(inputFolderButton);
	
	// - create an instance of inner classes ButtonHandler and TextFieldHandler to use for button event handling
	ButtonHandler handler = new ButtonHandler();
	inputFolderButton.addActionListener( handler );
	TextFieldHandler thandler  = new TextFieldHandler();
	inputFolderChoice.addActionListener(thandler);	
	try {gd.addPanel(inputPanel);}
	catch (NoSuchMethodError e) {
		String vMsg = "This plugin requires a newer version of ImageJ.\n";
		vMsg+="Please update and run again.";
		IJ.error(vMsg);
		return false;
	}
	
	Panel OptionPanel=new Panel();
	//  NormCB=new Checkbox ("Normalize to data recorded without stimulation (bleach correction)");
	NormCB=new Checkbox ("Subtract & normalize to"+AVL[AMethod]+"stack "+"'"+nhl[0]+AVS[AMethod]+".tif'");
	if (AMethod==SD_METHOD || AMethod==SE_METHOD) NormCB.setLabel("Subtract & normalize to"+AVL[AVG_METHOD]+"stack "+"'"+nhl[0]+AVS[AVG_METHOD]+".tif'");
	NormCB.addItemListener(new ItemListener() {
		public void itemStateChanged(ItemEvent e) {
			Normalizeb = NormCB.getState();
			RepaintIgnoreBoxes(null);
		}
	});
	
	OptionPanel.add(NormCB);
	gd.addPanel(OptionPanel);
	
	if (allowBlankArtifact) {
		if (2 > NbChkbxs) NbChkbxs = 2;
	} else {
		if (1 > NbChkbxs) NbChkbxs = 4;
	}
	MaxNbChkbxs = getMaxNbChkbxs(allowBlankArtifact);
	if (MaxNbChkbxs < NbChkbxs) NbChkbxs = MaxNbChkbxs;
	
	WButton = new Button(" < ");	 // " \u25c0 ");
	EButton = new Button(">      "); // "\u25b6      ");
	NButton = new Button("^");	 // "\u25b2");
	SButton = new Button("v");	 // "\u25bc");
	IgnoreBox = new JCheckBox [2][NbChkbxs+1];
	Panel FilePanel=new Panel();
	Panel leftpanel=new Panel();
	Panel rightpanel=new Panel();
	Panel leftpanel_1=new Panel();
	Panel leftpanel_2=new Panel();
	Panel rightpanel_1=new Panel();
	Panel rightpanel_2=new Panel();
	
	FilePanel.setLayout(new BorderLayout());
	leftpanel.setLayout(new BorderLayout());
	rightpanel.setLayout(new BorderLayout());
	leftpanel_1.setLayout(new GridLayout(0,1));
	leftpanel_2.setLayout(new GridLayout(0,1));
	rightpanel_1.setLayout(new GridLayout(0,1));
	rightpanel_2.setLayout(new GridLayout(0,1));

	IboxStartRow=0;
	IboxStartColumn = 2<nhl.length?1:0;
	if (1==nhl.length) {
		String s="";
		for (int k=nhl[0].length();k<16;k++) s+=" "; //   #### more precisely please!
		for (int i=0;i<2;i++) for(int j=1;j<=NbChkbxs; j++)
				IgnoreBox[i][j]=new JCheckBox (nhl[0]+(j+IboxStartRow+i*NbChkbxs)+".tif"+s, false);
	} else {
		String s="";
		for (int i=0;i<2;i++) {
			s="";
			for (int k=nhl[0].length();k<16;k++) s+=" "; //   #### more precisely please!
			for(int j=1;j<=NbChkbxs;j++) {
				IgnoreBox[i][j]=new JCheckBox (nhl[IboxStartColumn+i]+(j+IboxStartRow)+".tif"+s, false);
				if (1<nhl.length && hasControl && IboxStartColumn==0 && i==0) {
					Font f=IgnoreBox[i][j].getFont();
					int fstyle=(f.isBold()?Font.BOLD:Font.PLAIN);
					if (f.isItalic()) fstyle+=Font.ITALIC;
					f  = new Font(f.getName(), fstyle, f.getSize());
					IgnoreBox[i][j].setFont(f);
				}
			}
		}
	}
	for(int j=1;j< NbChkbxs/2 +1;j++) leftpanel_1.add(IgnoreBox[0][j]);
	leftpanel.add(leftpanel_1, BorderLayout.WEST);
	for(int j= NbChkbxs/2 +1; j<= NbChkbxs;j++) leftpanel_2.add(IgnoreBox[0][j]);
	leftpanel.add(leftpanel_2, BorderLayout.EAST);
	
	for(int j=1;j< NbChkbxs/2 +1;j++) rightpanel_1.add(IgnoreBox[1][j]);
	rightpanel.add(rightpanel_1, BorderLayout.WEST);
	for(int j= NbChkbxs/2 +1;j <= NbChkbxs;j++) rightpanel_2.add(IgnoreBox[1][j]);
	rightpanel.add(rightpanel_2, BorderLayout.EAST);
	
	Panel LatitutdePanel=new Panel();
	LatitutdePanel.setLayout(new BorderLayout());
	WButton.setEnabled(0<IboxStartColumn);
	Button LaB =new Button();
	LaB.setEnabled(false);
	EButton.setEnabled(2<nhl.length-IboxStartColumn);
	LatitutdePanel.add(WButton, BorderLayout.WEST);
	LatitutdePanel.add(LaB);
	LatitutdePanel.add(EButton, BorderLayout.EAST);
	
	Panel LongitutdePanel=new Panel();
	LongitutdePanel.setLayout(new BorderLayout());
	Button LoB =new Button();
	LoB.setEnabled(false);
	LongitutdePanel.add(NButton,BorderLayout.NORTH);
	LongitutdePanel.add(LoB);
	LongitutdePanel.add(SButton,BorderLayout.SOUTH);
	
	FilePanel.add(new Label("   IGNORE  LIST      (Please check the files you don't want to be included)"), BorderLayout.NORTH);
	FilePanel.add(leftpanel, BorderLayout.WEST);
	FilePanel.add(rightpanel);
	FilePanel.add(LongitutdePanel, BorderLayout.EAST);
	FilePanel.add(LatitutdePanel, BorderLayout.SOUTH);
	gd.addPanel(FilePanel);
	
	NButton.addActionListener( handler );
	EButton.addActionListener( handler );
	SButton.addActionListener( handler );
	WButton.addActionListener( handler );
	
	SetCheck(verboseMode);   // sets and enables/disables Checkboxes if files are found/ not found
	
	BaselineFirstField = new JTextField(""+baselinefirst, 4);
	BaselineLastField = new JTextField(""+baselinelast, 4);
	ResponseFirstField = new JTextField(""+responsefirst, 4);
	ResponseLastField = new JTextField(""+responselast, 4);
	if (blankArtifact && allowBlankArtifact) {
		blankFromField = new JTextField(""+blankFrom, 4);
		blankToField = new JTextField(""+blankTo, 4);
	} else {
		blankFromField = new JTextField("", 4);
		blankToField = new JTextField("", 4);
	}
	
	TextFieldHandler tfh = new TextFieldHandler();
	BaselineFirstField.addActionListener(tfh);
	BaselineLastField.addActionListener(tfh);
	ResponseFirstField.addActionListener(tfh);
	ResponseLastField.addActionListener(tfh);
	blankFromField.addActionListener(tfh);
	blankToField.addActionListener(tfh);
	
	Panel BaselinePanel=new Panel();
	BaselinePanel.add(new JLabel("Baseline Frame Range:     "));
	BaselinePanel.add(BaselineFirstField);
	BaselinePanel.add(new JLabel("to:    "));
	BaselinePanel.add(BaselineLastField);
	gd.addPanel(BaselinePanel);
	
	Panel ResponsePanel=new Panel();
	ResponsePanel.add(new JLabel("Response Frame Range:  "));
	ResponsePanel.add(ResponseFirstField);
	ResponsePanel.add(new JLabel("to:    "));
	ResponsePanel.add(ResponseLastField);
	gd.addPanel(ResponsePanel);
	
	Panel FilterLPanel=new Panel();
	FilterXYBox=new Checkbox("",DoFilterXY);
	if (!FilterMean) XYFilterField=new JTextField(""+FilterXy, 4);
	else XYFilterField = new JTextField(""+Math.round(FilterXy), 4);
	FilterLPanel.add(FilterXYBox);
	if (DoFilterXY) {
		if (!FilterMean) SigmaLabel = new JLabel("Filter in X,Y-direction,   sigma (pixel):");
		else SigmaLabel = new JLabel("Filter in X,Y-direction,  radius (pixel):");
	} else SigmaLabel = new JLabel("Filter in X,Y-direction                             ");
	FilterLPanel.add(SigmaLabel);
	FilterLPanel.add(XYFilterField);
	if (!FilterMean) FilterTypeLabel = new Label("  (Gaussian)  ");
	else FilterTypeLabel = new Label("  (Mean)  ");
	FilterLPanel.add(FilterTypeLabel);
	
	TextFieldHandler LHhandler = new TextFieldHandler();
	XYFilterField.addActionListener(LHhandler );
	XYFilterField.setEnabled(DoFilterXY);
	FilterXYBox.addItemListener(new ItemListener() {
		public void itemStateChanged(ItemEvent e) {
			DoFilterXY = FilterXYBox.getState();
			XYFilterField.setEnabled(DoFilterXY);
			DoFilterXY0 = DoFilterXY;
			if (DoFilterXY) {
				double FilterXyN = s2d (XYFilterField.getText(), FilterXy);
				if (FilterMean) {
					if (FilterXyN < 0.0 || Math.round(FilterXyN) != Math.round(FilterXy)) FilterXy = FilterXyN;
				} else if (FilterXyN < 0.0 || FilterXyN != FilterXy) FilterXy = FilterXyN;
				if (verboseMode && 0 == FilterXy) {
					if (!FilterMean) IJ.error("To allow filtering, set sigma to a value > 0.");
					else IJ.error("To allow filtering, set radius to a natural value >= 1.");
				}
				if (!FilterMean) SigmaLabel.setText("Filter in X,Y-direction,   sigma (pixel):");
				else SigmaLabel.setText("Filter in X,Y-direction,  radius (pixel):");
			} else SigmaLabel.setText("Filter in X,Y-direction                             ");
		}
	});
	gd.addPanel(FilterLPanel);
	
	Panel ZFilterPanel=new Panel();
	ZFilterAvBox = new Checkbox("", DoZFilterAv);
	ZFilterPanel.add(ZFilterAvBox);
	ZFilterField =new JTextField(""+FilterZ,4);
	if (DoZFilterAv) FramesLabel = new JLabel("Filter in Z-direction, average  voxel:");
	else FramesLabel = new JLabel("Filter in Z-direction                                 ");
	ZFilterPanel.add(FramesLabel);
	ZFilterPanel.add(ZFilterField);
	ZFilterAvBox.addItemListener(new ItemListener() {
		public void itemStateChanged(ItemEvent e) {
			DoZFilterAv=ZFilterAvBox.getState();
			ZFilterField.setEnabled(DoZFilterAv);
			DoZFilterAv0=DoZFilterAv;
			if (DoZFilterAv) {
				FramesLabel.setText("Filter in Z-direction, average  voxel:");
				FilterZ = s2i (ZFilterField2.getText(), FilterZ);
				if (verboseMode && 0 >= FilterZ) {
					String zMsg = "To allow filtering across frames,\n";
					zMsg +="set frame range ('average  voxel')\n";
					zMsg +="to a natural value >= 1.";
					IJ.error(zMsg);
				}
			} else FramesLabel.setText("Filter in Z-direction                                 ");
		}
	});
	TextFieldHandler Nhandler = new TextFieldHandler();
	ZFilterField.addActionListener(Nhandler );
	ZFilterField.setEnabled(DoZFilterAv);
	gd.addPanel(ZFilterPanel);
	
	Panel blankPanel=new Panel();
	doBlankBox = new Checkbox("", blankArtifact);
	boolean blankArtifact0 = blankArtifact;
	if (allowBlankArtifact) {
		blankPanel.add(doBlankBox);
		blankPanel.add(new JLabel("Interpolate Artifact Frames Range:   "));
		blankPanel.add(blankFromField);
		blankPanel.add(new JLabel("to:    "));
		blankPanel.add(blankToField);
		gd.addPanel(blankPanel);
		
		blankArtifact = blankArtifact && 0 < blankFrom &&  0 < blankTo;
		doBlankBox.setState(blankArtifact);
		if (blankArtifact) {
			blankFromField.setText (""+blankFrom);
			blankToField.setText (""+blankTo);
		} else {
			blankFromField.setText ("");
			blankToField.setText ("");
		}
		blankFromField.setEnabled(blankArtifact);
		blankToField.setEnabled(blankArtifact);
		
		doBlankBox.addItemListener(new ItemListener() {
			public void itemStateChanged(ItemEvent e) {
				blankArtifact = doBlankBox.getState();
				int xblankFrom = blankFrom;
				int xblankTo = blankTo;
				if (blankArtifact) {
					int tbaselinelast = s2i (BaselineLastField.getText(), baselinelast);
					int tresponsefirst = s2i (ResponseFirstField.getText(), responsefirst);
					int tresponselast = s2i (ResponseFirstField.getText(), responselast);
					if (2 > tresponsefirst - tbaselinelast) {
						String eMsg = "Interpolation range must be between 'baseline' and 'response' range.\n";
						eMsg += "               Please reduce upper limit of 'baseline' (to a value <"+tbaselinelast+")\n";
						eMsg += "               or increase lower limit of 'response' (to a value >"+tresponsefirst+").";
						IJ.error(eMsg);
						blankArtifact = false;
					}
					if (blankFrom <= tbaselinelast) blankFrom = tbaselinelast + 1;
					if (0 == blankTo || blankTo > tresponselast) 
							blankTo = (tresponsefirst + tbaselinelast) / 2;
					if (blankTo < blankFrom) blankTo = blankFrom;
					if (blankTo >= tresponsefirst) blankTo = tresponsefirst - 1;
					if (verboseMode && (blankFrom != xblankFrom || blankTo != xblankTo))
							new MessageDialog(new Frame(), PluginTitle, 
							"Interpolation limits changed.");
				}
				if (blankArtifact) {
					blankFromField.setText (""+blankFrom);
					blankToField.setText (""+blankTo);
				} else {
					blankFromField.setText ("");
					blankToField.setText ("");
				}
				doBlankBox.setState(blankArtifact);
				blankFromField.setEnabled(blankArtifact);
				blankToField.setEnabled(blankArtifact);
			}
		});
	} // allowBlankArtifact
	
	Panel ButtonPanel = new Panel();
	operationButton = new Button(" Output Options ");
	ButtonPanel.add(operationButton);
	filterSetButton = new Button(" Filter settings ");
	ButtonPanel.add(filterSetButton);
	optionButton = new Button(" Preferences ");
	ButtonPanel.add(optionButton);
	helpButton0 = new Button ("       Help       ");
	ButtonPanel.add(helpButton0);
	gd.addPanel(ButtonPanel);
	
	operationButton.addActionListener( handler );
	filterSetButton.addActionListener( handler );
	optionButton.addActionListener( handler );
	helpButton0.addActionListener( handler );
	
	//  workaround to allow tracking changes in inputFolderChoice
	DirIn0 = DirIn;  // -- keeps the initial selection
	
	//  keep other options that possibly may be changed without actual meaning
	//  	for later undo
	int blankFrom0 = blankFrom;
	int blankTo0 = blankTo;
	//  more ...	#######
	//  filter settings etc.
	
	gd.showDialog();
	
	// if (testMode) IJ.error((String) inputFolderChoice.getSelectedItem());  // +++++++++++
	if (Canceled || gd.wasCanceled()) {
		blankArtifact = blankArtifact0;
		IJ.showStatus("");
		return false;
	}
	if (!ShowAv&&!SaveAv && ((!ShowNo && !SaveNo) || !hasControl || 1==nhl.length)) 
		while (!ShowAv&&!SaveAv&& ((!ShowNo && !SaveNo) || !hasControl || 1==nhl.length) &&!ShowDelta&&!SaveDelta&&!ShowPerc&&!SavePerc&&!ShowDiv&&!SaveDiv&&!ShowRaw&&!ShowBase&&!SaveBase&&!ShowResp&&!SaveResp&&!ShowDiff&&!SaveDiff && !ShowPStack && !SavePStack && ((!ShowRatio && !SaveRatio) || 2 > nhl.length || (hasControl && 3 >nhl.length))  ) {
			if (verboseMode) IJ.error("Please select one option!");  //  #### let the user allow choosing ShowQual || SaveQual only ?
			if (showOperationDialog()) return showDialog();
		}
	baselinefirst = s2i (BaselineFirstField.getText(), baselinefirst);
	baselinelast = s2i (BaselineLastField.getText(), baselinelast);
	responsefirst = s2i (ResponseFirstField.getText(), responsefirst);
	responselast = s2i (ResponseLastField.getText(), responselast);
	DoFilterXY = FilterXYBox.getState();
	if (DoFilterXY) {
		double FilterXyN = s2d (XYFilterField.getText(), FilterXy);
		if (FilterMean) {
			if (FilterXyN < 0.0 || Math.round(FilterXyN) != Math.round(FilterXy)) FilterXy = FilterXyN;
		} else if (FilterXyN < 0.0 || FilterXyN != FilterXy) FilterXy = FilterXyN;
	}
	DoZFilterAv = ZFilterAvBox.getState();
	if (DoZFilterAv) FilterZ = s2i (ZFilterField.getText(), FilterZ);
	if (blankArtifact && allowBlankArtifact) {
		blankFrom = s2i (blankFromField.getText (), blankFrom);
		blankTo = s2i (blankToField.getText (), blankTo);
		blankArtifact = blankArtifact && 0 < blankFrom &&  0 < blankTo;
	}
	if (!blankArtifact || !allowBlankArtifact) {
		blankFrom = blankFrom0;
		blankTo = blankTo0;
	}
	if (!LayoutChanged) SetIgnore();
	if (NormCB.isEnabled()) Normalizeb = NormCB.getState();
	
	boolean showAgain = false;   // true if dialog should be shown again (will call the procedure recursively)
	String sel=""+inputFolderChoice.getSelectedItem();
	if (0!=DirIn0.compareTo(sel)) {
		File directory=null;
		SetIgnore();
		String tdir=sel.trim();
		while (tdir.endsWith("*") || tdir.endsWith(".")) tdir=tdir.substring(0, tdir.length()-1);
		pi=inputFolderChoice.getSelectedIndex();
		try {
			directory =new File(tdir);
			if (directory.isFile()) {
				directory = new File(directory.getParent());
				if (null==directory) directory=new File(tdir);
			}
			while (!directory.isDirectory() && minSubDirLevel<howOftenIn(BSL, tdir)) {
				tdir=tdir.substring(0, tdir.lastIndexOf(BSL));
				directory=new File(tdir);
			}
			if (directory.isDirectory()) {
				tdir=""+directory;
				DirIn=tdir;
				IboxStartRow=0;
				RepaintIgnoreBoxes("");
				SetCheck(!silentMode);
			} else {
				if (verboseMode) IJ.error("Folder ''"+ sel + "'' not found!");
				pi = -2;
			}
		}
		catch (SecurityException ex) {
			if (verboseMode) IJ.error("Could not access Folder ''"+ tdir + "'' !");
			pi = -2;
		}
		catch (NullPointerException ex) {
			if (verboseMode) IJ.error("Could not access Folder!");
			pi = -2;
		}
		if (0 <= pi) {
			inputFolderChoice.removeItemAt(pi);
			// --  select item if already in collection, otherwise add new item
			int si=-1;
			for (int i=0; i<inputFolderChoice.getItemCount(); i++) 	if (0==DirIn.compareTo(inputFolderChoice.getItemAt(i).toString())) si=i;
			if (si<0) {
				inputFolderChoice.insertItemAt(DirIn, pi);
				inputFolderChoice.setSelectedIndex(pi);
			} else inputFolderChoice.setSelectedIndex(si);
		} else if (-2 == pi) {
			pi = 0;
			for (int i=0; i<inputFolderChoice.getItemCount(); i++) 	if (0==DirIn.compareTo(inputFolderChoice.getItemAt(i).toString())) pi=i;
			inputFolderChoice.setSelectedIndex(pi);
		} else inputFolderChoice.insertItemAt(DirIn, 0);
		inputFolderChoice.repaint();
		// IJ.showStatus("Select files in " + DirIn);
		PDir= new String [inputFolderChoice.getItemCount()];
		for (int i=0;i<PDir.length;i++) PDir[i]=inputFolderChoice.getItemAt(i).toString();
		DirIn=""+directory;
		LayoutChanged=false;
		showAgain = true;
	}
	
	if (LayoutChanged) {
		LayoutChanged=false;
		showAgain = true;
	}
	
	//  ----------  test whether there are enough files with data to calculate in the selected folder
	if (!showAgain) {
		int xi=nhl.length;
		int xj=0;
		for (int i=0; i<xi; i++) if (iob[1][i].length>xj) xj=iob[1][i].length;
		boolean NoData = true;
		if (1<nhl.length && hasControl) {for (int i=1; i<xi; i++) if (NoData) for (int j = 1; j<xj; j++) if (!iob[1][i][j]) NoData = false;}
		else {for (int i=0; i<xi; i++) if (NoData) for (int j = 0; j<xj; j++) if (!iob[1][i][j]) NoData = false;}
		if (NoData) {
			IJ.error("No Files to load!");
			showAgain = true;
		}
	}
	if (showAgain) return showDialog();
	return true;
}

private int howOftenIn(String whatS, String inS) {
	if (0==whatS.length()) return 0;
	int cnt =0;
	for (int pos=0; pos<inS.length(); pos++) if (whatS.charAt(0)== inS.charAt(pos)) cnt++;
	return cnt;
}

private class ButtonHandler implements ActionListener {
	public void actionPerformed( ActionEvent e ) {
		if (e.getSource() == inputFolderButton){
			FileDialog fd=new FileDialog(IJ.getInstance(), "Select one file out of the file ensemble that you wish to open!");
			String Fset=""; // -- Let the user select only for files that match the mask.
			for (int i=0;i<nhl.length-1;i++) Fset=Fset+nhl[i]+"*.tif; ";
			Fset=Fset+nhl[nhl.length-1]+"*.tif";
			fd.setFile(Fset);
			File directory=null;
			String tdir=DirIn;
			try {
				directory=new File(DirIn);
				while (!directory.isDirectory() && minSubDirLevel<howOftenIn(BSL, tdir)) {
					tdir=tdir.substring(0,tdir.lastIndexOf(BSL));
					directory=new File(tdir);
				}
				if (directory.isDirectory()) fd.setDirectory(""+directory);
			}
			catch (SecurityException ex) {if (verboseMode) IJ.error("Could not access folder '"+DirIn+"'!");}
			catch (NullPointerException ex) {if (verboseMode) IJ.error("Could not access folder '"+DirIn+"'!");}
			
			if (null != fd.getFile()) {
				fd.setVisible(true);
				// update field and variable values
				directory =new File(fd.getDirectory());
				// reset checkboxes and fields only if the user has changed the folder
				if (DirIn != directory + "")  {
					DirIn = directory + "";
					int pi=-1;
					if (!IJ.isWindows()) {for (int pc=0;pc<PDir.length;pc++) if(0>pi) if (0==DirIn.compareTo(PDir[pc])) pi=pc;}
					if(0>pi) {for (int pc=0;pc<PDir.length;pc++) if(0>pi) if (0==DirIn.compareToIgnoreCase(PDir[pc])) pi=pc;}
					if (0>pi) {
						String [] TDir = new String [PDir.length];
						for (int tc=0;tc<TDir.length;tc++) TDir[tc]=PDir[tc];
						PDir= new String [PDir.length+1];
						PDir[0]=DirIn;
						for (int tc=0;tc<TDir.length;tc++) PDir[tc+1]=TDir[tc];
						inputFolderChoice.insertItemAt(DirIn, 0);
						inputFolderChoice.setSelectedIndex(0);
					} else inputFolderChoice.setSelectedIndex(pi);
					inputFolderChoice.repaint();
					IboxStartRow=0;
					RepaintIgnoreBoxes("");
					SetCheck(!silentMode);  // -- update checkboxes
					DirIn0 = DirIn;
					if ((filesize != filesize0 || 0 == filesize) && blankArtifact && allowBlankArtifact) {
						blankArtifact=false;
						doBlankBox.setState(false);
						blankFromField.setText ("");
						blankToField.setText ("");
						blankFromField.setEnabled(false);
						blankToField.setEnabled(false);
					}
				}
			}
			IJ.showStatus("Select files in " + DirIn);
		} else if (e.getSource() == operationButton) {
			showOperationDialog();
			if (AMethod==SD_METHOD || AMethod==SE_METHOD) {
				NormCB.setLabel("Subtract & normalize to"+AVL[AVG_METHOD]+"stack "+"'"+nhl[0]+AVS[AVG_METHOD]+".tif'");
			} else {
				NormCB.setLabel("Subtract & normalize to"+AVL[AMethod]+"stack "+"'"+nhl[0]+AVS[AMethod]+".tif'");
			}
			IJ.showStatus("Select files in " + DirIn);
		} else if (e.getSource() == filterSetButton) {
			IJ.showStatus("Modify Filter Settings...");
			if (DoFilterXY) {
				double FilterXyN = s2d (XYFilterField.getText(), FilterXy);
				if (FilterMean) {
					if (FilterXyN < 0.0 || Math.round(FilterXyN) != Math.round(FilterXy)) FilterXy = FilterXyN;
				} else if (FilterXyN < 0.0 || FilterXyN != FilterXy) FilterXy = FilterXyN;
			}
			if (DoZFilterAv) {
				int FilterZn = s2i (ZFilterField.getText(), FilterZ);
				if (FilterZn < 0 || FilterZn != FilterZ) {
					if (FilterZn >= 0) FilterZ = FilterZn;
					ZFilterField.setText(""+FilterZ);
				}
			}
			GenericDialog gd2 = new GenericDialog(PluginTitle, IJ.getInstance());
			Panel ZeroPanel=new Panel();
	   		ZeroPanel.add(new JLabel("Filter Settings"));
			try {gd2.addPanel(ZeroPanel);}
			catch (NoSuchMethodError er) {
				String vMsg = "This plugin requires a newer version of ImageJ.\n";
				vMsg+="Please update and run again.";
				IJ.error(vMsg);
				return;
			}
			
			Panel IPanel =new Panel();
			CheckboxGroup cbgi = new CheckboxGroup();
			ignoreNBox = new Checkbox(" ignore data", cbgi, !interpolateXY);
			interpolateBox = new Checkbox(" interpolate value", cbgi, interpolateXY);
			IPanel.setLayout(new GridLayout(3, 1));
	   		IPanel.add(new Label("Treatment of NaN and Infinity values (32 bit data)"));
			IPanel.add(ignoreNBox);
   			IPanel.add(interpolateBox);
			gd2.addPanel(IPanel);
			
			Panel FilterLPanel2=new Panel();
			final Checkbox FilterXYBox2=new Checkbox("Filter in X,Y-direction        ",DoFilterXY);
			SigmaLabel2 = new JLabel("  sigma:");
			if (FilterMean) XYFilterField2=new JTextField(""+FilterXy, 4);
			else XYFilterField2 = new JTextField(""+Math.round(FilterXy), 4);

			FilterLPanel2.add(FilterXYBox2);
			FilterLPanel2.add(SigmaLabel2);
			SigmaLabel2.setEnabled(DoFilterXY);
			FilterLPanel2.add(XYFilterField2);
			gd2.addPanel(FilterLPanel2);
			
			TextFieldHandler LHhandler2 = new TextFieldHandler();
			XYFilterField2.addActionListener(LHhandler2);
			XYFilterField2.setEnabled(DoFilterXY);
			
			Panel XFilterPanel=new Panel();
			XFilterPanel.setLayout(new GridLayout(2, 2));
			XFilterLabel = new Label("Filter Type:");
			XFilterPanel.add(XFilterLabel);
			XFilterLabel.setEnabled(DoFilterXY);
			CheckboxGroup cbgr = new CheckboxGroup();
			GaussianFBox = new Checkbox(" Gaussian  ", cbgr, DoFilterXY && !FilterMean);
			XFilterPanel.add(GaussianFBox);
			MeanFBox = new Checkbox(" Mean  ", cbgr, DoFilterXY && FilterMean);
			MeanFBox.setEnabled(DoFilterXY);
			XFilterPanel.add(new Label(""));
			XFilterPanel.add(MeanFBox);
			NoFilterBox = new Checkbox("", cbgr, !DoFilterXY);   // hidden checkbox
			boolean mustBeMean = true;
			try {mustBeMean = IJ.getVersion().compareTo("1.33c") < 0;}
			catch (NoSuchMethodError ex) {}
			if (mustBeMean) {
				MeanFBox.setState(DoFilterXY);
				MeanFBox.setEnabled(DoFilterXY);
				GaussianFBox.setState(false);
				GaussianFBox.setEnabled(false);
				XYFilterField2.setText(""+Math.round(FilterXy));
			} else {
				GaussianFBox.setEnabled(DoFilterXY);
				ItemListener itf = new ItemListener(){
					public void itemStateChanged(ItemEvent e) {
						double FilterXyN = s2d (XYFilterField2.getText(), FilterXy);
						if (GaussianFBox.getState()) {
							//  FilterTypeLabel.setText("  (Gaussian)  ");
							SigmaLabel2.setText("  sigma:");
							if (Math.round(FilterXyN) == Math.round(FilterXy)) FilterXyN = FilterXy;
							//  FilterXy0  #######
							XYFilterField2.setText(""+FilterXyN);
						} else if (MeanFBox.getState()) {
							//  FilterTypeLabel.setText("  (Mean)  ");
							XYFilterField2.setText(""+Math.round(FilterXyN));
							SigmaLabel2.setText(" radius:");
						}
					}
				};
				MeanFBox.addItemListener(itf);
				GaussianFBox.addItemListener(itf);
			}
			if (FilterMean) {
				SigmaLabel2.setText(" radius:");
				MeanFBox.setLabel(" Mean  ");
			}
			String fLabel=(0<counter[0][0]&&1<nhl.length&&Normalizeb && hasControl)?" (before normalization)":"";
			FilterAvBox = new Checkbox(" Filter averaged stacks"+fLabel, DoFilterAv && DoFilterXY);
			FilterAvBox.setEnabled(DoFilterXY);
			FilterBRBox = new Checkbox(" Filter baseline/ response projection (default)",DoFilterBR && DoFilterXY);
			FilterBRBox.setEnabled(DoFilterXY);
			fLabel=(0<counter[0][0]&&1<nhl.length&&Normalizeb && hasControl)?"F":" I ";
			fLabel+="/"+fLabel;
			FilterDeltaBox = new Checkbox(" Filter before "+DST+fLabel+SZS+" calculation",DoFilterDelta && DoFilterXY);
			FilterDeltaBox.setEnabled(DoFilterXY);
			gd2.addPanel(XFilterPanel);
			
			ItemListener ifxy = new ItemListener() {
				public void itemStateChanged(ItemEvent e) {
					boolean DoFilterXYn = FilterXYBox2.getState();
					boolean DoFilterAvN = FilterAvBox.getState();
					boolean DoFilterBRn = FilterBRBox.getState();
					boolean DoFilterDeltaN = FilterDeltaBox.getState();
					if (DoFilterXYn && !DoFilterAvN && !DoFilterBRn && !DoFilterDeltaN) {
						if (FilterXYBox2 == e.getSource()) {
							if (DoFilterAv || DoFilterBR || DoFilterDelta) {
								DoFilterAvN = DoFilterAv;
								DoFilterBRn = DoFilterBR;
								DoFilterDeltaN = DoFilterDelta;
							} else DoFilterBRn = true;
						} else DoFilterXYn = false;
					}
					
					// -- check sigma (or radius) presettings
					double FilterXyN = s2d (XYFilterField2.getText(), FilterXy);
					if (!FilterMean) FilterXyN = Math.rint(FilterXyN);
					if (verboseMode && 0 == FilterXyN && FilterXYBox2 == e.getSource()) {
						if (!FilterMean) IJ.error("To allow filtering, set sigma to a value > 0.");
						else IJ.error("To allow filtering, set radius to a natural value >= 1.");
					}
					
					FilterXYBox2.setState(DoFilterXYn);
					XYFilterField2.setEnabled(DoFilterXYn);
					SigmaLabel2.setEnabled(DoFilterXYn);
					XFilterLabel.setEnabled(DoFilterXYn);
					boolean mustBeMean2 = true;   //  make it  a global variable  ######
					try {mustBeMean2 = IJ.getVersion().compareTo("1.33c") < 0;}
					catch (NoSuchMethodError ex) {}
					if (mustBeMean2) {
						GaussianFBox.setEnabled(false);
						GaussianFBox.setState(false);
						MeanFBox.setEnabled(DoFilterXYn);
						MeanFBox.setState(DoFilterXYn);
						NoFilterBox.setState(!DoFilterXYn);  // hidden checkbox
					} else {
						GaussianFBox.setEnabled(DoFilterXYn);
						GaussianFBox.setState(DoFilterXYn && !FilterMean);
						MeanFBox.setEnabled(DoFilterXYn);
						MeanFBox.setState(DoFilterXYn && FilterMean);
						NoFilterBox.setState(!DoFilterXYn);  // hidden checkbox
					}
					FilterAvBox.setEnabled(DoFilterXYn);
					FilterAvBox.setState(DoFilterAvN && DoFilterXYn);
					FilterBRBox.setEnabled(DoFilterXYn);
					FilterBRBox.setState(DoFilterBRn && DoFilterXYn);
					FilterDeltaBox.setEnabled(DoFilterXYn);
					FilterDeltaBox.setState(DoFilterDeltaN && DoFilterXYn);
				}
			};
			FilterXYBox2.addItemListener(ifxy);
			FilterAvBox.addItemListener(ifxy);
			FilterBRBox.addItemListener(ifxy);
			FilterDeltaBox.addItemListener(ifxy);
			Panel FOPanel=new Panel();
			FOPanel.setLayout(new GridLayout(0,1));
			FOPanel.add(FilterAvBox);
			FOPanel.add(FilterBRBox);
			FOPanel.add(FilterDeltaBox);
			FOPanel.add(new Label (""));
			gd2.addPanel(FOPanel);
			
			Panel ZFilterPanel2=new Panel();
			ZFilterAvBox2 = new Checkbox("Filter in Z-direction       ", DoZFilterAv);
			ZFilterPanel2.add(ZFilterAvBox2);
			FramesLabel2 = new JLabel("  frames:");
			FramesLabel2.setEnabled(DoZFilterAv);
			ZFilterPanel2.add(FramesLabel2);
			ZFilterField2 =new JTextField(""+FilterZ,4);
			ZFilterField2.setEnabled(DoZFilterAv);
			ZFilterPanel2.add(ZFilterField2);
			ZFilterAvBox2.addItemListener(new ItemListener() {
				public void itemStateChanged(ItemEvent e) {
					boolean DoZFilterAv2=ZFilterAvBox2.getState();
					DoZFilterAv=DoZFilterAv2;
					ZFilterField2.setEnabled(DoZFilterAv);
					FramesLabel2.setEnabled(DoZFilterAv);

					// -- check radius (+/- frames) presettings
					int FilterZn = s2i (ZFilterField2.getText(), FilterZ);;
					if (verboseMode && 0 >= FilterZn) {
						String zMsg = "To allow filtering across frames,\n";
						zMsg +="set frame range ('average  voxel')\n";
						zMsg +="to a natural value >= 1.";
						IJ.error(zMsg);
					}
				}
			});
			gd2.addPanel(ZFilterPanel2);
			
			Panel HPanel2 =new Panel();
			HPanel2.add(new Label("                                                                     "));
			helpButton2 = new Button ("  Help  ");
			HPanel2.add(helpButton2);
			ButtonHandler handler = new ButtonHandler();
			helpButton2.addActionListener( handler );
			gd2.addPanel(HPanel2);
			
			gd2.showDialog();
			if (!gd2.wasCanceled()) {
				// -- update values in main dialog with changes applied by this filter dialog
				IJ.showStatus("Filter settings changed or updated.");
				DoFilterXY = FilterXYBox2.getState();
				FilterXYBox.setState(DoFilterXY);
				XYFilterField.setEnabled(DoFilterXY);
				if (DoFilterXY) {
					if (GaussianFBox.getState()) {
						FilterMean=false;
						FilterTypeLabel.setText("  (Gaussian)  ");
					} else if (MeanFBox.getState()) {
						FilterMean=true;
						FilterTypeLabel.setText("  (Mean)  ");
					}
					DoFilterAv=FilterAvBox.getState();
					DoFilterBR=FilterBRBox.getState();
					DoFilterDelta=FilterDeltaBox.getState();
					double FilterXyN = s2d (XYFilterField2.getText(), FilterXy);
					if (!FilterMean) {
						if (FilterXyN < 0.0 || Math.round(FilterXyN) != Math.round(FilterXy)) FilterXy = FilterXyN;
						SigmaLabel.setText("Filter in X,Y-direction,  radius (pixel):");
						XYFilterField.setText(""+Math.round(FilterXy));
					} else {
						if (FilterXyN < 0.0 || FilterXyN != FilterXy) FilterXy = FilterXyN;
						SigmaLabel.setText("Filter in X,Y-direction,   sigma (pixel):");
						XYFilterField.setText(""+FilterXy);
					}
				} else {
					// XYFilterField.setText("0");
					SigmaLabel.setText("Filter in X,Y-direction                             ");
				}
				// -- z filter values
				DoZFilterAv=ZFilterAvBox2.getState();
				ZFilterAvBox.setState(DoZFilterAv);
				ZFilterField.setEnabled(DoZFilterAv);
				if (DoZFilterAv) {
					FilterZ = s2i (ZFilterField2.getText(), FilterZ);
					ZFilterField.setText(""+FilterZ);
					FramesLabel.setText("Filter in Z-direction, average  voxel:");
				} else {
					// ZFilterField.setText("0");
					FramesLabel.setText("Filter in Z-direction                                 ");
				}
				// -- interpolation (instead of ignoring) NaN values
				interpolateXY = interpolateBox.getState();
			} else IJ.showStatus("Select files in " + DirIn);
		} else if (e.getSource() == optionButton) {
			IJ.showStatus("Modify plugin preferences...");
			GenericDialog gd2 = new GenericDialog(PluginTitle, IJ.getInstance());
			Panel ZeroPanel=new Panel();
			ZeroPanel.add(new JLabel("Layout  preferences:"));
			try {gd2.addPanel(ZeroPanel);}
			catch (NoSuchMethodError er) {
				String vMsg = "This plugin requires a newer version of ImageJ.\n";
				vMsg+="Please update and run again.";
				IJ.error(vMsg);
				return;
			}
			
			Panel NCPanel=new Panel();
			NCField=new JTextField(""+NbChkbxs, 4);
			NCPanel.add(new Label("Number of ''Ignore''-boxes    "));
			NCPanel.add(NCField);
			TextFieldHandler Nbhandler  = new TextFieldHandler();
			NCField.addActionListener(Nbhandler);
			gd2.add(NCPanel);
			
			Panel NHLPanel=new Panel();
			String nhls=nhl[0];
			for (int i=1; i<nhl.length; i++) nhls+=(", "+nhl[i]);
			NHLField=new JTextField(nhls, 8);
			NHLPanel.add(new Label("File name(s),''#.tif''   "));
			NHLPanel.add(NHLField);
			TextFieldHandler Nhandler = new TextFieldHandler();
			NHLField.addActionListener(Nhandler);
			
			Panel HconPanel=new Panel();
			hasConBox=new Checkbox(" Control files    ", hasControl && 1<nhl.length);
			hasConBox.setEnabled(1<nhl.length);
			String [] nhlC = new String [nhl.length];
			String sp = "";
			for (int i=nhl[0].length()*2;i<21; i++) sp+=" ";
			nhlC[0]=nhl[0]+sp;
			for (int i=1;i<nhl.length;i++) nhlC[i]=nhl[i];

			// Choice noChoice = new Choice();
			// for (int i=0;i<nhl.length;i++) noChoice.add(nhl[i]);

			conSelCBox = new JComboBox(nhlC);
			hasConBox.addItemListener(new ItemListener() {
				public void itemStateChanged(ItemEvent e) {
					if (hasConBox.getState()) {
						conSelCBox.setSelectedIndex(0);
						conSelCBox.setEnabled(true);
					} else {
						// noChoice.removeAllItems();
						conSelCBox.setSelectedIndex(-1);
						conSelCBox.setEnabled(false);
					}
				}
			});
			TextFieldHandler chandler  = new TextFieldHandler();
			conSelCBox.addActionListener(chandler);
			
			HconPanel.add(hasConBox);
			HconPanel.add(conSelCBox);
			//  HconPanel.add(noChoice);
			if (1<nhl.length && hasControl) {
				conSelCBox.setSelectedIndex(0);
				conSelCBox.setEnabled(true);
			} else {
				conSelCBox.setSelectedIndex(-1);
				conSelCBox.setEnabled(false);
			}
			
			Panel LtPanel=new Panel();
			LtPanel.setLayout(new GridLayout(0,1));
			LtPanel.add(NCPanel);
			LtPanel.add(NHLPanel);
			LtPanel.add(HconPanel);
			gd2.addPanel(LtPanel);
			
			Panel bPanel =new Panel();
			aBlankBox = new Checkbox("Allow  artifact  z-interpolating  option", allowBlankArtifact);
			//  may be neded for later tracking of changes in allowBlankArtifact
			boolean currentAllowBlankArtifact = allowBlankArtifact;
			bPanel.add(aBlankBox);
			//  disable displaying additional elements on main dialog 
			//	if there is not enough space on screen for it 
			if (2 == getMaxNbChkbxs(true)) {
				allowBlankArtifact = false;   // no more elements
   				aBlankBox.setEnabled(false);
   				aBlankBox.setState(false);
			}
   			gd2.addPanel(bPanel);
			
			Panel MidPanel=new Panel();
			MidPanel.add(new JLabel("Operational  preferences:"));
			gd2.addPanel(MidPanel);
			
			Panel RPanel =new Panel();
			CheckboxGroup cbg2 = new CheckboxGroup();
			verboseBox = new Checkbox("verbose",cbg2,verboseMode);
			normmBox = new Checkbox("normal",cbg2,!silentMode&&!verboseMode);
			silentBox = new Checkbox("silent",cbg2,silentMode);
			RPanel.setLayout(new GridLayout(2,2));
	   		RPanel.add(new Label("Responsiveness:"));
			RPanel.add(verboseBox);
   			RPanel.add(normmBox);
			RPanel.add(silentBox);
			gd2.addPanel(RPanel);
			
			Panel KPanel =new Panel();
			CheckboxGroup cbg = new CheckboxGroup();
			dontKeepBox = new Checkbox("don't keep  ",cbg,DONT_KEEP==storeModus);
			localKeepBox = new Checkbox("local",cbg,KEEP_LOCAL==storeModus);
			personalKeepBox = new Checkbox("personal",cbg,KEEP_PERSONAL==storeModus);
			globalKeepBox = new Checkbox("global",cbg,KEEP_GLOBAL==storeModus);
			customKeepBox = new Checkbox("custom",cbg,KEEP_CUSTOM==storeModus);
			int storeModus0 = storeModus;
			String Pth0 = Pth;
			ItemListener itl = new ItemListener() {
				public void itemStateChanged(ItemEvent e) {
					int tstoreModus = KEEP_GLOBAL;
					if (dontKeepBox.getState()) tstoreModus=DONT_KEEP;
					else if (localKeepBox.getState()) tstoreModus=KEEP_LOCAL;
					else if (personalKeepBox.getState()) tstoreModus=KEEP_PERSONAL;
					else if (customKeepBox.getState()) tstoreModus=KEEP_CUSTOM;
					if (tstoreModus != storeModus) {
						if (tstoreModus != DONT_KEEP) {
							String tPth = Pth;
							String pPth = Pth;
							int pstoreModus = storeModus;
							storeModus = tstoreModus;
							
							if (storeModus==KEEP_CUSTOM) {
								boolean askAgain = true;
								do {
									IJ.showStatus("New preferences path");
									File tdirectory = new File(tPth);
									if (null != tdirectory) {
										IJ.showStatus("please wait...");
										JFileChooser chooser = new JFileChooser();
										chooser.setCurrentDirectory(tdirectory);
										chooser.setDialogTitle("Select position (directory) where to keep settings - "+PluginTitle);
										chooser.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
										chooser.setAcceptAllFileFilterUsed(false);
										IJ.showStatus("New preferences path");
										if (chooser.showOpenDialog(null) == JFileChooser.APPROVE_OPTION) { 
											tPth = ""+ chooser.getSelectedFile();
										} else {
											tPth = null;
										}
									} else {
										tPth = (String)JOptionPane.showInputDialog(
										null, "Enter new path to save preferences:", PluginTitle, 
										JOptionPane.PLAIN_MESSAGE, null, null, tPth);
									}
									if (null != tPth) {
										tdirectory = new File(tPth);
										tPth = "" + tdirectory;
									} else tdirectory = null;
									if (null == tdirectory) askAgain = false;
									else {
										askAgain = !checkSettingsPath(tPth);
										if (askAgain) {
											Object[] options = {"OK", "Cancel"};
											String dPth = tPth;
											if (30 < dPth.length()) dPth = dPth.substring(0,27) + "...";
											if (1 == JOptionPane.showOptionDialog(null, 
												"Failed to use path\n'"+ dPth + "'\nfor keeping settings", 
												PluginTitle,  JOptionPane.DEFAULT_OPTION, 
												JOptionPane.WARNING_MESSAGE, null, options, options[0])) {
													askAgain = false;
													tPth = null;
											}
										} else {
											IJ.showStatus("Preferences will be saved in " + tPth);
										}
									}
								} while (askAgain);
								gotSet = null != tPth;
								if (!gotSet) {
									storeModus = pstoreModus;
									IJ.showStatus("Modify plugin preferences...");
								} else Pth = tPth;
							} else {
								gotSet = findPathAndDoSettings ("", CHECK, !silentMode);
								if (!gotSet) {
									storeModus = pstoreModus;
									IJ.showStatus("Modify plugin preferences...");
								}
							}
							if (storeModus != pstoreModus) {
								File iniFile = null;
								try {
									if (Pth.endsWith(BSL)) iniFile = new File(Pth+IniFileName);
									else iniFile = new File(Pth+BSL+IniFileName);
									if (!iniFile.isFile()) iniFile = null;
								}
								catch (SecurityException ex) {}
								catch (NullPointerException ex) {}
								if (null != iniFile) {
									if (silentMode) {		// ###  !verboseMode  ? 
										if (doSettings(Pth, GET)) {
											updateDlg2 ();
											newSettings = true;
										} else {
											if (!silentMode) IJ.error("Failed to retrive settings after path change\nfrom '"+pPth+"'\nto '"+Pth+"'!");
											errString+=(", Failed to retrive settings after path change from '"+pPth+"' to '"+Pth+"'");
											Pth = pPth;
											gotSet = false;
											storeModus = pstoreModus;
											IJ.showStatus("Modify plugin preferences...");
										}
 									} else {
										String ques = "Reload settings from\n'"+Pth+"'?";
										switch (JOptionPane.showConfirmDialog(null, ques, PluginTitle,JOptionPane.YES_NO_CANCEL_OPTION)) {
											case JOptionPane.YES_OPTION:
												if (doSettings(Pth, GET)) {
													updateDlg2 ();
													newSettings = true;
												} else {
													if (!silentMode) IJ.error("Failed to retrive settings after path change\nfrom '"+pPth+"'\nto '"+Pth+"'!");
													errString+=(", Failed to retrive settings after path change from '"+pPth+"' to '"+Pth+"'");
													Pth = pPth;
													gotSet = false;
													storeModus = pstoreModus;
													IJ.showStatus("Modify plugin preferences...");
												}
												break;
											case JOptionPane.NO_OPTION:
												break;
											default:
												Pth = pPth;
												storeModus = pstoreModus;
												IJ.showStatus("Modify plugin preferences...");
										} // switch
									} // if(silentMode) {} else
								} // iniFile
							} // if (storeModus != pstoreModus)
							ResetButton.setEnabled(gotSet);
							dontKeepBox.setState(storeModus==DONT_KEEP);
							localKeepBox.setState(storeModus==KEEP_LOCAL);
							personalKeepBox.setState(storeModus==KEEP_PERSONAL);
							globalKeepBox.setState(storeModus==KEEP_GLOBAL);
							customKeepBox.setState(storeModus==KEEP_CUSTOM);
						} else {
							String retMsg="Setting changes won't be kept anymore for future Plugin runs.";
							if (verboseMode) new MessageDialog(new Frame(), PluginTitle, retMsg);
							IJ.showStatus(retMsg);
						}
					}
				}
			};  // ItemListener itl
			dontKeepBox.addItemListener(itl);
			localKeepBox.addItemListener(itl);
			personalKeepBox.addItemListener(itl);
			globalKeepBox.addItemListener(itl);
			customKeepBox.addItemListener(itl);
			
			KPanel.setLayout(new GridLayout(2,3));
	   		KPanel.add(new Label("Settings:"));
			KPanel.add(dontKeepBox);
   			// KPanel.add(new Label(""));
			KPanel.add(localKeepBox);
			KPanel.add(personalKeepBox);
			KPanel.add(globalKeepBox);
			KPanel.add(customKeepBox);
			gd2.addPanel(KPanel);
			
			Panel HPanel =new Panel();
			ApplyButton = new Button ("     Apply      ");  // -- Keep settings
			HPanel.add(ApplyButton);
			ResetButton = new Button (2 < storedSettings.length()?" Undo Reset ":"     Reset      ");
			ResetButton.setEnabled(gotSet);
			HPanel.add(ResetButton);
			// HPanel.add(new Label("         "));
			helpButton3 = new Button ("  Help  ");
			HPanel.add(helpButton3);
			ButtonHandler handler = new ButtonHandler();
			ApplyButton.addActionListener( handler );
			if (gotSet) ResetButton.addActionListener( handler );
			helpButton3.addActionListener( handler );
			gd2.addPanel(HPanel);
			gd2.showDialog();
			
			if (!gd2.wasCanceled()) {
				IJ.showStatus("Preferences changed or updated.");
				String [] nhl2 = null;
				nhl2=s2array(NHLField.getText(), nhl, false);
				if (0 < nhl2.length) {
					nhl = new String [nhl2.length];
					for (int i=0; i<nhl2.length; i++) nhl[i]=nhl2[i];
				}
				if (hasConBox.isEnabled()) hasControl=hasConBox.getState();
				if (!hasControl && 1<nhl.length) IboxStartColumn = 1;
				else IboxStartColumn = (2<nhl.length)?1:0;
				hasConBox.setEnabled (1<nhl.length);
				
				verboseMode=verboseBox.getState();
				if (verboseMode) silentMode=false;
				else silentMode = silentBox.getState();
				
				if (dontKeepBox.getState()) storeModus=DONT_KEEP;
				else if (localKeepBox.getState()) storeModus=KEEP_LOCAL;
				else if (personalKeepBox.getState()) storeModus=KEEP_PERSONAL;
				else if (customKeepBox.getState()) storeModus=KEEP_CUSTOM;
				else storeModus=KEEP_GLOBAL;
				
				IboxStartRow=0;
				Font f=IgnoreBox[0][1].getFont();
				int fstyle=(f.isBold()?Font.BOLD:Font.PLAIN);
				if (1<nhl.length && hasControl && IboxStartColumn==0) fstyle+=Font.ITALIC;
				f  = new Font(f.getName(), fstyle, f.getSize());
				for(int j=1;j<=NbChkbxs;j++) IgnoreBox[0][j].setFont(f);
				RepaintIgnoreBoxes("");
				SetCheck(verboseMode);
				WButton.setEnabled(0<IboxStartColumn);
				EButton.setEnabled(iob[1].length>IboxStartColumn+2);
				
				int NextNbChkbxs = s2i (NCField.getText(), NbChkbxs);
				if (aBlankBox.isEnabled()) allowBlankArtifact = aBlankBox.getState();
				if (storeModus==storeModus0) newSettings = false;
				if (NextNbChkbxs!=NbChkbxs || allowBlankArtifact !=currentAllowBlankArtifact || newSettings) {
					if (2 > NextNbChkbxs) NextNbChkbxs = NbChkbxs;
					MaxNbChkbxs = getMaxNbChkbxs(allowBlankArtifact);
					if (MaxNbChkbxs < NextNbChkbxs) NextNbChkbxs = MaxNbChkbxs;
					if (NextNbChkbxs!=NbChkbxs || allowBlankArtifact !=currentAllowBlankArtifact || newSettings) {
						LayoutChanged= true;
						newSettings = false;
						NbChkbxs = NextNbChkbxs;
						IJ.showStatus("restarting the " + PluginTitle);
						gd.dispose();  // restart the user dialog
					}
				}
			} else {
				IJ.showStatus("Select files in " + DirIn);
				storeModus = storeModus0;
				Pth = Pth0;
			}
		} else if (e.getSource() == NButton) {
			// -- one or more step(s) up
			int pRow=IboxStartRow;
			SetIgnore();
			boolean FileFound=false;
			// -- step up until in the new range to display is a file to select
			do {
				if (1==nhl.length) {
					IboxStartRow-=(2*NbChkbxs);
					if (0 > IboxStartRow) IboxStartRow = 0;
					for (int j=1; j<=2*NbChkbxs;  j++) if (!iob[0][IboxStartColumn][j+IboxStartRow]) FileFound=true;
				} else {
					IboxStartRow-=NbChkbxs;
					if (0 > IboxStartRow) IboxStartRow = 0;
					for (int i=IboxStartColumn; i<IboxStartColumn+2; i++)  for (int j=1; j <=NbChkbxs; j++) if (!iob[0][i][j+IboxStartRow]) FileFound=true;
				}
			} while (!FileFound && 0 < IboxStartRow);
			// -- enable/disable Buttons if applicable
			if (pRow!=IboxStartRow) {
				NButton.setEnabled(0<IboxStartRow);
				// if (1==nhl.length || (2==nhl.length && !hasControl)) SButton.setEnabled(iob[1][0].length-1>IboxStartRow+2*NbChkbxs);
				if (1==nhl.length) SButton.setEnabled(iob[1][0].length-1>IboxStartRow+2*NbChkbxs);
				else SButton.setEnabled(iob[1][0].length-1>IboxStartRow+NbChkbxs);
				RepaintIgnoreBoxes("");
			}
		} else if (e.getSource() == SButton) {
			// -- one or more step(s) down
			int pRow=IboxStartRow;
			SetIgnore();
			boolean FileFound=false;
			// -- step down until in the new range to display is a file to select
			if (1==nhl.length) {
				do {
					IboxStartRow+=(2*NbChkbxs);
					for (int j=1; j<=2*NbChkbxs;  j++) if(j+IboxStartRow<iob[0][0].length) if (!iob[0][IboxStartColumn][j+IboxStartRow]) FileFound=true;
				} while (!FileFound && iob[0][0].length-1>IboxStartRow+2*NbChkbxs);
			} else {
				do {
					IboxStartRow+=NbChkbxs;
					for (int i=IboxStartColumn; i<IboxStartColumn+2; i++)  for (int j=1; j <=NbChkbxs; j++) {
						if(j+IboxStartRow<iob[0][i].length)  if (!iob[0][i][j+IboxStartRow]) FileFound=true;
					}
				} while (!FileFound && iob[1][0].length-1>IboxStartRow+NbChkbxs);
			}
			// -- enable/disable Buttons if applicable
			if (pRow!=IboxStartRow) {
				NButton.setEnabled(0<IboxStartRow);
				// if (1==nhl.length || (2==nhl.length && !hasControl)) SButton.setEnabled(iob[1][0].length-1>IboxStartRow+2*NbChkbxs);
				if (1==nhl.length) SButton.setEnabled(iob[1][0].length-1>IboxStartRow+2*NbChkbxs);
				else SButton.setEnabled(iob[1][0].length-1>IboxStartRow+NbChkbxs);
				RepaintIgnoreBoxes("");
			}
		} else if (e.getSource() == WButton) {
			// -- one step right
			int pCol=IboxStartColumn;
			SetIgnore();
			IboxStartColumn--;
			if (0>IboxStartColumn) IboxStartColumn=0;
			if (pCol!=IboxStartColumn) {
				// WButton.setEnabled((hasControl && 0<IboxStartColumn)||(1<IboxStartColumn));
				WButton.setEnabled(0<IboxStartColumn);
				EButton.setEnabled(iob[1].length>IboxStartColumn+2);
				if (1<nhl.length && hasControl && IboxStartColumn==0) {
					Font f=IgnoreBox[0][1].getFont();
					int fstyle=(f.isBold()?Font.BOLD:Font.PLAIN);
					fstyle+=Font.ITALIC;
					f  = new Font(f.getName(), fstyle, f.getSize());
					for (int j=1; j<IgnoreBox[0].length; j++) IgnoreBox[0][j].setFont(f);
				} 
				RepaintIgnoreBoxes("");
			}
		} else if (e.getSource() == EButton) {
			// -- one step left
			int pCol=IboxStartColumn;
			SetIgnore();
			IboxStartColumn++;
			if (pCol!=IboxStartColumn) {
				WButton.setEnabled(0<IboxStartColumn);
				EButton.setEnabled(iob[1].length>IboxStartColumn+2);
				if (1<nhl.length && hasControl && IboxStartColumn==1) {
					Font f=IgnoreBox[0][1].getFont();
					int fstyle=(f.isBold()?Font.BOLD:Font.PLAIN);
					f  = new Font(f.getName(), fstyle, f.getSize());
					for (int j=1; j<IgnoreBox[0].length; j++) IgnoreBox[0][j].setFont(f);
				}
				RepaintIgnoreBoxes("");
			}
		} else if (e.getSource() == helpButton0) {
			String [] Msgs = new String[50];
			int c=0;
			Msgs[c++]="Instructions:\n";
			Msgs[c++]=" \n";
			Msgs[c]="Input Folder - Select the directory of files, i.e. "; 
			int strLen = 0;
			for (int i=0; i<nhl.length-1; i++) if (strLen >= 0) {
				strLen += 5 + nhl[i].length();
				if (strLen < truncateFileNameColectionAt) Msgs[c]+="'"+nhl[i]+"#.tif'; ";
				else {
					Msgs[c]+="...; '";
					strLen = -1;
				}
			}
			Msgs[c++]+=nhl[nhl.length-1]+".tif', to read\n";
			Msgs[c++]="            wheras '#' should be a natural (integer) number > 0.\n";
			Msgs[c++]="            Change 'Preferences' for defining more or other file names\n";
			Msgs[c++]="            and to define which one of them serves as control.\n";
			if (NormCB.isEnabled()) Msgs[c++]="Subtract and normalize... - data normalization to control\n";
			Msgs[c++]="            (used i.e. for correction for phosphorescence effects in VSD signal data).\n";
			Msgs[c++]="Ignore list - select checkboxes to exclude single files from data averaging.\n";
			Msgs[c++]="            use arrow buttons at left and bottom to browse through selectable files.\n";
			Msgs[c++]="            Files names of control images used for normalization are drawn italic.\n";
			Msgs[c++]="Baseline frame range - set minimum and maximum frame number of the range\n";
			Msgs[c++]="            that should be considered as 'baseline' before stimulation.\n";
			Msgs[c++]="Response frame Range - set minimum and maximum frame number of the range\n";
			Msgs[c++]="            that should be considered as 'response' during and after stimulation.\n";
			Msgs[c++]="Filter in X,Y-direction - reduce noise by filtering in X,Y direction;\n";
			Msgs[c]="            if enabled, enter ";
			if (DoFilterXY) {
				if (!FilterMean) Msgs[c++]+="sigma (pixel).\n";
				else Msgs[c++]+="radius (pixel).\n";
			} else Msgs[c++]+="sigma (Gaussian filter) or radius (Mean).\n";
			Msgs[c++]="Filter in Z direction - reduce noise in output stacks by average frames across stack\n";
			Msgs[c++]="            if enabled, enter number of  frames (voxel) to average.\n";
			if (allowBlankArtifact) {
				Msgs[c++]="Interpolate artifact frames range - replace voxels in frames by interpolating\n";
				Msgs[c++]="            the voxels of the frames of the surrounding frames.\n";
				Msgs[c++]="            This optional feature is thought to be used to blank an intrinsic artifact.\n";
				Msgs[c++]="            If applicable, enter the range of frames to be replaced.\n";
			}
			Msgs[c++]="Button 'Output options' - Choose what this plugin should calculate and return.\n";
			Msgs[c++]="Button 'Filter settings' - Choose between 'Gaussian' and 'Mean' filter (if applicable),\n";
			Msgs[c++]="            enable/disable filtering in Z direction etc.\n";
			Msgs[c++]="Button 'Preferences' - Set operational settings for the plugin run.  This includes\n";
			Msgs[c++]="            layout options, preferences saving behavior, and the name mask of files\n";
			Msgs[c]="            to look up (e.g. ";
			strLen = 0;
			for (int i=0; i<nhl.length-1; i++) if (strLen >= 0) {
				strLen += 5 + nhl[i].length();
				if (strLen < truncateFileNameColectionAt) Msgs[c]+="'"+nhl[i]+".tif'; ";
				else {
					Msgs[c]+="...; '";
					strLen = -1;
				}
			}
			Msgs[c++]+=nhl[nhl.length-1]+".tif', current setting).\n";
			Msgs[c++]="Button 'Help' - that's the button you have pressed.\n";
			Msgs[c++]=" \n";
			Msgs[c++]="Download the newest version of this plugin and (possibly) more detailed instructions\n";
			Msgs[c++]="            at  http://tinyurl.com/vsdproc.\n";
			Msgs[c]="                                                       Plugin Version "+PluginVersion+":r"+src+"@IJ";
			try {Msgs[c++]+=IJ.getVersion()+"\n";}
			catch (NoSuchMethodError ex) {Msgs[c++]+="<1.32\n";}
			int maxlines=getMaxNbLines();
			int ni=1;
			do { 
				String Msg="";
				for (int ci =maxlines*(ni-1); ci<c && ci<maxlines*ni; ci++) Msg+=Msgs[ci];
				if (maxlines*ni < c) {
					Msg+="[...]\n";
					Msgs[maxlines*ni]="[...] "+Msgs[maxlines*ni];
				}
				new MessageDialog(new Frame(), PluginTitle, Msg);
			} while (maxlines*ni++ < c);
		} else if (e.getSource() == helpButton1) {
			String [] Msgs = new String[50];
			int c=0;
			Msgs[c++] = "OUTPUT OPTIONS,\n";
			Msgs[c++]="set action(s) to perform on each plugin run:\n";
			Msgs[c++]=" \n";
			Msgs[c++]="Show raw image - displays images (on screen) while data is read;\n";
			Msgs[c++]="            keeping these windows open.  Be careful with large numbers of files.\n";
			if (1<nhl.length&&hasControl) {
				Msgs[c++]="Show/Save ''"+nhl[0]+"'' average - displays/saves (to file) averaged data\n";
				Msgs[c++]="            for control images (e.g. recorded without stimulation), if applicable.\n";
			}
			Msgs[c++]="Show/Save average - displays/saves average data of trials except controls.\n";
			Msgs[c++]="Show/Save base (B) - displays/saves ''baseline'' projection of the average,\n";
			Msgs[c++]="            for frames recorded before stimulation.\n";
			Msgs[c++]="Show/Save response (R) - displays/saves ''response'' projection of the average,\n";
			Msgs[c++]="            for frames recorded during and after stimulation.\n";
			Msgs[c++]="Show/Save difference (R-B) - displays/saves the difference of ''response'' and ''baseline''.\n";
			Msgs[c++]="Show/Save % ratio  (R/B) - displays/saves the percent ratio of ''response'' and ''baseline'';\n";
			Msgs[c++]="            use this option as default, e.g. to generate maps of evoked cortical activity.\n";
			Msgs[c++]="Show/Save % ratio stack - displays/saves a stack of percent ratios ''response''/ ''baseline'';\n";
			Msgs[c++]="            in which each slice represents the response and baseline response for one trial.\n";
			Msgs[c++]="            This option is thought be used for identifying outliners.  It will yield one stack\n";
			Msgs[c++]="            for each file series containing one slice for each file processed per file series.\n";
			Msgs[c++]="            Use the ''common base[line]'' option to compare the response component only.\n";
			String fLabel=(1<nhl.length&&Normalizeb && hasControl)?"F":" I ";
			fLabel+="/"+fLabel;
			Msgs[c++]="Show/Save % "+DST+fLabel+SZS+" - displays/saves the percent normalized difference to ''baseline'',\n";
			Msgs[c++]="             divided by ''baseline''. Use this option to generate multi frame images (videos)\n";
			Msgs[c++]="             and to display temporal changes in activity during and after stimulation.\n";
			if (2<nhl.length || (!hasControl && 1<nhl.length)) {
				Msgs[c]="Show/Save Ratio";
				if (3<nhl.length || (!hasControl && 2<nhl.length)) Msgs[c]+="s";
				int c1=(hasControl && 1<nhl.length)?1:0;
				Msgs[c]+=" ''"+nhl[c1+1]+"'' / "+nhl[c1]+"'' - displays/saves the ratio";
				if (3<nhl.length || (!hasControl && 2<nhl.length)) Msgs[c]+="s";
				Msgs[c]+=" of responses of the results for ''";
				for (int i=c1+1; i<nhl.length-1; i++) Msgs[c]+=nhl[i]+"'' / ''"+nhl[1]+"'';  ''";
				Msgs[c]+= nhl[nhl.length-1]+"'' / ''"+nhl[1]+"'' files\n             (ratio";
				if (3<nhl.length || (!hasControl && 2<nhl.length)) Msgs[c]+="s";
				Msgs[c++]+=" of ratio of  ''response'' and ''baseline'').\n";
			}
			Msgs[c++]="Show/Save Deviation log - displays in a log window /saves to text file\n";
			Msgs[c++]="             for each file the average percent deviation from mean of all vicinal files.\n";
			Msgs[c++]="             This value should be high if processing outlining data\n";
			Msgs[c++]="             (i.e. due to artifacts or events of spontaneous activity etc).\n";
			Msgs[c++]="Show/Save log - displays in a log window /saves to a text file detailed information\n";
			Msgs[c++]="             on the settings chosen for this particular plugin run. If applicable, text files\n";
			Msgs[c++]="             will be saved together with results in the same folder used for data input.\n";
			Msgs[c++]="Calculation Mode - set how the plugin should proceed with multiple input files\n";
			Msgs[c++]="             By default is calculates the average.  Other options may run slower\n";
			Msgs[c++]="             and require more memory. (Be careful at 32-bit systems!)\n";
			Msgs[c++]="Button 'Help' - that's the button you have pressed.\n";
			Msgs[c++]=" \n";
			Msgs[c++]="Download the newest version of this plugin and (possibly) more detailed instructions\n";
			Msgs[c++]="             at  http://tinyurl.com/vsdproc.\n";
			Msgs[c]="                                                       Plugin Version "+PluginVersion+":r"+src+"@IJ";
			try {Msgs[c++]+=IJ.getVersion()+"\n";}
			catch (NoSuchMethodError ex) {Msgs[c++]+="<1.32\n";}
			int maxlines=getMaxNbLines();
			int ni=1;
			do { 
				String Msg="";
				for (int ci =maxlines*(ni-1); ci<c && ci<maxlines*ni; ci++) Msg+=Msgs[ci];
				if (maxlines*ni < c) {
					Msg+="[...]\n";
					Msgs[maxlines*ni]="[...] "+Msgs[maxlines*ni];
				}
				new MessageDialog(new Frame(), PluginTitle, Msg);
			} while (maxlines*ni++ < c);
		} else if (e.getSource() == helpButton2) {
			String [] Msgs = new String[50];
			int c=0;
			Msgs[c++] = "Instructions to the Filter Settings\n";
			Msgs[c++]=" \n";
			Msgs[c++]="Treatment of NaN and Infinity values - define how to mask values of pixels\n";
			Msgs[c++]="             that have invalid value (i.e. NaN or Infinity). This applies to 32-bit data only.\n";
			Msgs[c++]="             Default option is to preferably 'ignore' those voxels in a stack. Alternatively replace\n";
			Msgs[c++]="             these pixels in each frame by an average of surrounding pixels values.\n";
			Msgs[c++]="             Choose 'interpolate' in case input files of different trials have different amplitude.'\n";
			Msgs[c++]="Filter in X,Y-direction - enable/disable all filtering in X,Y direction;\n";
			Msgs[c]="             if applicable, enter ";
			if (DoFilterXY) {
				if (!FilterMean) Msgs[c++]+="sigma (in pixel).\n";
				else Msgs[c++]+="radius (in pixel).\n";
			} else Msgs[c++]+="sigma (for Gaussian filter) or radius (for Mean).\n";
			Msgs[c++]="Filter type - select  filter model.\n";
			Msgs[c++]="Filter averaged stacks - filter raw images before projection.\n";
			Msgs[c++]="             may be useful to reduce noise in difference-ratio stacks.\n";
			Msgs[c++]="Filter baseline / response - filter projections.\n";
			Msgs[c++]="             use this option to reduce noise in forms of single-frame results.\n";
			String fLabel=(0<counter[0][0]&&1<nhl.length&&Normalizeb && hasControl)?"F":" I ";
			fLabel+="/"+fLabel;
			Msgs[c++]="Filter before ''"+DST+fLabel+SZS+"'' calculation - additional filtering, to further\n";
			Msgs[c++]="             reduce noise in difference-ratio stacks.\n";
			Msgs[c++]="Filter in Z-direction - average frames by a running z-projection.\n";
			Msgs[c++]="             set number of frames before and after each frame (except edges).\n";
			Msgs[c++]="             This option may allow reducing noise,\n";
			Msgs[c++]="             i.e. in normalized ratio-of-difference stacks.\n";
			Msgs[c++]="Button 'Help' - that's the button you obviously have pressed.\n";
			Msgs[c++]=" \n";
			Msgs[c++]="Download the newest version of this plugin and (possibly) more detailed instructions\n";
			Msgs[c++]="             at  http://tinyurl.com/vsdproc.\n";
			Msgs[c]="                                                       Plugin Version "+PluginVersion+":r"+src+"@IJ";
			try {Msgs[c++]+=IJ.getVersion()+"\n";}
			catch (NoSuchMethodError ex) {Msgs[c++]+="<1.32\n";}
			int maxlines=getMaxNbLines();
			int ni=1;
			do { 
				String Msg="";
				for (int ci =maxlines*(ni-1); ci<c && ci<maxlines*ni; ci++) Msg+=Msgs[ci];
				if (maxlines*ni < c) {
					Msg+="[...]\n";
					Msgs[maxlines*ni]="[...] "+Msgs[maxlines*ni];
				}
				new MessageDialog(new Frame(), PluginTitle, Msg);
			} while (maxlines*ni++ < c);
		} else if (e.getSource() == helpButton3) {
			String [] Msgs = new String[50];
			int c=0;
			
			// -- first part
			Msgs[c++] = "Layout Preferences:\n";
			Msgs[c++]="settings for plugin layout, input file names and their process order:\n";
			Msgs[c++]=" \n";
			Msgs[c++]="Number of ''Ignore'' boxes - Set number of boxes that are displayed simultaneously\n";
			Msgs[c++]="             in the Ignore List (Main dialog) per file series.\n";
			Msgs[c++]="File name (s), ''#.tif'' - Preset names of the files series to be load by the plugin.\n";
			Msgs[c++]="             whereas the index '#' should be a natural (integer) number > 0.\n";
			Msgs[c++]="             Omit indices and file extension in this field.  You may separate file names.\n";
			Msgs[c++]="             by space or other typical list separators (comma, semicolon, etc.)\n"; 
			Msgs[c++]="             Default presetting is ''no, hi, lo''.\n";
			Msgs[c++]="Control files - Select among file names one that is to be used as control file series\n";
			Msgs[c++]="             if applicable. Data in this file series may be used to normalize the other data,\n";
			Msgs[c]="             i.e., typically for VSD recordings. ";
			if (hasControl) Msgs[c++]+="(i.e.''"+nhl[0]+"'').\n";
			else Msgs[c++]+="Default setting is ''no''.\n";
			Msgs[c++]="Allow artifact z-interpolating option - if enabled, the main dialog screen will be extended\n";
			Msgs[c++]="            with an additional option that allows to define which frames should be replaced\n";
			Msgs[c++]="            by voxels with value that is an interpolation of values in the frames before and\n";
			Msgs[c++]="            after the range of frames to be interpolated in stack.\n";
			int maxlines=getMaxNbLines();
			int ni=1;
			
			// -- divide message dialog here into 2 parts if screen height is not large enough to display both
			if (maxlines*ni < c + 23) {  //  -- number of lines in second dialog should be equal or smaller than 23
				do { 
					String Msg="";
					for (int ci =maxlines*(ni-1); ci<c && ci<maxlines*ni; ci++) Msg+=Msgs[ci];
					Msg+="[...]\n";
					Msgs[maxlines*ni]="[...] "+Msgs[maxlines*ni];
					new MessageDialog(new Frame(), PluginTitle, Msg);
				} while (maxlines*ni++ < c);
				Msgs = new String[50];
				c=0;
				ni=1;
			} else Msgs[c++]=" \n";
			
			// -- second part
			Msgs[c++] = "Operational Preferences:\n";
			Msgs[c++]=" \n";
			Msgs[c++]="Responsiveness - Preset how the plugin should response to errors and unexpected events\n";
			Msgs[c++]="             Choose 'verbose' for many messages and hints; 'silent' for an (almost) quiet run.'\n";
			Msgs[c++]="Settings (saving behavior) - By default, this plugin keeps settings for future runs.\n";
			Msgs[c++]="             In 'global' mode, settings are generally shared by all users of the computer.'\n";
			Msgs[c++]="             (Administrator or advanced user privileges may be required).'\n";
			Msgs[c++]="             In 'personal' mode, settings are stored for all ImageJ installations for the current user.'\n";
			Msgs[c++]="             In 'local' mode, settings are stored for the current ImageJ installation (for all users).'\n";
			Msgs[c++]="             Some limitations in this functionality may apply for Macintosh(TM) computers.'\n";
			Msgs[c++]="Apply (button) - Saves all current settings as preferences for next run.\n";
			Msgs[c++]="             These include also settings done in the main dialog of this plugin\n";
			Msgs[c++]="             (except if keeping settings is switched off) but do not apply to \n";
			Msgs[c++]="             any selection of individual files to be ignored at /excluded from loading.\n";
			if (2 < storedSettings.length()) {
				Msgs[c++]="Undo Reset (button) - Restore the saved settings that were cleared by ''Reset''.\n";
				Msgs[c++]="             Otherwise, they will be lost when closing the dialogs of this plugin.\n";
				Msgs[c++]="             (I.e. if you had pressed the ''Reset'' button mistakenly before.)\n";
			} else {
				Msgs[c++]="Reset (button) - Clear all saved settings and reset the plugin to default.\n";
				Msgs[c++]="             use this option only if required or to clean up an erroneous presetting.\n";
			}
			Msgs[c++]="Help (button) - That's the button you have pressed.\n \n";
			Msgs[c++]="Download the newest version of this plugin and (possibly) more detailed instructions\n";
			Msgs[c++]="             at  http://tinyurl.com/vsdproc.\n";
			Msgs[c]="                                                       Plugin Version "+PluginVersion+":r"+src+"@IJ";
			try {Msgs[c++]+=IJ.getVersion()+"\n";}
			catch (NoSuchMethodError ex) {Msgs[c++]+="<1.32\n";}
			do {
				String Msg="";
				for (int ci =maxlines*(ni-1); ci<c && ci<maxlines*ni; ci++) Msg+=Msgs[ci];
				if (maxlines*ni < c) {
					Msg+="[...]\n";
					Msgs[maxlines*ni]="[...] "+Msgs[maxlines*ni];
				}
				new MessageDialog(new Frame(), PluginTitle, Msg);
			} while (maxlines*ni++ < c);
		} else if (e.getSource() == ResetButton) {
			try {
				if (2 < storedSettings.length()) {
					String Msg="Do you want to UNDO the Reset?\n";
					Msg += "Previously saved settings will be retrieved and available at next run.";
					if (JOptionPane.YES_OPTION == JOptionPane.showConfirmDialog(new Frame(), Msg ,PluginTitle,JOptionPane.YES_NO_OPTION, JOptionPane.WARNING_MESSAGE)) {
						boolean SetSet = restoreSettings(Pth);
						if (SetSet) {
							storedSettings="";
							MessageDialog d = new MessageDialog(new Frame(), ShortPluginTitle,"Previously deleted prefernces are recovered and again saved.");
							ResetButton.setLabel("     Reset      ");
						} else IJ.error("Failed to undo the reset!");
					}
				} else {
					String Msg="Do you want to delete ALL saved settings and reset the plugin to default?\n";
					Msg +="(current changes won't be affected.)";
					if (JOptionPane.YES_OPTION == JOptionPane.showConfirmDialog(new Frame(), Msg ,PluginTitle,JOptionPane.YES_NO_OPTION, JOptionPane.WARNING_MESSAGE)) {
						boolean SetSet = deleteSettings(Pth);
						if (SetSet) {
							MessageDialog d = new MessageDialog(new Frame(), ShortPluginTitle,"Previously saved preferences are removed.\nYou may undo this or cancel the plugin run to finalize it.");
							ResetButton.setLabel(" Undo Reset ");
						} else IJ.error("Failed to reset!");
					} 
				}
			}
			catch (UnsupportedOperationException ex) {
				IJ.error(Canceled?"Undo of Reset failed!":"Reset failed!");
				errString+=(Canceled?"Undo of Reset failed":"Reset failed");
			}
		} else if (e.getSource() == ApplyButton) {
			String Msg="Do you want to keep all current settings for the next plugin run now?";
			if (JOptionPane.YES_OPTION == JOptionPane.showConfirmDialog(new Frame(), Msg ,PluginTitle,JOptionPane.YES_NO_OPTION)) {
				// -- get recent changes the user may have done in current dlg
				String [] currentNhl = new String [nhl.length];
				for (int i=0; i<nhl.length; i++) currentNhl[i]=nhl[i];
				nhl=s2array(NHLField.getText(), nhl, false);
				if (1 > nhl.length) {
					nhl = new String [currentNhl.length];
					for (int i=0; i<currentNhl.length; i++) nhl[i]=currentNhl[i];
				}
				
				boolean currentHasControl = hasControl;
				if (hasConBox.isEnabled()) hasControl=hasConBox.getState();
				
				int currentNbChkbxs = NbChkbxs;
				NbChkbxs = s2i (NCField.getText(), NbChkbxs);
				boolean currentAllowBlankArtifact = allowBlankArtifact;
				if (aBlankBox.isEnabled()) allowBlankArtifact = aBlankBox.getState();
				if (currentNbChkbxs!=NbChkbxs || currentAllowBlankArtifact != allowBlankArtifact) {
					if (2 > NbChkbxs) NbChkbxs = currentNbChkbxs;
					MaxNbChkbxs = getMaxNbChkbxs(allowBlankArtifact);
					if (MaxNbChkbxs < NbChkbxs) NbChkbxs = MaxNbChkbxs;
					NCField.setText(""+NbChkbxs);
				}
				
				verboseMode=verboseBox.getState();
				if (verboseMode) silentMode=false;
				else silentMode = silentBox.getState();
				
				int tstoreModus=storeModus;
				if (dontKeepBox.getState()) tstoreModus=DONT_KEEP;
				if (localKeepBox.getState()) tstoreModus=KEEP_LOCAL;
				if (personalKeepBox.getState()) tstoreModus=KEEP_PERSONAL;
				if (globalKeepBox.getState()) tstoreModus=KEEP_GLOBAL;
				if (customKeepBox.getState()) tstoreModus=KEEP_CUSTOM;
				if (tstoreModus != storeModus) Pth="";
				storeModus = tstoreModus;
				
				// -- get recent changes the user may have done in main dlg
				int currentBaselinefirst = baselinefirst;
				int currentBaselinelast = baselinelast;
				int currentResponsefirst = responsefirst;
				int currentResponselast = responselast;
				
				baselinefirst = s2i (BaselineFirstField.getText(), baselinefirst);
				baselinelast = s2i (BaselineLastField.getText(), baselinelast);
				responsefirst = s2i (ResponseFirstField.getText(), responsefirst);
				responselast = s2i (ResponseLastField.getText(), responselast);
				
				boolean currentDoFilterXY = DoFilterXY;
				double currentFilterXy = FilterXy;
				boolean currentDoZFilterAv = DoZFilterAv;
				int currentFilterZ = FilterZ;
				
				DoFilterXY = FilterXYBox.getState();

				if (DoFilterXY) {
					double FilterXyN = s2d (XYFilterField2.getText(), FilterXy);
					if (!FilterMean) {
						if (FilterXyN < 0.0 || Math.round(FilterXyN) != Math.round(FilterXy)) FilterXy = FilterXyN;
					} else {
						if (FilterXyN < 0.0) FilterXy = FilterXyN;
					}
				}
				DoZFilterAv = ZFilterAvBox.getState();
				if (DoZFilterAv) FilterZ = s2i (ZFilterField.getText(), FilterZ);
				
				boolean currentNormalizeb = Normalizeb;
				if (NormCB.isEnabled()) Normalizeb = NormCB.getState();
				
				boolean currentBlankArtifact = blankArtifact;
				if (allowBlankArtifact && currentAllowBlankArtifact)
						blankArtifact = doBlankBox.getState();
				
				String currentDirIn = DirIn;
				String [] currentPDir = new String [PDir.length];
				for (int tc=0;tc<currentPDir.length;tc++) currentPDir[tc]=PDir[tc];
							
				if (0!=DirIn.compareTo((String)inputFolderChoice.getSelectedItem())) {
					File directory = new File((String)inputFolderChoice.getSelectedItem());
					try {
						if (!directory.isDirectory()) {
							String tdir=DirIn;
							int pos=0;
							while (!directory.isDirectory() && minSubDirLevel<howOftenIn(BSL, tdir)) {
								tdir=tdir.substring(0,tdir.lastIndexOf(BSL));
								directory=new File(tdir);
							}
						}
						if (directory.isDirectory()) {
							DirIn=""+directory;
							int pi=inputFolderChoice.getSelectedIndex();
							if (0>pi) {
								String [] TDir = new String [PDir.length];
								for (int tc=0;tc<TDir.length;tc++) TDir[tc]=PDir[tc];
								PDir= new String [PDir.length+1];
								PDir[0]=DirIn;
								for (int tc=0;tc<TDir.length;tc++) PDir[tc+1]=TDir[tc];
							} else PDir[pi]=DirIn;
						}
					} 
					catch (SecurityException ex) {
						DirIn =  currentDirIn; 
						PDir = new String [currentPDir.length];
						for (int tc=0;tc<PDir.length;tc++) PDir[tc]=currentPDir[tc];
					}
					catch (NullPointerException ex) {
						DirIn =  currentDirIn; 
						PDir = new String [currentPDir.length];
						for (int tc=0;tc<PDir.length;tc++) PDir[tc]=currentPDir[tc];
					}
					if ((filesize != filesize0 || 0==filesize) && blankArtifact && allowBlankArtifact)
							blankArtifact=false;
				}
				if (allowBlankArtifact && blankArtifact) {
					blankFrom = s2i (blankFromField.getText(), blankFrom);
					blankTo = s2i (blankToField.getText(), blankTo);
				}
				
				// -- save settings as preferences for next plugin run
				if (DONT_KEEP==storeModus) {
					try {
						Prefs.set("VSDproc_store.int", (double)storeModus);
						Prefs.set("VSDproc_lastRun.long", ""+System.currentTimeMillis());
					} catch (NoSuchMethodError er) {}
				} else {
					gotSet=findPathAndDoSettings (Pth, SET, true);
					ResetButton.setEnabled(gotSet);
				}
				
				// -- reset values of current user dlg to previous settings (for the case the user is cancelling the dlg)
				NbChkbxs = currentNbChkbxs;
				allowBlankArtifact = currentAllowBlankArtifact;
				blankArtifact = currentBlankArtifact;
				nhl = new String [currentNhl.length];
				for (int i=0; i<currentNhl.length; i++) nhl[i]=currentNhl[i];
				hasControl = currentHasControl;
				
				// -- reset values of current main dlg to previous settings
				baselinefirst = currentBaselinefirst;
				baselinelast = currentBaselinelast;
				responsefirst = currentResponsefirst;
				responselast = currentResponselast ;
				DoFilterXY = currentDoFilterXY;
				FilterXy = currentFilterXy;
				DoZFilterAv = currentDoZFilterAv;
				FilterZ = currentFilterZ;
				Normalizeb = currentNormalizeb;
				DirIn =  currentDirIn; 
				PDir = new String [currentPDir.length];
				for (int tc=0;tc<PDir.length;tc++) PDir[tc]=currentPDir[tc];
			}
		}
	}
}

private void updateDlg2 () {											
	NCField.setText(""+NbChkbxs);
	String nhls=nhl[0];
	for (int i=1; i<nhl.length; i++) nhls+=(", "+nhl[i]);
	if (0!=nhls.compareTo(NHLField.getText())) {
		NHLField.setText(nhls);
		conSelCBox.removeAllItems();
		for (int i=0; i<nhl.length; i++)  conSelCBox.insertItemAt(nhl[i], i);
	}
	if (hasControl && 1<nhl.length) {
		hasConBox.setState(true);
		hasConBox.setEnabled(1<nhl.length);
		conSelCBox.setEnabled(true);
		conSelCBox.setSelectedIndex(0);
	} else {
		hasConBox.setState(false);
		hasConBox.setEnabled(1<nhl.length);
		conSelCBox.setSelectedIndex(-1);
		conSelCBox.setEnabled(false);
	}
	if (2 == getMaxNbChkbxs(true)) {
		allowBlankArtifact = false;   // no more elements
		aBlankBox.setEnabled(false);
		aBlankBox.setState(false);
	} else aBlankBox.setState(allowBlankArtifact);
	verboseBox.setState(verboseMode);
	normmBox.setState(!silentMode&&!verboseMode);
	silentBox.setState(silentMode);
	ignoreNBox.setState(!interpolateXY);
	interpolateBox.setState(interpolateXY);
}


private void RepaintIgnoreBoxes(String spc) {
	if (null != spc) {
		try {
			for(int i=0;i<2;i++) for(int j=1;j<=NbChkbxs;j++) {
				if (1==nhl.length) IgnoreBox[i][j].setText(nhl[IboxStartColumn]+(j+IboxStartRow+i*NbChkbxs)+".tif "+spc);
				else IgnoreBox[i][j].setText(nhl[IboxStartColumn+i]+(j+IboxStartRow)+".tif "+spc);
			}
		} catch (NoSuchMethodError ex) { 
			for(int i=0;i<2;i++) for(int j=1;j<=NbChkbxs;j++) {
				if (1==nhl.length) IgnoreBox[i][j].setLabel(nhl[IboxStartColumn]+(j+IboxStartRow+i*NbChkbxs)+".tif "+spc);  // -----  NO WORRIES, THIS IS SOLELY FOR ...
				else IgnoreBox[i][j].setLabel(nhl[IboxStartColumn+i]+(j+IboxStartRow)+".tif "+spc);   //   ... DOWNCOMPATIBILITY TO LOWER IJ VERSIONS.  YOU CAN IGNORE THESE TWO COMPILING WARNINGS.  -------
			}
		}
	}
	
	if (1==nhl.length) {
		for(int i=0;i<2;i++) for(int j=1;j<=NbChkbxs;j++) {
			if (null!=iob) {
				if (iob[1][0].length>j+IboxStartRow+i*NbChkbxs) {
					IgnoreBox[i][j].setEnabled(!iob[0][0][j+IboxStartRow+i*NbChkbxs]);
					IgnoreBox[i][j].setSelected(iob[1][0][j+IboxStartRow+i*NbChkbxs]);
				} else {
					IgnoreBox[i][j].setSelected(true);
					IgnoreBox[i][j].setEnabled(false);
				}
			} else {
				IgnoreBox[i][j].setSelected(true);
				IgnoreBox[i][j].setEnabled(false);
			}
		}
	} else {
		for(int i=0;i<2;i++) for(int j=1;j<=NbChkbxs;j++) {
			if (null!=iob) {
				if (iob[1].length>IboxStartColumn+i) {
					if (iob[1][i].length>j+IboxStartRow) {
						if (0 == i && 1<nhl.length && hasControl && !Normalizeb && 0 == IboxStartColumn) IgnoreBox[0][j].setEnabled(false);
						else IgnoreBox[i][j].setEnabled(!iob[0][i+IboxStartColumn][j+IboxStartRow]);
						IgnoreBox[i][j].setSelected(iob[1][i+IboxStartColumn][j+IboxStartRow]);
					} else {
						IgnoreBox[i][j].setSelected(true);
						IgnoreBox[i][j].setEnabled(false);
					}
				} else {
					IgnoreBox[i][j].setSelected(true);
					IgnoreBox[i][j].setEnabled(false);
				}
			} else {
				IgnoreBox[i][j].setSelected(true);
				IgnoreBox[i][j].setEnabled(false);
			}
		}
	}
}

private class TextFieldHandler implements ActionListener {
	public void actionPerformed( ActionEvent e ) {
		File directory;
		if( e.getSource() == inputFolderChoice) {
			String sel=""+inputFolderChoice.getSelectedItem();
			if (0!=DirIn.compareTo(sel)) {
				DirIn0=DirIn;
				String tdir=sel.trim();
				while (tdir.endsWith("*") || tdir.endsWith(".")) tdir=tdir.substring(0, tdir.length()-1);
				int pi=inputFolderChoice.getSelectedIndex();
				try {
					directory =new File(tdir);
					if (directory.isFile()) {
						directory = new File(directory.getParent());
						if (null==directory) directory=new File(tdir);
					}
					while (!directory.isDirectory() && minSubDirLevel<howOftenIn(BSL, tdir)) {
						tdir=tdir.substring(0, tdir.lastIndexOf(BSL));
						directory=new File(tdir);
					}
					if (directory.isDirectory()) {
						tdir=""+directory;
						DirIn=tdir;
						IboxStartRow=0;
						RepaintIgnoreBoxes("");
						SetCheck(!silentMode);
					} else {
						if (verboseMode) IJ.error("Folder ''"+ sel + "'' not found!");
						pi = -2;
					}
				}
				catch (SecurityException ex) {
					if (verboseMode) IJ.error("Could not access Folder ''"+ tdir + "'' !");
					pi = -2;
				}
				catch (NullPointerException ex) {
					if (verboseMode) IJ.error("Could not access Folder!");
					pi = -2;
				}
				if (0 <= pi) {
					inputFolderChoice.removeItemAt(pi);
					// --  select item if already in collection, otherwise add new item
					int si=-1;
					for (int i=0; i<inputFolderChoice.getItemCount(); i++) 	if (0==DirIn.compareTo(inputFolderChoice.getItemAt(i).toString())) si=i;
					if (si<0) {
						inputFolderChoice.insertItemAt(DirIn, pi);
						inputFolderChoice.setSelectedIndex(pi);
					} else inputFolderChoice.setSelectedIndex(si);
				} else if (-2 == pi) {
					pi = 0;
					for (int i=0; i<inputFolderChoice.getItemCount(); i++) 	if (0==DirIn.compareTo(inputFolderChoice.getItemAt(i).toString())) pi=i;
					inputFolderChoice.setSelectedIndex(pi);
				} else inputFolderChoice.insertItemAt(DirIn, 0);
				inputFolderChoice.repaint();
				IJ.showStatus("Select files in " + DirIn);
				PDir= new String [inputFolderChoice.getItemCount()];
				for (int i=0;i<PDir.length;i++) PDir[i]=inputFolderChoice.getItemAt(i).toString();
				if ((filesize != filesize0 || 0==filesize) && blankArtifact && allowBlankArtifact) {
					blankArtifact=false;
					doBlankBox.setState(false);
					blankFromField.setText ("");
					blankToField.setText ("");
					blankFromField.setEnabled(false);
					blankToField.setEnabled(false);
				}
			}
		} else if( e.getSource() == conSelCBox ){
			int ci=conSelCBox.getSelectedIndex();
			if (0 < ci) {
				String [] nhl2 = s2array(NHLField.getText(), nhl, !silentMode);
				if (null != nhl2) if (0 < nhl2.length) {
					String [] nhlC = new String [nhl2.length];
					for (int i=0;i<nhlC.length;i++) nhlC[i]=nhl2[i];
					nhl2[0]=nhlC[ci];
					for (int i=0;i<ci;i++) nhl2[i+1]=nhlC[i];
					for (int i=ci+1;i<nhl2.length;i++) nhl2[i]=nhlC[i];
					conSelCBox.removeItemAt(ci);
					conSelCBox.insertItemAt(nhl2[0], 0);
					conSelCBox.setSelectedIndex(0);
					String nhls =nhl2[0];
					for (int i=1; i<nhl2.length; i++) nhls+=(", "+nhl2[i]);
					NHLField.setText(nhls);
				}
			}
		} else if( e.getSource() == BaselineFirstField ){
			baselinefirst = s2i (BaselineFirstField.getText(), baselinefirst);
			if (baselinefirst < 1) baselinefirst = 1;
			if (baselinefirst > baselinelast) baselinefirst = baselinelast;
			if (0 < framecnt && baselinefirst>framecnt-1) baselinefirst=framecnt-1;
			BaselineFirstField.setText(""+baselinefirst);
		} else if( e.getSource() == BaselineLastField ){
			baselinelast = s2i (BaselineLastField.getText(), baselinelast);
			if (baselinelast < baselinefirst + 1) baselinelast = baselinefirst + 1;
			if (baselinelast > responsefirst - 1) baselinelast = responsefirst - 1;
			if (0<framecnt && baselinelast>framecnt-1) {
				baselinelast=framecnt-1;
				if (baselinefirst > baselinelast) baselinefirst = baselinelast;
				BaselineFirstField.setText(""+baselinefirst);
			}
			BaselineLastField.setText(""+baselinelast);
			if (blankFrom >= baselinelast ) {
				blankArtifact=false;
				doBlankBox.setState(false);
				blankFromField.setText ("");
				blankToField.setText ("");
				blankFromField.setEnabled(false);
				blankToField.setEnabled(false);
			}
		} else if( e.getSource() == ResponseFirstField ){
			responsefirst = s2i (ResponseFirstField.getText(), responsefirst);
			if (responsefirst > responselast) responsefirst = responselast;
			if (responsefirst < baselinelast +1) responsefirst = baselinelast + 1;
			if (0<framecnt &&  responsefirst>framecnt) {
				responsefirst=framecnt;
				if (responsefirst > baselinelast+1) baselinelast = responsefirst - 1;
				if (baselinefirst < baselinelast) baselinefirst = baselinelast;
				BaselineLastField.setText(""+baselinelast);
				BaselineFirstField.setText(""+baselinefirst);
			}
			ResponseFirstField.setText(""+responsefirst);
			if (blankTo <= responsefirst) {
				blankArtifact=false;
				doBlankBox.setState(false);
				blankFromField.setText ("");
				blankToField.setText ("");
				blankFromField.setEnabled(false);
				blankToField.setEnabled(false);
			}
		} else if( e.getSource() == ResponseLastField ){
			responselast = s2i (ResponseLastField.getText(), responselast);
			if (responsefirst > responselast) responselast = responsefirst;
			if (0<framecnt && responselast>framecnt) {
				responselast=framecnt;
				if (responsefirst > responselast) responsefirst = responselast;
				if (responsefirst < baselinelast+1) baselinelast = responsefirst - 1;
				if (baselinefirst > baselinelast) baselinefirst = baselinelast;
				ResponseFirstField.setText(""+responsefirst);
				BaselineLastField.setText(""+baselinelast);
				BaselineFirstField.setText(""+baselinefirst);
			}
			ResponseLastField.setText(""+responselast);
		} else if( e.getSource() == blankFromField){
			blankFrom = s2i (blankFromField.getText(), blankFrom);
			if (1 > blankFrom) blankFrom = 1;
			if (blankFrom > blankTo) blankFrom = blankTo;
			if (blankFrom <= baselinelast) blankFrom = baselinelast +1;
			blankFromField.setText(""+blankFrom);
		} else if( e.getSource() == blankToField){
			blankTo = s2i (blankToField.getText(), blankTo);
			if (blankFrom > blankTo) blankTo = blankFrom;
			if (blankTo >= responsefirst) blankTo = responsefirst - 1;
			blankToField.setText(""+blankTo);
		} else if( e.getSource() == XYFilterField ){
			double FilterXyN = s2d (XYFilterField.getText(), FilterXy);
			if (!FilterMean) {
				if (Math.round(FilterXyN) != Math.round(FilterXy)) FilterXy = FilterXyN;
			} else FilterXy = FilterXyN;
			if (FilterXy < 0.0) FilterXy = 0.0;
			if (!FilterMean) XYFilterField.setText(""+Math.round(FilterXy));
			else XYFilterField.setText(""+FilterXy);
		} else if( e.getSource() == ZFilterField ){
			FilterZ = s2i (ZFilterField.getText(), FilterZ);
			if (FilterZ < 0) FilterZ = 0;
			ZFilterField.setText(""+FilterZ);
		} else if( e.getSource() == XYFilterField2 ){
			double FilterXyN = s2d (XYFilterField.getText(), FilterXy);
			if (!FilterMean) {
				if (FilterXyN >= 0.0) XYFilterField2.setText(""+Math.round(FilterXyN));
			} else {
				if (FilterXyN >= 0.0) XYFilterField2.setText(""+FilterXyN);
			}
		} else if( e.getSource() == ZFilterField2 ){
			int FilterZn = s2i (ZFilterField2.getText(), FilterZ);
			if (FilterZn < 0) ZFilterField2.setText(""+FilterZ);
			else ZFilterField2.setText(""+FilterZn);
		} else if( e.getSource() == NCField ){
			int NextNbChkbxs = s2i (NCField.getText(), NbChkbxs);
			if (2 > NextNbChkbxs) NextNbChkbxs = NbChkbxs;
			if (aBlankBox.isEnabled()) allowBlankArtifact = aBlankBox.getState();
			MaxNbChkbxs =  getMaxNbChkbxs(allowBlankArtifact);
			if (MaxNbChkbxs < NextNbChkbxs) NextNbChkbxs = MaxNbChkbxs;
			NCField.setText(""+NextNbChkbxs);
			if (aBlankBox.isEnabled() && 2 == MaxNbChkbxs) {
				aBlankBox.setEnabled(false);
				aBlankBox.setState(false);
			}
		} else if( e.getSource() == NHLField ){
			String nhls="";
			String [] nhl2 = null;
			// if (!hasControl && 1<nhl.length) {
			// 	nhl2 = new String [nhl.length-1];
			// 	for (int i=0; i<nhl2.length; i++) nhl2[i]=nhl[i+1];
			// 	nhl2=s2array(NHLField.getText(), nhl2, false);
			// } else 
			nhl2=s2array(NHLField.getText(), nhl, false);
			if (0<nhl2.length) {
				nhls=nhl2[0];
				for (int i=1; i<nhl2.length; i++) nhls+=(", "+nhl2[i]);
				hasConBox.setEnabled(1<nhl2.length);
				hasConBox.setState(hasControl && 1<nhl2.length);
				conSelCBox.removeAllItems();
				for (int i=nhl2.length-1; i>=0; i--) conSelCBox.insertItemAt(nhl2[i],0);
				conSelCBox.setEnabled(hasControl && 1<nhl2.length);
				if (hasControl && 1<nhl2.length) conSelCBox.setSelectedIndex(0);
				else conSelCBox.setSelectedIndex(-1);
			} else {
				nhls=nhl[0];
				for (int i=1; i<nhl.length; i++) nhls+=(", "+nhl[i]);
			}
			NHLField.setText(nhls);
		}
	}
}

private boolean showOperationDialog() {
	IJ.showStatus(PluginTitle+": select actions to do");
	boolean tShowRaw= ShowRaw;
	boolean tShowBase=  ShowBase;
	boolean tSaveBase=  SaveBase;
	boolean tShowResp=  ShowResp;
	boolean tSaveResp=  SaveResp;
	boolean tShowDiff=  ShowDiff;
	boolean tSaveDiff=  SaveDiff;
	boolean tShowPerc=  ShowPerc;
	boolean tSavePerc=  SavePerc;
	boolean tShowPStack=  ShowPStack;
	boolean tSavePStack=  SavePStack;
	boolean commonBase0 = commonBase;
	// boolean tShowDiv=  ShowDiv;
	// boolean tSaveDiv=  SaveDiv;
	boolean tShowNo=  ShowNo;
	boolean tSaveNo=  SaveNo;
	boolean tShowAv=  ShowAv;
	boolean tSaveAv=  SaveAv;
	boolean tShowDelta= ShowDelta;
	boolean tSaveDelta= SaveDelta;
	boolean tShowRatio= ShowRatio;
	boolean tSaveRatio= SaveRatio;
	if (2>nhl.length || (hasControl && 3>nhl.length)) {
		tShowRatio=  false;
		tSaveRatio=  false;
	}
	boolean tShowQual=  ShowQual;
	boolean tSaveQual=  SaveQual;
	boolean tDoShowLog= DoShowLog;
	boolean tDoSaveLog= DoSaveLog;
	int tAMethod=AMethod;
	boolean setOk=false;
	Checkbox  ShowRatioBox=null, SaveRatioBox=null, ShowNoBox = null, SaveNoBox = null;
	do {
		GenericDialog gd4 = new GenericDialog(PluginTitle, IJ.getInstance());
		Panel ZeroPanel=new Panel();	
		ZeroPanel.add(new JLabel("Output Options"));
		Label titleLabel = new Label("           -  Action(s) to perform :");
		ZeroPanel.add(titleLabel);
		Panel OperationPanel=new Panel();	
		Checkbox  ShowRawBox = new Checkbox ("Show raw images", tShowRaw);
		String aLabel="--";
		switch (tAMethod) {
			case AVG_METHOD: aLabel="average"; break;
			case MEDIAN_METHOD: aLabel="median"; break;
			case SD_METHOD: aLabel="av. & stdev."; break;
			case SE_METHOD: aLabel="av. & sterr.";
		}
		ShowAvBox = new Checkbox ("Show "+aLabel, tShowAv);
		SaveAvBox = new Checkbox ("Save "+aLabel, tSaveAv);
		ShowNoBox = new Checkbox ("Show ''"+nhl[0]+"'' "+aLabel, tShowNo && (hasControl && 1< nhl.length));
		SaveNoBox = new Checkbox ("Save ''"+nhl[0]+"'' "+aLabel, tSaveNo && (hasControl && 1< nhl.length));
		Checkbox  ShowBaseBox = new Checkbox ("Show base(B)", tShowBase);
		Checkbox  SaveBaseBox = new Checkbox ("Save base (B)", tSaveBase);
		Checkbox  ShowRespBox = new Checkbox ("Show response (R)", tShowResp);
		Checkbox  SaveRespBox = new Checkbox ("Save response (R)", tSaveResp);
		Checkbox  ShowDiffBox = new Checkbox ("Show difference R-B", tShowDiff);
		Checkbox  SaveDiffBox = new Checkbox ("Save difference R-B", tSaveDiff);
		// Checkbox  ShowDivBox = new Checkbox ("Show ratio R/B", tShowDiv);
		// Checkbox  SaveDivBox = new Checkbox ("Save ratio R/B", tSaveDiv);
		Checkbox  ShowPercBox = new Checkbox ("Show % ratio R/B", tShowPerc);
		Checkbox  SavePercBox = new Checkbox ("Save % ratio R/B", tSavePerc);
		ShowPStackBox = new Checkbox ("Show % ratio Stack", tShowPStack);
		SavePStackBox = new Checkbox ("Save % ratio Stack", tSavePStack);
		String fLabel="F";
		if (null!=counter) if (0==counter[0][0] || 2>nhl.length || !Normalizeb || !hasControl) fLabel=" I ";
		fLabel+= "/"+fLabel;
		Checkbox  ShowDeltaBox = new Checkbox ("Show % "+DST+fLabel+SZS, tShowDelta);
		Checkbox  SaveDeltaBox = new Checkbox ("Save % "+DST+fLabel+SZS, tSaveDelta);
		fLabel="ratio";
		if (3<nhl.length || (!hasControl && 2<nhl.length)) fLabel+="s";
		int c1=(hasControl && 1<nhl.length)?1:0;
		if (2<nhl.length || (!hasControl && 1<nhl.length)) 
				if (c1<nhl.length) fLabel+=" ''"+nhl[c1+1]+"'' / ''"+nhl[c1]+"''";
		if (3<nhl.length || (!hasControl && 2<nhl.length)) fLabel+=",...";
		ShowRatioBox = new Checkbox ("Show "+fLabel, tShowRatio && (2<nhl.length || (!hasControl && 1<nhl.length)) );
		SaveRatioBox = new Checkbox ("Save "+fLabel, tSaveRatio && (2<nhl.length || (!hasControl && 1<nhl.length)) );
		Checkbox  ShowQualBox = new Checkbox ("Show Deviation", tShowQual);
		Checkbox  SaveQualBox = new Checkbox ("Save Deviation log", tSaveQual);
		Checkbox  ShowLogBox = new Checkbox ("Show Log ", tDoShowLog);
		Checkbox  SaveLogBox = new Checkbox ("Save Log to file", tDoSaveLog);
		cBaseBox = new Checkbox ("  to common base", commonBase && (ShowPStack || SavePStack));
		cBaseBox.setEnabled(ShowPStack || SavePStack);
		
		Panel ShowPanel=new Panel();
		ShowPanel.setLayout(new GridLayout(0,1));
		ShowPanel.add(ShowRawBox);
		ShowPanel.add(ShowNoBox);
		ShowPanel.add(ShowAvBox);
		ShowPanel.add(ShowBaseBox);
		ShowPanel.add(ShowRespBox);
		ShowPanel.add(ShowDiffBox);
		// ShowPanel.add(ShowDivBox);
		ShowPanel.add(ShowPercBox);
		ShowPanel.add(ShowPStackBox);
		ShowPanel.add(new Label(""));
		ShowPanel.add(ShowDeltaBox);
		ShowPanel.add(ShowRatioBox);
		ShowPanel.add(ShowQualBox);
		ShowPanel.add(ShowLogBox);
		Panel SavePanel=new Panel();
		SavePanel.setLayout(new GridLayout(0,1, 30,0));
		SavePanel.add(new Label(""), BorderLayout.NORTH);
		SavePanel.add(SaveNoBox);
		SavePanel.add(SaveAvBox);
		SavePanel.add(SaveBaseBox);
		SavePanel.add(SaveRespBox);
		SavePanel.add(SaveDiffBox);
		// SavePanel.add(SaveDivBox);  // you may reactivate this option if it makes sense to you
		SavePanel.add(SavePercBox);
		SavePanel.add(SavePStackBox);
		SavePanel.add(cBaseBox);
		SavePanel.add(SaveDeltaBox);
		SavePanel.add(SaveRatioBox);
		SavePanel.add(SaveQualBox);
		
		ItemListener ct = new ItemListener() {
			public void itemStateChanged(ItemEvent e) {
				if (cBaseBox.isEnabled()) commonBase = cBaseBox.getState();
				boolean p = ShowPStackBox.getState() || SavePStackBox.getState();
				cBaseBox.setEnabled(p);
				cBaseBox.setState(p && commonBase);
			}
		};
		ShowPStackBox.addItemListener(ct);
		SavePStackBox.addItemListener(ct);
		
		ShowNoBox.setEnabled(hasControl && 1< nhl.length);
		SaveNoBox.setEnabled(hasControl && 1< nhl.length);
		ShowRatioBox.setEnabled(2<nhl.length || (!hasControl && 1<nhl.length));
		SaveRatioBox.setEnabled(2<nhl.length || (!hasControl && 1<nhl.length));
		
		SavePanel.add(SaveLogBox);
		OperationPanel.add(ShowPanel,BorderLayout.WEST);
		OperationPanel.add(SavePanel,BorderLayout.EAST);
		try {gd4.addPanel(ZeroPanel);}
		catch (NoSuchMethodError e) {
			String vMsg = "This plugin requires a newer version of ImageJ.\n";
			vMsg+="Please update and run again.";
			IJ.error(vMsg);
			return false;
		}
		gd4.addPanel(OperationPanel);
		
		Panel ModusPanel =new Panel();
		CheckboxGroup cbg = new CheckboxGroup();
		AverageBox = new Checkbox("Average   ",cbg,AVG_METHOD==tAMethod);
		MedianBox = new Checkbox("Median   ",cbg,MEDIAN_METHOD==tAMethod);
		StdevBox = new Checkbox("Av. & Stdev.",cbg,SD_METHOD==tAMethod);
		SterrBox = new Checkbox("Av. & Sterr.",cbg,SE_METHOD==tAMethod);
		ItemListener it = new ItemListener() {
			public void itemStateChanged(ItemEvent e) {
				String sLabel="average";
				if (!silentMode) if (MedianBox.getState()) {
					sLabel="median";
					String ms="Be careful with using MEDIAN calculation.\n";
					ms+="Results may be used for overview purpose\n";
					ms+="but not for scientific data analysis!\n";
					ms+="Remember rules of good scientific practice.\n";
					ms+="                                                the author\n";
					new MessageDialog(new Frame(), PluginTitle, ms);
				}
				else if (StdevBox.getState()) sLabel="av. & stdev.";
				else if (SterrBox.getState()) sLabel="av. & sterr.";
				ShowAvBox.setLabel ("Show "+sLabel);
				SaveAvBox.setLabel ("Save "+sLabel);
			}
		};
		AverageBox.addItemListener(it);
		MedianBox.addItemListener(it);
		StdevBox.addItemListener(it);
		SterrBox.addItemListener(it);
		ModusPanel.setLayout(new GridLayout(2,3));
   		ModusPanel.add(new Label("Calculate:"));
		ModusPanel.add(AverageBox);
		ModusPanel.add(MedianBox);
		ModusPanel.add(new Label(" "));
   		ModusPanel.add(StdevBox);
   		ModusPanel.add(SterrBox);
   		// ModusPanel.add(new Label(" "));
		gd4.addPanel(ModusPanel);
		
		Panel HPanel =new Panel();
		HPanel.add(new Label("                                                                     "));
		helpButton1 = new Button ("Help");
		HPanel.add(helpButton1);
		ButtonHandler handler = new ButtonHandler();
		helpButton1.addActionListener( handler );
		try {gd4.addPanel(HPanel);}
		catch (NoSuchMethodError e) {
			// gd4.add(HPanel);
		}
		
		gd4.showDialog();
		checkboxHeight = (int) ShowRawBox.getSize().getHeight()+1;
		try {
			Prefs.set("checkboxHeight.int", (double) checkboxHeight);
		} catch (NoSuchMethodError e) {}
		
		if (gd4.wasCanceled()) {
			IJ.showStatus("Options change canceled.");
			commonBase = commonBase0;
			return false;
		} else IJ.showStatus("Options changed and settings to be updated.");
		tShowQual=ShowQualBox.getState();
		tSaveQual=SaveQualBox.getState();
		tDoShowLog=ShowLogBox.getState();
		tDoSaveLog=SaveLogBox.getState();
		tShowRaw=ShowRawBox.getState();
		if (ShowNoBox.isEnabled()) tShowNo=ShowNoBox.getState();
		if (SaveNoBox.isEnabled()) tSaveNo=SaveNoBox.getState();
		tShowAv=ShowAvBox.getState();
		tSaveAv=SaveAvBox.getState();
		tShowBase=ShowBaseBox.getState();
		tSaveBase=SaveBaseBox.getState();
		tShowResp=ShowRespBox.getState();
		tSaveResp=SaveRespBox.getState();
		tShowDiff=ShowDiffBox.getState();
		tSaveDiff=SaveDiffBox.getState();
		// tShowDiv=ShowDivBox.getState();
		// tSaveDiv=SaveDivBox.getState();
		tShowPerc=ShowPercBox.getState();
		tSavePerc=SavePercBox.getState();
		tShowPStack=ShowPStackBox.getState();
		tSavePStack=SavePStackBox.getState();
		if (cBaseBox.isEnabled()) commonBase = cBaseBox.getState();
		else commonBase = commonBase0;
		tShowDelta=ShowDeltaBox.getState();
		tSaveDelta=SaveDeltaBox.getState();
		if (ShowRatioBox.isEnabled()) tShowRatio=ShowRatioBox.getState();
		if (SaveRatioBox.isEnabled()) tSaveRatio=SaveRatioBox.getState();
		//  if (null!=counter) if (0<counter[0][0]&&1<nhl.length) {
		//	NormCB.setEnabled(true);
		//	NormCB.setState(Normalizeb);
		//  }
		if (tShowAv||tSaveAv||tShowDelta||tSaveDelta||tShowPerc||tSavePerc||tShowPStack||tSavePStack) setOk=true;
		else if ((tShowNo||tSaveNo) && hasControl && 1<nhl.length) setOk=true;
		else if ((tShowRatio||tSaveRatio) && SaveRatioBox.isEnabled()) setOk=true;
		//  else if (tShowDiv||tSaveDiv||tShowRaw||tShowBase||tSaveBase||tShowResp||tSaveResp||tShowDiff||tSaveDiff) setOk=true;
		else if (tShowRaw||tShowBase||tSaveBase||tShowResp||tSaveResp||tShowDiff||tSaveDiff) setOk=true;
		else if (tShowQual||tSaveQual) {
			tShowAv=true;
			tShowNo=hasControl && 1<nhl.length;
			if (verboseMode) IJ.error("Selection changed, please reconfirm!");
		} else {
			tShowRaw=ShowRaw;
			tShowNo=ShowNo;
			tSaveNo=SaveNo;
			tShowAv=ShowAv;
			tSaveAv=SaveAv;
			tShowBase=ShowBase;
			tSaveBase=SaveBase;
			tShowResp=ShowResp;
			tSaveResp=SaveResp;
			tShowDiff=ShowDiff;
			tSaveDiff=SaveDiff;
			// tShowDiv=ShowDiv;
			// tSaveDiv=SaveDiv;
			tShowPerc=ShowPerc;
			tSavePerc=SavePerc;
			tShowPStack=ShowPStack;
			tSavePStack=SavePStack;
			tShowDelta=ShowDelta;
			tSaveDelta=SaveDelta;
			tShowRatio=ShowRatio;
			tSaveRatio=SaveRatio;
			if (!silentMode) IJ.error("Select at least one operation!");
		}
		if (MedianBox.getState()) tAMethod=MEDIAN_METHOD;
		else if (StdevBox.getState()) tAMethod=SD_METHOD;
		else if (SterrBox.getState()) tAMethod=SE_METHOD;
		else tAMethod=AVG_METHOD;
	} while (!setOk);
	IJ.showStatus("Options changed and/or settings updated.");
	AMethod=tAMethod;
	ShowRaw=tShowRaw;
	if (ShowNoBox.isEnabled()) ShowNo=tShowNo;
	if (SaveNoBox.isEnabled()) SaveNo=tSaveNo;
	ShowAv=tShowAv;
	SaveAv=tSaveAv;
	ShowBase=tShowBase;
	SaveBase=tSaveBase;
	ShowResp=tShowResp;
	SaveResp=tSaveResp;
	ShowDiff=tShowDiff;
	SaveDiff=tSaveDiff;
	// ShowDiv=tShowDiv;
	// SaveDiv=tSaveDiv;
	ShowPerc=tShowPerc;
	SavePerc=tSavePerc;
	ShowPStack=tShowPStack;
	SavePStack=tSavePStack;
	ShowDelta=tShowDelta;
	SaveDelta=tSaveDelta;
	if (ShowRatioBox.isEnabled()) ShowRatio=tShowRatio;
	if (SaveRatioBox.isEnabled()) SaveRatio=tSaveRatio;
	DoShowLog=tDoShowLog;
	DoSaveLog=tDoSaveLog;
	ShowQual=tShowQual;
	SaveQual=tSaveQual;
	return true;
}

private void SetCheck(boolean verboseb) {
	// -- find minimal xj  (xj is the maximal index in all file names)
	int xj = 2>nhl.length ? 2*NbChkbxs+1 : NbChkbxs+1;
	
	// -- get input directory
	String [] sUc = null;   // --  array of file names (keeping upper case letters)
	String [] sLc = null;   // --  file name array  (all strings in lower case)
	int [][] fjs = null;   // --  file index array  (smart file  finding)  [i-index][j-index]
	try {
		File directory = new File(DirIn);
		if (!directory.isDirectory()) {
			String tdir=DirIn;
			while (!directory.isDirectory() && minSubDirLevel<howOftenIn(BSL, tdir)) {
				tdir=tdir.substring(0,tdir.lastIndexOf(BSL));
				directory=new File(tdir);
			}
			if (!directory.isDirectory() && null !=PDir) { 
				for (int pc=0;pc<PDir.length;pc++)  if(!directory.isDirectory()) {
					directory=new File(PDir[pc]);
					if(directory.isDirectory()) DirIn = PDir[pc] ;
				}
			}
			if (!directory.isDirectory()) { 
				for (int pc=0;pc<PDir.length;pc++)  if(!directory.isDirectory()) {
					tdir=PDir[pc];
					while (!directory.isDirectory() && minSubDirLevel<howOftenIn(BSL, tdir)) {
						tdir=tdir.substring(0,tdir.lastIndexOf(BSL));
						directory=new File(tdir);
					}
					if(directory.isDirectory()) DirIn = PDir[pc];
				}
			}
			if (!directory.isDirectory()) { 
				tdir=DirIn;
				while (!directory.isDirectory() && minSubDirLevel<howOftenIn(BSL, tdir)) {
					tdir=tdir.substring(0,tdir.lastIndexOf(BSL));
					directory=new File(tdir);
				}
			}
			if (!directory.isDirectory() && null !=PDir) { 
				for (int pc=0;pc<PDir.length;pc++)  if(!directory.isDirectory()) {
					tdir=PDir[pc];
					while (!directory.isDirectory() && 0<howOftenIn(BSL, tdir)) {
						tdir=tdir.substring(0,tdir.lastIndexOf(BSL));
						directory=new File(tdir);
					}
					if(directory.isDirectory()) DirIn = PDir[pc];
				}
			}
			if (directory.isDirectory()) {
				DirIn=""+directory;
				int pi=-1;
				if (!IJ.isWindows()) {for (int pc=0;pc<PDir.length;pc++) if(0>pi) if (0==DirIn.compareTo(PDir[pc])) pi=pc;}
				if (0>pi) {for (int pc=0;pc<PDir.length;pc++) if(0>pi) if (0==DirIn.compareToIgnoreCase(PDir[pc])) pi=pc;}
				if (0>pi) {
					String [] TDir = new String [PDir.length];
					for (int tc=0;tc<TDir.length;tc++) TDir[tc]=PDir[tc];
					PDir= new String [PDir.length+1];
					PDir[0]=DirIn;
					for (int tc=0;tc<TDir.length;tc++) PDir[tc+1]=TDir[tc];
					inputFolderChoice.insertItemAt(DirIn, 0);
					inputFolderChoice.setSelectedIndex(0);
				} else inputFolderChoice.setSelectedIndex(pi);
				inputFolderChoice.repaint();
			} else DirIn="";
		}
		if (0<DirIn.length()) IJ.showStatus("Select files in " + DirIn);
		else IJ.showStatus("Select files...");
		
		// String [] sUc = null;   // --  array of file names (keeping upper case letters)
		if (directory.isDirectory())  sUc = directory.list();
		
		// String [] sLc = null;   // --  file name array  (all strings in lower case)
		// int [][] fjs = null;   // --  file index array  (smart file  finding)  [i-index][j-index]
		if (null!=sUc) {
			sLc = new String[sUc.length];
			fjs = new int[2][sUc.length];
			for (int k=0; k<sLc.length; k++) {
				fjs[0][k]= -1;
				sLc[k] = sUc[k].toLowerCase();
				// if (sLc[k].endsWith(".tif")) for (int i=0; i<nhl.length; i++) if (hasControl || 1==nhl.length || 0<i) if (sLc[k].startsWith(nhl[i].toLowerCase())) {
				if (sLc[k].endsWith(".tif")) for (int i=0; i<nhl.length; i++) if (sLc[k].startsWith(nhl[i].toLowerCase())) {
					fjs[1][k]= -1;  //  -- for the case the minimum index  j  is 0 (currently it is 1)  ########
					try {fjs[1][k]=Integer.parseInt(sLc[k].substring(nhl[i].length(),sLc[k].length()-4));}
					catch (NumberFormatException e) {}
					if (fjs[1][k]>=0) fjs[0][k] = i;
					if (fjs[1][k]>=xj) if (sLc[k].length() == (nhl[i] + fjs[1][k]).length() +4) xj=fjs[1][k]+1;
				}
			}
		}
	}
	catch (SecurityException ex) {if (verboseb) IJ.error("Could not check out the input folder(s)!"); return;}
	catch (NullPointerException ex) {if (verboseb) IJ.error("Could not check out the input folder!"); return;}
	
	//  -- set  iob  values;  true if  file is missing or invalid
	iob=new boolean[2][nhl.length][xj];
	counter=new int [4][nhl.length];
	long [][] fileSizes=new long[nhl.length][xj];
	long [] fsz = new long [1];
	int [] fct = new int [1];
	// make histogram of file sizes; get the predominant size
	if (null != fjs) for (int i=0;i<nhl.length;i++) {
	// if (null != fjs) for (int i=0;i<nhl.length;i++) if (hasControl || 1==nhl.length || 0<i) {
		//  for (int k=0;k<3;k++) counter[k][i]=0;
		for (int j=1;j<xj;j++) {
			for (int k=0;k<fjs[1].length ;k++)   if (0== fileSizes[i][j] && fjs[0][k] == i && fjs[1][k] == j) {
				if (0==sLc[k].compareTo(nhl[i].toLowerCase()+j+".tif")) {   // making sure the file name does match the mask exactly
					try {
						File F= new File (DirIn+BSL+sUc[k]);
						if (F.isFile()) fileSizes[i][j] = F.length();
					}
					catch (SecurityException ex) {}
					catch (NullPointerException ex) {}
				}
			}
			iob[0][i][j] = 0== fileSizes[i][j];   //  files with size 0 Byte are fully ignored ####
			if(!iob[0][i][j]) {
				counter[0][i]++;
				boolean ffd=false;
				int le=fsz.length;
				for (int f=1;f<le;f++) {
					if (fsz[f]==fileSizes[i][j] && !ffd) {
						ffd=true;
						fct[f]++;
					}
				}
				if (!ffd) {
					long [] tl = new long[le];
					int [] ti = new int[le];
					for (int f=1;f<le;f++) {
						tl[f]=fsz[f];
						ti[f]=fct[f];
					}
					le++;
					fsz=new long[le];
					fct = new int [le];
					for (int f=1;f<le-1;f++) {
						fsz[f]=tl[f];
						fct[f]=ti[f];
					}
					fsz[le-1]= fileSizes[i][j];
					fct [le-1]=1;
				}
			}
		}
	}
	
	int MaxC=0;
	int Maxf=-1;
	for(int f=1;f<fct.length;f++) if(fct[f]>MaxC) {
		Maxf=f;
		MaxC=fct[f];
	}
	
	if(Maxf<0) for(int f=0;f<fct.length-1;f++) if(fct[f]>MaxC) Maxf=f;
	if(Maxf>0) {
		filesize=fsz[Maxf];
		for (int i=0;i<nhl.length;i++) for (int j=1;j<xj;j++) iob[1][i][j] = iob[0][i][j] || filesize != fileSizes[i][j]; 
	} else {
		filesize = 0;
		for (int i=0;i<nhl.length;i++) for (int j=1;j<xj;j++) iob[1][i][j] = true;
	}
	
	if (0==counter[0][0] || 1==nhl.length || !hasControl) {
		NormCB.setState(false);
		NormCB.setEnabled(false);
	} else {
		NormCB.setState(Normalizeb);
		NormCB.setEnabled(true);
	}
	if (0<filesize) {
		//  lookup filesize in presets
		int iidx=-1;
		for (int ni=0;ni<tps;ni++) if (filesize==filesizeA[ni]) iidx=ni;
		if (iidx>-1) {
			framecnt = framecntA[iidx];
			baselinefirst = baselinefirstA[iidx];
			baselinelast = baselinelastA[iidx];
			responsefirst = responsefirstA[iidx];
			responselast = responselastA[iidx];
			FilterZ = FilterZA[iidx];
			FilterXy = FilterXyA[iidx];
		} else if (0<counter[0][0]&&1<nhl.length && hasControl) {
			baselinefirst = baselinefirstVSD;
			baselinelast = baselinelastVSD;
			responsefirst = responsefirstVSD;
			responselast = responselastVSD;
			FilterXy = FilterXyVSD;
			FilterZ = FilterZVSD;
		} else {
			baselinefirst = baselinefirstIO;
			baselinelast = baselinelastIO;
			responsefirst = responsefirstIO;
			responselast = responselastIO;
			FilterXy = FilterXyIO;
			FilterZ = FilterZIO;
		}
	}
	
	if ((!DoFilterAv && !DoFilterBR && !DoFilterDelta) || 0 >= FilterXy) DoFilterXY = false;
	else DoFilterXY = DoFilterXY0;
	if (0 >= FilterZ) DoZFilterAv = false;
	else DoZFilterAv = DoZFilterAv0;

	// if (allowBlankArtifact && blankArtifact)
	//	if (2 > responsefirst - baselinelast) blankArtifact=false;

	try {
		BaselineFirstField.setText(""+baselinefirst);
		BaselineLastField.setText(""+baselinelast);
		ResponseFirstField.setText(""+responsefirst);
		ResponseLastField.setText(""+responselast);
		if (DoFilterXY) {
			if (!FilterMean) {
				XYFilterField.setText(""+FilterXy);
				SigmaLabel.setText("Filter in X,Y-direction,   sigma (pixel):");
			} else {
				XYFilterField.setText(""+Math.round(FilterXy));
				SigmaLabel.setText("Filter in X,Y-direction,  radius (pixel):");
			}
		} else SigmaLabel.setText("Filter in X,Y-direction                             ");
		ZFilterField.setText(""+FilterZ);
		ZFilterField.setEnabled(DoZFilterAv);
		if (DoZFilterAv) {
			FramesLabel.setText("Filter in Z-direction, average  voxel:");
			FilterZ = s2i (ZFilterField2.getText(), FilterZ);
		} else FramesLabel.setText("Filter in Z-direction                                 ");
		// if (allowBlankArtifact && blankArtifact) {
		//	doBlankBox.setState(blankArtifact);
		//	if (blankArtifact) {
		//		blankFromField.setText (""+blankFrom);
		//		blankToField.setText (""+blankTo);
		//	} else {
		//		blankFromField.setText ("");
		//		blankToField.setText ("");
		//	}
		//	blankFromField.setEnabled(blankArtifact);
		//	blankToField.setEnabled(blankArtifact);
		// }
	} catch (NullPointerException e) {}
	
	RepaintIgnoreBoxes(null);
	if (null!=iob) {
		IboxStartRow=0;
		boolean FileFound=true;
		if (1==nhl.length) {
			do {
				if (!FileFound)  IboxStartRow+=(2*NbChkbxs);
				FileFound=false;
				for (int j=1; j<=2*NbChkbxs;  j++) if(j+IboxStartRow<iob[0][0].length) if (!iob[0][IboxStartColumn][j+IboxStartRow]) FileFound=true;
			} while (!FileFound && iob[0][0].length-1>IboxStartRow+2*NbChkbxs);
		} else {
			do {
				if (!FileFound) IboxStartRow+=NbChkbxs;
				FileFound=false;
				for (int i=IboxStartColumn; i<IboxStartColumn+2; i++)  for (int j=1; j <=NbChkbxs; j++) {
					if(j+IboxStartRow<iob[0][i].length)  if (!iob[0][i][j+IboxStartRow]) FileFound=true;
				}
			} while (!FileFound && iob[1][0].length-1>IboxStartRow+NbChkbxs);
		}
		if (!FileFound) IboxStartRow=0;
		if (0<IboxStartRow) {
			String s="";
			for (int k=nhl[nhl.length-1].length();k<14;k++) s+=" ";   //  ################
			RepaintIgnoreBoxes(s);
		}
	}
	NButton.setEnabled(0<IboxStartRow);
	if (1==nhl.length) SButton.setEnabled(iob[1][0].length-1>IboxStartRow+2*NbChkbxs);
	else SButton.setEnabled(iob[1][0].length-1>IboxStartRow+NbChkbxs);
}

private boolean checkSettingsPath(String pth) {
	if (pth==null) return false;
	if (0==pth.length()) return false;
	File aFile = null;
	try {
		if (pth.endsWith(BSL)) aFile = new File(pth+IniFileName);
		else aFile = new File(pth+BSL+IniFileName);
		if (!aFile.isFile()) {
			if (!aFile.createNewFile()) return false;
		} else return aFile.canRead();
	}
	catch (IOException ex) {return false;}
	catch (SecurityException ex) {return false;}
	catch (NullPointerException ex) {return false;}
	
	try {
		aFile.delete();
		return !aFile.isFile();
	}
	catch (SecurityException ex) {}
	catch (NullPointerException ex) {}
	return false;
}

private boolean doSettings(String pth, int method) {
	if (pth==null) return false;
	if (0==pth.length()) return false;
	switch(method) {
		case GET:
			return getSettings(pth);
		case SET:
			return setSettings(pth);
		case CHECK:
			return checkSettingsPath(pth);
	}
	return false;
}

private boolean getSettings (String pth) {
	File aFile = null;
	try {
		if (pth.endsWith(BSL)) aFile = new File(pth+IniFileName);
		else aFile = new File(pth+BSL+IniFileName);
		if (!aFile.isFile()) return false;
		if (!aFile.canRead()) return false;
	}
	catch (SecurityException ex) {return false;}
	catch (NullPointerException ex) {return false;}
	if (null==aFile) return false;
	
	BufferedReader input = null;
	boolean Result=true;
	try {
		// use buffering, reading one line at a time
		input = new BufferedReader(new FileReader(""+aFile));
		String line = "";
		// readLine is a bit quirky :
		// - it returns the content of a line MINUS the newline.
		// - it returns null only for the END of the stream.
		// - it returns an empty String if two newlines appear in a row.
		while (( line = input.readLine()) != null){
			if(line.startsWith("Dir")) {
				if(line.startsWith("DirIn=")) {
					DirIn=GetString(line,DirIn);
					HDir[0] = DirIn;
				} else if(line.startsWith("DirWin=")) {
					winDir = GetString(line, winDir);
				} else if(line.startsWith("DirHomeDrive=")) {
					homeDrive = GetString(line, homeDrive);
				} else if(line.startsWith("DirPublic=")) {
					publicDir = GetString(line, publicDir);
				} else if(line.startsWith("DirProgr=")) {
					progrDir = GetString(line, progrDir);
				} else if(keepComInIni && 2<line.length()) {
					if (IJ.isWindows()) iniComments+=(char)13;
					iniComments+="\n";
					iniComments+=line;
				}
			} else if(line.startsWith("HDir")) {
				int hix=GetIdxInt(line,-1);
				if (hix>=0 && hix <HDir.length-1) HDir[hix+1] = GetString(line,HDir[hix+1]);
				else if(keepComInIni && 2<line.length()) {
					if (IJ.isWindows()) iniComments+=(char)13;
					iniComments+="\n";
					iniComments+=line;
				}
			} else if(line.startsWith("b")) {
				if(line.startsWith("baselinefirst")) {
					if(line.startsWith("baselinefirstIO=")) baselinefirstIO=GetInt(line,baselinefirstIO);
					else if(line.startsWith("baselinefirstVSD=")) baselinefirstVSD=GetInt(line,baselinefirstVSD);
					else baselinefirst=GetInt(line,baselinefirst);
				} else if(line.startsWith("baselinelast")) {
					if(line.startsWith("baselinelastIO=")) baselinelastIO=GetInt(line,baselinelastIO);
					else if(line.startsWith("baselinelastVSD=")) baselinelastVSD=GetInt(line,baselinelastVSD);
					else baselinelast=GetInt(line,baselinelast);
				} else if(line.startsWith("baselinelast")) {
					if(line.startsWith("blankArtifact=")) blankArtifact=GetBoolean(line,blankArtifact);
					if(line.startsWith("blankFrom=")) blankFrom=GetInt(line, blankFrom);
					if(line.startsWith("blankTo=")) blankTo=GetInt(line, blankTo);
				} else if(keepComInIni && 2<line.length()) {
					if (IJ.isWindows()) iniComments+=(char)13;
					iniComments+="\n";
					iniComments+=line;
				}
			} else if(line.startsWith("r")) {
				if(line.startsWith("responsefirst")) {
					if(line.startsWith("responsefirstIO=")) responsefirstIO=GetInt(line,responsefirstIO);
					else if(line.startsWith("responsefirstVSD=")) responsefirstVSD=GetInt(line,responsefirstVSD);
					else responsefirst=GetInt(line,responsefirst);
				} else if(line.startsWith("responselast")) {
					if(line.startsWith("responselastIO=")) responselastIO=GetInt(line,responselastIO);
					else if(line.startsWith("responselastVSD=")) responselastVSD=GetInt(line,responselastVSD);
					else responselast=GetInt(line,responselast);
				} else if(keepComInIni && 2<line.length()) {
					if (IJ.isWindows()) iniComments+=(char)13;
					iniComments+="\n";
					iniComments+=line;
				}
			} else if(line.startsWith("M")) {
				// keeps downcompatibility to previous plugin versions
				if(line.startsWith("MeanNo=")) FilterZ=GetInt(line,FilterZ);
				else if(line.startsWith("MeanLoHi=")) FilterXy=GetDouble(line,FilterXy);
				else if(line.startsWith("MeanLoHiIO=")) FilterXyIO=GetDouble(line,FilterXyIO);
				else if(line.startsWith("MeanLoHiVSD=")) FilterXyVSD=GetDouble(line,FilterXyVSD);
				else if(line.startsWith("Method")) AMethod=GetInt(line,AMethod);
				else if(keepComInIni && 2<line.length()) {
					if (IJ.isWindows()) iniComments+=(char)13;
					iniComments+="\n";
					iniComments+=line;
				}
			} else if(line.startsWith("Filter")) {
				if(line.startsWith("FilterZNo=")) FilterZ=GetInt(line,FilterZ);
				else if(line.startsWith("FilterZ=")) FilterZ=GetInt(line,FilterZ);
				else if(line.startsWith("FilterZIO=")) FilterZIO=GetInt(line,FilterZIO);
				else if(line.startsWith("FilterZVSD=")) FilterZVSD=GetInt(line,FilterZVSD);
				else if(line.startsWith("FilterXy=")) FilterXy=GetDouble(line,FilterXy);
				else if(line.startsWith("FilterXyIO=")) FilterXyIO=GetDouble(line,FilterXyIO);
				else if(line.startsWith("FilterXyVSD=")) FilterXyVSD=GetDouble(line,FilterXyVSD);
				else if(line.startsWith("FilterType=")) FilterMean=GetBoolean(line,FilterMean);
				else if(line.startsWith("FilterMean=")) FilterMean=GetBoolean(line,FilterMean);
				else if(keepComInIni && 2<line.length()) {
					if (IJ.isWindows()) iniComments+=(char)13;
					iniComments+="\n";
					iniComments+=line;
				}
			} else if(line.startsWith("Do")) {
				if(line.startsWith("DoFilterAv=")) DoFilterAv=GetBoolean(line,DoFilterAv);
				else if(line.startsWith("DoFilterXY=")) DoFilterXY=GetBoolean(line,DoFilterXY);
				else if(line.startsWith("DoFilterBR=")) DoFilterBR=GetBoolean(line,DoFilterBR);
				else if(line.startsWith("DoFilterDelta=")) DoFilterDelta=GetBoolean(line,DoFilterDelta);
				else if(line.startsWith("DoZFilterAv=")) DoZFilterAv=GetBoolean(line,DoZFilterAv);
				//  else if(line.startsWith("DoZFilterNo=")) DoZFilterNo=GetBoolean(line,DoZFilterNo);
				else if(keepComInIni && 2<line.length()) {
					if (IJ.isWindows()) iniComments+=(char)13;
					iniComments+="\n";
					iniComments+=line;
				}
			} else if(line.startsWith("N")) {
				if(line.startsWith("Normalizeb=")) Normalizeb=GetBoolean(line,Normalizeb);
				else if(line.startsWith("NbChkbxs=")) NbChkbxs=GetInt(line,NbChkbxs);
				else {
					int iidx=-1;	
					long fsze=GetIdxLong(line,0);
					if (fsze>0) {
						for (int ni=0;ni<tps;ni++) if (fsze==filesizeA[ni]) iidx=ni;
						if (iidx<0) {
							long [] tl = new long[tps];
							int [] ti = new int[tps];
							double [] td = new double[tps];
							
							for (iidx=0; iidx<tps; iidx++) tl[iidx]=filesizeA[iidx];
							filesizeA=new long[tps+1];
							for (iidx=0; iidx<tps; iidx++) filesizeA[iidx]=tl[iidx];
							filesizeA[tps]=fsze;
								
							for (iidx=0; iidx<tps; iidx++) ti[iidx]=framecntA[iidx];
							framecntA=new int[tps+1];
							for (iidx=0; iidx<tps; iidx++) framecntA[iidx]=ti[iidx];
							
							for (iidx=0; iidx<tps; iidx++) ti[iidx]=baselinefirstA[iidx];
							baselinefirstA=new int[tps+1];
							for (iidx=0; iidx<tps; iidx++) baselinefirstA[iidx]=ti[iidx];
							
							for (iidx=0; iidx<tps; iidx++) ti[iidx]=baselinelastA[iidx];
							baselinelastA=new int[tps+1];
							for (iidx=0; iidx<tps; iidx++) baselinelastA[iidx]=ti[iidx];
							
							for (iidx=0; iidx<tps; iidx++) ti[iidx]=responsefirstA[iidx];
							responsefirstA=new int[tps+1];
							for (iidx=0; iidx<tps; iidx++) responsefirstA[iidx]=ti[iidx];
							
							for (iidx=0; iidx<tps; iidx++) ti[iidx]=responselastA[iidx];
							responselastA=new int[tps+1];
							for (iidx=0; iidx<tps; iidx++) responselastA[iidx]=ti[iidx];
							
							for (iidx=0; iidx<tps; iidx++) ti[iidx]=FilterZA[iidx];
							FilterZA=new int[tps+1];
							for (iidx=0; iidx<tps; iidx++) FilterZA[iidx]=ti[iidx];
							
							for (iidx=0; iidx<tps; iidx++) td[iidx]=FilterXyA[iidx];
							FilterXyA=new double [tps+1];
							for (iidx=0; iidx<tps; iidx++) FilterXyA[iidx]=td[iidx];
							
							iidx=tps;
							tps++;
						} // --  if (iidx<0)
						if(line.startsWith("Nframecnt_")) framecntA[iidx]=GetInt(line,framecntA[iidx]);
						else if(line.startsWith("NF")) {
							if(line.startsWith("NFilterZNo_")) FilterZA[iidx]=GetInt(line,FilterZA[iidx]);
							else if(line.startsWith("NFilterZ_")) FilterZA[iidx]=GetInt(line,FilterZA[iidx]);
							else if(line.startsWith("NFilterXy_")) FilterXyA[iidx]=GetDouble(line,FilterXyA[iidx]);
						} else if(line.startsWith("NM")) {
							if(line.startsWith("NMeanNo_")) FilterZA[iidx]=GetInt(line,FilterZA[iidx]);
							else if(line.startsWith("NMeanLoHi_")) FilterXyA[iidx]=GetDouble(line,FilterXyA[iidx]);
						} else if(line.startsWith("Nb")) {
							if(line.startsWith("Nbaselinefirst_")) baselinefirstA[iidx]=GetInt(line,baselinefirstA[iidx]);
							else if(line.startsWith("Nbaselinelast_")) baselinelastA[iidx]=GetInt(line,baselinelastA[iidx]);
						} else if(line.startsWith("Nr")) {
							if(line.startsWith("Nresponsefirst_")) responsefirstA[iidx]=GetInt(line,responsefirstA[iidx]);
							else if(line.startsWith("Nresponselast_")) responselastA[iidx]=GetInt(line,responselastA[iidx]);
						}
					} else if(keepComInIni && 2<line.length()) {
						if (IJ.isWindows()) iniComments+=(char)13;
						iniComments+="\n";
						iniComments+=line;
					}
				}
			} else if(line.startsWith("Show")) {
				if(line.startsWith("ShowD")) {
					if(line.startsWith("ShowDiff=")) ShowDiff=GetBoolean(line,ShowDiff);
					//  else if(line.startsWith("ShowDiv=")) ShowDiv=GetBoolean(line,ShowDiv);
					else if(line.startsWith("ShowDelta=")) ShowDelta=GetBoolean(line,ShowDelta);
					else if(keepComInIni && 2<line.length()) {
						if (IJ.isWindows()) iniComments+=(char)13;
						iniComments+="\n";
						iniComments+=line;
					}
				} else if(line.startsWith("ShowR")) {
					if(line.startsWith("ShowRaw=")) ShowRaw=GetBoolean(line,ShowRaw);
					else if(line.startsWith("ShowResp=")) ShowResp=GetBoolean(line,ShowResp);
					else if(line.startsWith("ShowRatio=")) ShowRatio=GetBoolean(line,ShowRatio);
					else if(keepComInIni && 2<line.length()) {
						if (IJ.isWindows()) iniComments+=(char)13;
						iniComments+="\n";
						iniComments+=line;
					}
				} else if(line.startsWith("ShowP")) {
					if(line.startsWith("ShowPerc=")) ShowPerc=GetBoolean(line,ShowPerc);
					else if(line.startsWith("ShowPStack=")) ShowPStack=GetBoolean(line,ShowPStack);
					else if(keepComInIni && 2<line.length()) {
						if (IJ.isWindows()) iniComments+=(char)13;
						iniComments+="\n";
						iniComments+=line;
					}
				} else if(line.startsWith("ShowQual=")) ShowQual=GetBoolean(line,ShowQual);
				else if(line.startsWith("ShowBase=")) ShowBase=GetBoolean(line,ShowBase);
				else if(line.startsWith("ShowNo=")) ShowNo=GetBoolean(line,ShowNo);
				else if(line.startsWith("ShowAv=")) ShowAv=GetBoolean(line,ShowAv);
				else if(line.startsWith("ShowLog=")) DoShowLog=GetBoolean(line,DoShowLog);
				else if(keepComInIni && 2<line.length()) {
					if (IJ.isWindows()) iniComments+=(char)13;
					iniComments+="\n";
					iniComments+=line;
				}
			} else if(line.startsWith("Save")) {
				if(line.startsWith("SaveD")) {
					if(line.startsWith("SaveDiff=")) SaveDiff=GetBoolean(line,SaveDiff);
					//  else if(line.startsWith("SaveDiv=")) SaveDiv=GetBoolean(line,SaveDiv);
					else if(line.startsWith("SaveDelta=")) SaveDelta=GetBoolean(line,SaveDelta);
					else if(keepComInIni && 2<line.length()) {
						if (IJ.isWindows()) iniComments+=(char)13;
						iniComments+="\n";
						iniComments+=line;
					}
				} else if(line.startsWith("SaveR")) {
					if(line.startsWith("SaveResp=")) SaveResp=GetBoolean(line,SaveResp);
					else if(line.startsWith("SaveRatio=")) SaveRatio=GetBoolean(line,SaveRatio);
					else if(keepComInIni && 2<line.length()) {
						if (IJ.isWindows()) iniComments+=(char)13;
						iniComments+="\n";
						iniComments+=line;
					}
				} else if(line.startsWith("SaveP")) {
					if(line.startsWith("SavePerc=")) SavePerc=GetBoolean(line,SavePerc);
					else if(line.startsWith("SavePStack=")) SavePStack=GetBoolean(line,SavePStack);
					else if(keepComInIni && 2<line.length()) {
						if (IJ.isWindows()) iniComments+=(char)13;
						iniComments+="\n";
						iniComments+=line;
					}
				} else if(line.startsWith("SaveLog=")) DoSaveLog=GetBoolean(line,DoSaveLog);
				else if(line.startsWith("SaveBase=")) SaveBase=GetBoolean(line,SaveBase);
				else if(line.startsWith("SaveNo=")) SaveNo=GetBoolean(line,SaveNo);
				else if(line.startsWith("SaveAv=")) SaveAv=GetBoolean(line,SaveAv);
				else if(line.startsWith("SaveQual=")) SaveQual=GetBoolean(line,SaveQual);
				else if(keepComInIni && 2<line.length()) {
					if (IJ.isWindows()) iniComments+=(char)13;
					iniComments+="\n";
					iniComments+=line;
				}
			} else if(line.startsWith("AMethod")) AMethod=GetInt(line,AMethod);  //  name change for downcompatibility to previous plugin versions
			else if(line.startsWith("interpolateXY=")) interpolateXY=GetBoolean(line,interpolateXY);
			else if(line.startsWith("allowBlankArtifact=")) allowBlankArtifact=GetBoolean(line,allowBlankArtifact);
			else if(line.startsWith("commonBase=")) commonBase=GetBoolean(line,commonBase);
			else if(line.startsWith("t=")) {
				if (0==lastRun) lastRun=GetLong(line,lastRun);
			} else if(line.startsWith("src")) src=GetInt(line,src);
			// else if(line.startsWith("lineHeight")) lineHeight=GetInt(line,lineHeight);
			else if(line.startsWith("checkboxHeight")) checkboxHeight=GetInt(line,checkboxHeight);
			else if(line.startsWith("lastError=")) {
				// do nothing;  we can do some fresh ones in this run...
			} else if(line.startsWith("nhl")) nhl=s2array(GetString(line,""), nhl, false);
			else if(line.startsWith("hasControl=")) hasControl=GetBoolean(line,hasControl);
			else if(line.startsWith("filesize=")) filesize0=GetLong(line,filesize);
			else if(line.startsWith("verboseMode=")) {
				verboseMode=GetBoolean(line,hasControl);
				if (verboseMode) silentMode=false;
			} else if(line.startsWith("silentMode=")) {
				silentMode=GetBoolean(line,hasControl);
				if (silentMode) verboseMode=false;
			} else if(keepComInIni) {
				if(2<line.length() && !line.startsWith("Ver=") && !line.startsWith("/n")) {
					if (IJ.isWindows()) iniComments+=(char)13;
					iniComments+="\n";
					iniComments+=line;
				}
			}
		} //  while (( line = input.readLine()) != null)
	} //  try
	catch (FileNotFoundException ex) {Result=false;}
	catch (IOException ex){Result=false;}
	finally {
		// flush and close both "input" and its underlying FileReader
		try {if (input!= null) input.close();}
		catch (IOException ex) {}
	}
	src++;
	return Result;
}

private boolean deleteSettings(String pth) {
	if (pth==null) return false;
	if (0==pth.length()) return false;
	File aFile = null;
	try {
		if (pth.endsWith(BSL)) aFile = new File(pth+IniFileName);
		else aFile = new File(pth+BSL+IniFileName);
		if (!aFile.isFile()) return true;
		if (!aFile.canRead()) return false;
	}
	catch (SecurityException ex) {return false;}
	catch (NullPointerException ex) {return false;}
	if (null==aFile) return false;
	
	String ret = "";
	if (IJ.isWindows()) ret+=(char)13;
	ret +="\n";
	// String ret = "\n";
	// if (IJ.isWindows()) ret =""+(char)13+(char)10;
	
	// try to read out the ini file and keep settings stored in it
	BufferedReader input = null;
	boolean Result=true;
	try {
		input = new BufferedReader(new FileReader(""+aFile));
		String line = "";
		while (( line = input.readLine()) != null){
			if (2 < line.length()) {
				storedSettings += line;
				storedSettings += ret;
			}
		}
	}
	catch (FileNotFoundException ex) {Result=false;}
	catch (IOException ex){Result=false;}
	catch (SecurityException ex) {Result=false;}
	catch (NullPointerException ex) {Result=false;}
	finally {
		// flush and close both "input" and its underlying FileReader
		try {if (input!= null) input.close();}
		catch (IOException ex) {}
	}
	
	// try to delete the ini file
	if (Result) {
		try {
			aFile.delete();
			if (aFile.isFile()) return false;
		}
		catch (SecurityException e) {return false;}
		catch (NullPointerException ex) {return false;}
	} else return false;
	return Result;
}

private boolean restoreSettings(String pth) {
	if (pth==null) return false;
	if (3 > storedSettings.length()) return false;
	
	File aFile = null;
	try {
		if (pth.endsWith(BSL)) aFile = new File(pth+IniFileName);
		else aFile = new File(pth+BSL+IniFileName);
		if (aFile.isFile()) aFile.delete();
		if (aFile.isFile()) return false;
	}
	catch (SecurityException ex) {return false;}
	catch (NullPointerException ex) {return false;}
	
	boolean Result=true;
	FileWriter output = null;
	try {
		if (!aFile.createNewFile()) return false;
		output =new FileWriter(aFile);
		output.write(storedSettings);
	}
	catch (IOException ex) {Result=false;}
	finally {
		//  - flush and close both "output" and its underlying FileWriter
		try {if (output != null) output.close();}
		catch (IOException ex) {}
	}
	return Result;
}

private boolean findPathAndDoSettings (String path, int method, boolean verb) {

	if (DONT_KEEP==storeModus) return false;
	switch(method) {
		case GET:
			IJ.showStatus("Retrieve previous settings...");
			break;
		case SET:
			IJ.showStatus("Save settings...");
			break;
		case CHECK:
			IJ.showStatus("Check settings...");
			break;
		default:
			//  not defined in his version
			return false;
	}
	
	boolean gotSet=false;
	if (0< path.length()) gotSet=doSettings(path, method);
	
	String TDir="";
	String reportMsg="";
	if (!gotSet) {
		int tstoreModus = storeModus;
		switch (storeModus) {
			case KEEP_GLOBAL:
				if (IJ.isWindows()) {
					if (0==winDir.length()) winDir = GetEnvironS("WINDIR");
					TDir = winDir;
					if(0<TDir.length()) gotSet=doSettings(TDir, method);
					if (!gotSet) {
						winDir = GetEnvironS("WINDIR");
						TDir = winDir;
						if(0<TDir.length()) gotSet=doSettings(TDir, method);
					}
				}
				if (!gotSet) {
					//  TDir = System.getProperty("user.home");
					try {
					 	if (IJ.getVersion().compareTo("1.34s") < 0)
							TDir = System.getProperty("user.home");
						else TDir = IJ.getDirectory("home");
					}
					catch(NoSuchMethodError e) {
						TDir = System.getProperty("user.home");
						if (0==TDir.length() && IJ.isWindows()) TDir = GetEnvironS("HOMEPATH");
					}
					if (null != TDir) if(1<TDir.length()) gotSet=doSettings(TDir, method);
					if (gotSet) tstoreModus=KEEP_PERSONAL;
				}
				if (!gotSet) {
					try {
						TDir = Prefs.getHomeDir();
					} 
					catch (NoSuchMethodError e) {
						if (null != TDir) if(1<TDir.length()) gotSet=doSettings(TDir, method);
						if (gotSet) tstoreModus=KEEP_LOCAL;
					}
				}
				if (!gotSet) {
					try {
						// TDir = IJ.getDirectory("imagej");
						TDir = Menus.getPlugInsPath();
						if (TDir !=null) TDir = (new File(TDir)).getParent();
					} catch (NoSuchMethodError e) {}
					
					if (null != TDir) if(1<TDir.length()) gotSet=doSettings(TDir, method);
					if (gotSet) tstoreModus=KEEP_LOCAL;
				}
				break;
			case KEEP_PERSONAL:
				//  TDir = System.getProperty("user.home");
				try {
				 	if (IJ.getVersion().compareTo("1.34s") < 0)
						TDir = System.getProperty("user.home");
					else TDir = IJ.getDirectory("home");
				}
				catch(NoSuchMethodError e) {
					TDir = System.getProperty("user.home");
					if (0==TDir.length() && IJ.isWindows())  TDir = GetEnvironS("HOMEPATH");
				}
				if (null != TDir) if(1<TDir.length()) gotSet=doSettings(TDir, method);
				if (!gotSet) {
					try {
						TDir = Prefs.getHomeDir();
						if (null != TDir) if(1<TDir.length()) gotSet=doSettings(TDir, method);
					} catch (NoSuchMethodError e) {}
					if (gotSet) tstoreModus=KEEP_LOCAL;
				}
				if (!gotSet) {
					try {
						// TDir = IJ.getDirectory("imagej");
						TDir = Menus.getPlugInsPath();
						if (TDir !=null) TDir = (new File(TDir)).getParent();
					} catch (NoSuchMethodError e) {}
					if (null != TDir) if(1<TDir.length()) gotSet=doSettings(TDir, method);
					if (gotSet) tstoreModus=KEEP_LOCAL;
				}
				if (!gotSet && IJ.isWindows()) {
					if (0 < winDir.length()) winDir = GetEnvironS("WINDIR");
					TDir = winDir;
					if (0<TDir.length()) gotSet=doSettings(TDir, method);
					if (!gotSet) {
						winDir = GetEnvironS("WINDIR");
						TDir = winDir;
						if (0<TDir.length()) gotSet=doSettings(TDir, method);
					}
					if (gotSet) tstoreModus=KEEP_GLOBAL;
				}
				break;
			case KEEP_LOCAL:
				try {
					// TDir = IJ.getDirectory("imagej");
					TDir = Menus.getPlugInsPath();
					if (TDir !=null) TDir = (new File(TDir)).getParent();
				} catch (NoSuchMethodError e) {}
				if (null != TDir) if(1<TDir.length()) gotSet=doSettings(TDir, method);
				if (!gotSet) {
					try {
						TDir = Prefs.getHomeDir();
						if (null != TDir) if(1<TDir.length()) gotSet=doSettings(TDir, method);
					} catch (NoSuchMethodError e) {}
				}
				if (!gotSet) {
					//  TDir = System.getProperty("user.home");
					try {
					 	if (IJ.getVersion().compareTo("1.34s") < 0)
							TDir = System.getProperty("user.home");
						else TDir = IJ.getDirectory("home");
					}
					catch(NoSuchMethodError e) {
						TDir = System.getProperty("user.home");
						if (0==TDir.length() && IJ.isWindows())  TDir = GetEnvironS("HOMEPATH");
					}
					if (null != TDir) if(1<TDir.length()) gotSet=doSettings(TDir, method);
					if (gotSet) tstoreModus=KEEP_PERSONAL;
				}
				if (!gotSet && IJ.isWindows()) {
					if (0 < winDir.length()) winDir = GetEnvironS("WINDIR");
					TDir = winDir;
					if (0<TDir.length()) gotSet=doSettings(TDir, method);
					if (!gotSet) {
						winDir = GetEnvironS("WINDIR");
						TDir = winDir;
						if (0<TDir.length()) gotSet=doSettings(TDir, method);
					}
					if (gotSet) tstoreModus=KEEP_GLOBAL;
				}
		} // switch
		if (gotSet) {
			Pth = TDir;
			if (storeModus != tstoreModus) {
				switch (tstoreModus) {
					case KEEP_GLOBAL: reportMsg="GLOBALLY "; break;
					case KEEP_PERSONAL: reportMsg="PERSONALLY "; break;
					case KEEP_LOCAL: reportMsg="LOCALY "; break;
					default: reportMsg="custom ";
				}
				storeModus = tstoreModus;
			} 
		}
	}
	switch(method) {
		case GET:
			if (gotSet) {
				String Msg = "Settings are retrieved "+reportMsg+"from '"+Pth+"'.";
				IJ.showStatus(Msg);
				if (verboseMode) new MessageDialog(null, PluginTitle, Msg);
			} else if (verb) IJ.error("Failed to retrieve previous settings from ini file!");
			break;
		case SET:
			if (gotSet) {
				String Msg = "Settings are saved "+reportMsg+"to '"+Pth+"'.";
				IJ.showStatus(Msg);
				if (verboseMode) new MessageDialog(null, PluginTitle, Msg);
			} else if (verb) IJ.error("Failed to save settings to ini file!");
			try {
				Prefs.set("VSDproc_lastRun.long", ""+System.currentTimeMillis());
				Prefs.set("VSDproc_store.int", (double)storeModus);
				Prefs.set("VSDproc_pth.string", Pth);
				Prefs.set("checkboxHeight.int", (double) checkboxHeight);
			} catch (NoSuchMethodError e) {}
			break;
		case CHECK:
			if (gotSet) {
				String Msg = "Settings can be saved "+reportMsg+"in '"+Pth+"'.";
				IJ.showStatus(Msg);
				if (verboseMode) new MessageDialog(null, PluginTitle, Msg);
			} else if (verb) IJ.error("Failed to access previous settings in ini file!");
	}
	return gotSet;
}

private boolean setSettings(String pth) {
	if (pth==null) return false;
	File aFile = null;
	try {
		if (pth.endsWith(BSL)) aFile = new File(pth+IniFileName);
		else aFile = new File(pth+BSL+IniFileName);
		if (aFile.isFile()) aFile.delete();
		if (aFile.isFile()) return false;
	}
	catch (SecurityException ex) {return false;}
	catch (NullPointerException ex) {return false;}
	
	//  ----------  manage settings for comaptibility in later runs  -----------
	if (0<counter[0][0]&&1<nhl.length && hasControl) {
		baselinefirstVSD = baselinefirst;
		baselinelastVSD = baselinelast;
		responsefirstVSD = responsefirst;
		responselastVSD = responselast;
		FilterXyVSD = FilterXy;
		FilterZVSD = FilterZ;
	} else {
		baselinefirstIO = baselinefirst;
		baselinelastIO = baselinelast;
		responsefirstIO = responsefirst;
		responselastIO = responselast;
		FilterXyIO = FilterXy;
		FilterZIO = FilterZ;
	}
	for (int ni=0;ni<tps;ni++) if (filesize==filesizeA[ni]) {
		baselinefirstA[ni] = baselinefirst;
		baselinelastA[ni] = baselinelast;
		responsefirstA[ni] = responsefirst;
		responselastA[ni] = responselast;
		FilterZA[ni] = FilterZ;
		FilterXyA[ni] = FilterXy;
	}
	
	boolean Result=true;
	FileWriter output = null;
	String ret = "";
	if (IJ.isWindows()) ret+=(char)13;
	ret +="\n";
	try {
		if (!aFile.createNewFile()) return false;
		output =new FileWriter(aFile);
		output.write("DirIn="+DirIn+ret);
		String vst = "Ver="+PluginVersion+":r"+src+"@";
		try {vst+=IJ.getVersion();}
		catch (NoSuchMethodError ex) {vst+="<1.32";}
		output.write(vst+ret);
		int pc=0;
		for (int hc=0; hc<HDir.length; hc++) if (0<HDir[hc].length()) if (0!=HDir[hc].compareTo(DirIn)) output.write("HDir_"+(pc++)+"="+HDir[hc]+ret);
		if (0<winDir.length()) output.write("DirWin="+winDir+ret);
		if (0<homeDrive.length()) output.write("DirHomeDrive="+homeDrive+ret);
		if (0<publicDir.length()) output.write("DirPublic="+publicDir+ret);
		if (0<progrDir.length()) output.write("DirProgr="+progrDir+ret);
		output.write("baselinefirst="+baselinefirst+ret);
		output.write("baselinelast="+baselinelast+ret);
		output.write("baselinefirstIO="+baselinefirstIO+ret);
		output.write("baselinelastIO="+baselinelastIO+ret);
		output.write("baselinefirstVSD="+baselinefirstVSD+ret);
		output.write("baselinelastVSD="+baselinelastVSD+ret);
		output.write("responsefirst="+responsefirst+ret);
		output.write("responselast="+responselast+ret);
		output.write("responsefirstIO="+responsefirstIO+ret);
		output.write("responselastIO="+responselastIO+ret);
		output.write("responsefirstVSD="+responsefirstVSD+ret);
		output.write("responselastVSD="+responselastVSD+ret);
		output.write("FilterMean="+FilterMean+ret);
		if(Normalizeb) output.write("Normalizeb=1"+ret);
		else output.write("Normalizeb=0"+ret);
		output.write("FilterZ="+FilterZ+ret);
		output.write("FilterZIO="+FilterZIO+ret);
		output.write("FilterZVSD="+FilterZVSD+ret);
		output.write("FilterXy="+FilterXy+ret);
		output.write("FilterXyIO="+FilterXyIO+ret);
		output.write("FilterXyVSD="+FilterXyVSD+ret);
		if(DoFilterXY) output.write("DoFilterXY=1"+ret);
		else output.write("DoFilterXY=0"+ret);
		if(DoFilterAv) output.write("DoFilterAv=1"+ret);
		else output.write("DoFilterAv=0"+ret);
		if(DoFilterBR) output.write("DoFilterBR=1"+ret);
		else output.write("DoFilterBR=0"+ret);
		if(DoFilterDelta) output.write("DoFilterDelta=1"+ret);
		else output.write("DoFilterDelta=0"+ret);
		if(DoZFilterAv) output.write("DoZFilterAv=1"+ret);
		else output.write("DoZFilterAv=0"+ret);
		for (int iidx=0;iidx<tps;iidx++) {
			output.write("Nframecnt_"+filesizeA[iidx]+"="+framecntA[iidx]+ret);
			output.write("Nbaselinefirst_"+filesizeA[iidx]+"="+baselinefirstA[iidx]+ret);
			output.write("Nbaselinelast_"+filesizeA[iidx]+"="+baselinelastA[iidx]+ret);
			output.write("Nresponsefirst_"+filesizeA[iidx]+"="+responsefirstA[iidx]+ret);
			output.write("Nresponselast_"+filesizeA[iidx]+"="+responselastA[iidx]+ret);
			output.write("NFilterZ_"+filesizeA[iidx]+"="+FilterZA[iidx]+ret);
			output.write("NFilterXy_"+filesizeA[iidx]+"="+FilterXyA[iidx]+ret);
		}
		if (0<filesize) {
			boolean NewSet=true;
			for (int iidx=0;iidx<tps;iidx++) if (filesize==filesizeA[iidx]) NewSet=false;
			if (NewSet) {
				output.write("Nframecnt_"+filesize+"="+framecnt+ret);
				output.write("Nbaselinefirst_"+filesize+"="+baselinefirst+ret);
				output.write("Nbaselinelast_"+filesize+"="+baselinelast+ret);
				output.write("Nresponsefirst_"+filesize+"="+responsefirst+ret);
				output.write("Nresponselast_"+filesize+"="+responselast+ret);
				output.write("NFilterZ_"+filesize+"="+FilterZ+ret);
				output.write("NFilterXy_"+filesize+"="+FilterXy+ret);
			}
		}
		if (ShowRaw) output.write("ShowRaw=1"+ret);
		else output.write("ShowRaw=0"+ret);
		if(ShowBase)output.write("ShowBase=1"+ret);
		else output.write("ShowBase=0"+ret);
		if(ShowResp)output.write("ShowResp=1"+ret);
		else output.write("ShowResp=0"+ret);
		if(ShowDiff)output.write("ShowDiff=1"+ret);
		else output.write("ShowDiff=0"+ret);
		if(ShowPerc)output.write("ShowPerc=1"+ret);
		else output.write("ShowPerc=0"+ret);
		//  if(ShowDiv)output.write("ShowDiv=1"+ret);
		//  else output.write("ShowDiv=0"+ret);
		if(ShowNo)output.write("ShowNo=1"+ret);
		else output.write("ShowNo=0"+ret);
		if(ShowPStack)output.write("ShowPStack=1"+ret);
		else output.write("ShowPStack=0"+ret);
		if(ShowAv)output.write("ShowAv=1"+ret);
		else output.write("ShowAv=0"+ret);
		if(ShowDelta)output.write("ShowDelta=1"+ret);
		else output.write("ShowDelta=0"+ret);
		if(ShowRatio)output.write("ShowRatio=1"+ret);
		else output.write("ShowRatio=0"+ret);

		if(SaveBase)output.write("SaveBase=1"+ret);
		else output.write("SaveBase=0"+ret);
		if(SaveResp)output.write("SaveResp=1"+ret);
		else output.write("SaveResp=0"+ret);
		if(SaveDiff)output.write("SaveDiff=1"+ret);
		else output.write("SaveDiff=0"+ret);
		if(SavePerc)output.write("SavePerc=1"+ret);
		else output.write("SavePerc=0"+ret);
		//  if(SaveDiv)output.write("SaveDiv=1"+ret);
		//  else output.write("SaveDiv=0"+ret);
		if(SaveNo)output.write("SaveNo=1"+ret);
		else output.write("SaveNo=0"+ret);
		if(SavePStack)output.write("SavePStack=1"+ret);
		else output.write("SavePStack=0"+ret);
		if(SaveAv)output.write("SaveAv=1"+ret);
		else output.write("SaveAv=0"+ret);
		if(SaveDelta)output.write("SaveDelta=1"+ret);
		else output.write("SaveDelta=0"+ret);
		if(SaveRatio)output.write("SaveRatio=1"+ret);
		else output.write("SaveRatio=0"+ret);

		if(ShowQual)output.write("ShowQual=1"+ret);
		else output.write("ShowQual=0"+ret);
		if(SaveQual)output.write("SaveQual=1"+ret);
		else output.write("SaveQual=0"+ret);
		if(DoShowLog)output.write("ShowLog=1"+ret);
		else output.write("ShowLog=0"+ret);
		if(DoSaveLog)output.write("SaveLog=1"+ret);
		else output.write("SaveLog=0"+ret);
		if(verboseMode)output.write("verboseMode=1"+ret);
		else output.write("verboseMode=0"+ret);
		if(silentMode)output.write("silentMode=1"+ret);
		else output.write("silentMode=0"+ret);		
		if(interpolateXY)output.write("interpolateXY=1"+ret);
		else output.write("interpolateXY=0"+ret);
		output.write("src="+src+ret);
		output.write("t="+System.currentTimeMillis()+ret);
		// output.write("lineHeight="+lineHeight+ret);
		output.write("checkboxHeight="+checkboxHeight+ret);
		output.write("Method="+AMethod+ret);
		String nns=nhl[0];
		for (int i=1; i<nhl.length;i++) nns+=(","+nhl[i]);
		output.write("nhl="+nns+ret);
		if(hasControl)output.write("hasControl=1"+ret);
		else output.write("hasControl=0"+ret);
		output.write("NbChkbxs="+NbChkbxs+ret);
		if(blankArtifact) output.write("blankArtifact=1"+ret);
		else output.write("blankArtifact=0"+ret);
		if(allowBlankArtifact)output.write("allowBlankArtifact=1"+ret);
		else output.write("allowBlankArtifact=0"+ret);
		if(commonBase)output.write("commonBase=1"+ret);
		else output.write("commonBase=0"+ret);
		output.write("blankFrom="+blankFrom+ret);
		output.write("blankTo="+blankTo+ret);
		output.write("filesize="+filesize+ret);
		if (0<errString.length()) output.write("lastError="+errString+ret);
		if (keepComInIni && 0<iniComments.length()) output.write(iniComments+ret);
	}
	catch (IOException ex) {Result=false;}
	catch (SecurityException ex) {Result=false;}
	catch (NullPointerException ex) {Result=false;}
	finally {
		//  - flush and close both "output" and its underlying FileWriter
		try {if (output != null) output.close();}
		catch (IOException ex) {}
	}
	return Result;
}

private void ScaleTheMiddle(ImagePlus imp, int frame, double ScD) {
	if (null==imp) return;
	ImageProcessor ip;
	if (1>frame) frame=1;
	String impTitle="";
	try {
		impTitle=" '"+ imp.getShortTitle() +"''";
		if (1<imp.getStackSize()) ip = imp.getStack().getProcessor(frame);
		else ip = imp.getProcessor();
	}
	catch (IllegalArgumentException e) {
			if (!silentMode && !Canceled) IJ.error("Failed to scale image '"+impTitle+"'!");
			if (!Canceled) errString+=(", Failed to scale image '"+impTitle+"'");
			return;
	}
	catch (NullPointerException e) {
			if (!silentMode && !Canceled) IJ.error("Failed to scale images '"+impTitle+"'!");
			if (!Canceled) errString+=(", Failed to scale images '"+impTitle+"'");
			return;
	}
	try {
		int w=imp.getWidth();
		int h=imp.getHeight();
		if (h > 7 && w > 7) ip.setRoi(new Roi(w/8, h/8, w*3/4, h*3/4));
		else ip.setRoi(new Roi(0, 0, w, h));
	}
	catch (NoSuchMethodError ex) {
		int w=imp.getWidth();
		int h=imp.getHeight();
		if (h > 7 && w > 7) ip.setRoi(new Rectangle(w/8, h/8, w*3/4, h*3/4));
		else ip.setRoi(new Rectangle(0, 0, w, h));
	}
	catch (IllegalArgumentException e) {}
	catch (NullPointerException e) {}
	try {
		ImageStatistics stats = ImageStatistics.getStatistics(ip, ImageStatistics.MIN_MAX, null);
		int hmin, hmax, threshold;
		int[] histogram = stats.histogram;
		if (ScD>0.0) threshold = (int)(stats.pixelCount*ScD/200.0);
		else threshold = 0;
		int i = -1;
		boolean found = false;
		int count = 0;
		do {
			i++;
			count += histogram[i];
			found = count>threshold;
		} while (!found && i<255);
		hmin = i;
		
		i = 256;
		count = 0;
		do {
			i--;
			count += histogram[i];
			found = count>threshold;
		} while (!found && i>0);
		hmax = i;
		
		if (hmax>hmin) {
			double min = stats.histMin+hmin*stats.binSize;
			double max = stats.histMin+hmax*stats.binSize;
			imp.getProcessor().setMinAndMax(min, max);
			
			//  imp.updateAndRepaintWindow();
			// - Calls updateAndDraw to update from the pixel data and draw the image, 
			// - and also repaints the image window to force the information displayed above the image (dimension, type, size) to be updated.
			imp.updateAndDraw();
			// -  Updates this image from the pixel data in its associated ImageProcessor, then displays it. 
			// -  Does nothing if there is no window associated with this image (i.e. show() has not been called).
		}
	}
	catch (IllegalArgumentException e) {
			if (!silentMode && !Canceled) IJ.error("Failed to scale image '"+impTitle+"'!");
			if (!Canceled) errString+=(", Failed to scale image '"+impTitle+"'");
	}
	catch (NullPointerException e) {
			if (!silentMode && !Canceled) IJ.error("Failed to scale images '"+impTitle+"'!");
			if (!Canceled) errString+=(", Failed to scale images '"+impTitle+"'");
	}
}

private static final String GetEnvironS(String env) {
	String S="";
	if (0==env.length()) return S;
	try {
		Process p = Runtime.getRuntime().exec( "cmd /c echo %"+env+"%" );
		BufferedReader in = new BufferedReader(new InputStreamReader(p.getInputStream()) );
		S=in.readLine();
		if(!new File(S).isDirectory()) S="";  //  test whether S is name of an existing directory
	}
	catch(SecurityException e) {return "";}
	catch (NullPointerException e) {return "";}
	catch(IOException e) {return "";}
	return S;
}
private int GetInt(String line, int oldint) {
	int newint=oldint;
	try {newint=Integer.parseInt(line.substring(line.indexOf("=")+1,line.length()));}
	catch(NumberFormatException e) {return oldint;}
	catch(IndexOutOfBoundsException e) {return oldint;}
	return newint;
}
private long GetLong(String line, long oldlong) {
	long newlong=oldlong;
	try {newlong=Long.parseLong(line.substring(line.indexOf("=")+1,line.length()));}
	catch(NumberFormatException e) {return oldlong;}
	catch(IndexOutOfBoundsException e) {return oldlong;}
	return newlong;
}
private long GetIdxLong(String line, long oldidx) {
	long newidx=oldidx;
	try {newidx=Long.parseLong(line.substring(line.indexOf("_")+1,line.indexOf("=")));}
	catch(NumberFormatException e) {return oldidx;}
	catch(IndexOutOfBoundsException e) {return oldidx;}
	return newidx;
}
private int GetIdxInt(String line, int oldidx) {
	int newidx=oldidx;
	try {newidx=Integer.parseInt(line.substring(line.indexOf("_")+1,line.indexOf("=")));}
	catch(NumberFormatException e) {return oldidx;}
	catch(IndexOutOfBoundsException e) {return oldidx;}
	return newidx;
}
private String GetString(String line,String oldString) {
	String newString=oldString;
	try {newString=line.substring(line.indexOf("=")+1,line.length());}
	catch(NumberFormatException e) {return oldString;}
	catch(IndexOutOfBoundsException e) {return oldString;}
	return newString;
}
private double GetDouble(String line, double olddouble) {
	double newdouble=olddouble;
	try {newdouble=Double.parseDouble(line.substring(line.indexOf("=")+1,line.length()));}
	catch(NumberFormatException e) {return olddouble;}
	catch(IndexOutOfBoundsException e) {return olddouble;}
	return newdouble;
}
private boolean GetBoolean(String line, boolean oldboolean) {
	boolean newboolean=oldboolean;
	try {if(0==line.substring(line.indexOf("=")+1,line.length()).compareTo("1")) newboolean=true;
	if(0==line.substring(line.indexOf("=")+1,line.length()).compareTo("0")) newboolean=false; }
	catch(NumberFormatException e) {return oldboolean;}
	catch(IndexOutOfBoundsException e) {return oldboolean;}
	return newboolean;
}
private int s2i (String s, int oldi) {
	int i = oldi;
	try { i = Integer.parseInt(s.trim());}
	catch(NumberFormatException e) {return oldi;}
	catch(IndexOutOfBoundsException e) {return oldi;}
	return i;
}
private double s2d (String s, double oldd) {
	double d = oldd;
	try {d = Double.parseDouble(s.trim());}
	catch(NumberFormatException e) {return oldd;}
	catch(IndexOutOfBoundsException e) {return oldd;}
	return d;
}
private String [] s2array (String nns, String [] s0, boolean verboseb) {
	int a=0;
	String [] s1=new String [0];
	nns=";"+nns+";";
	for (int p=0; p<nns.length(); p++) {
		int chi = nns.charAt(p);
		if (33>chi || 34 ==chi || 37==chi || 42==chi || 44==chi || 47 == chi || (57<chi && 61!=chi && 64> chi) || 92==chi || 124==chi ) {
			if (p-a>1) {
				String sn=nns.substring(a+1,p);
				
				//  trim numbers from end of string if there are any
				int pp = p-a-1; // sn.length();
				do {
					chi = sn.charAt(pp-1);
					if (47 < chi && 58 > chi) {
						pp--;
						if (0 == pp) sn="";
					} else {
						if (pp < sn.length()) sn = sn.substring(0, pp);
						pp = -1;
					}
				} while (pp > 0);
				
				//  make sure there is something left
				boolean stringOK = 0 < sn.length();
				//  if (stringOK) {IJ.error(sn); }  // +++++++++++
				
				//  exclude invalid file names
				if (IJ.isWindows()) {
					//  exclude names if similar already in array 
					for (int ii=0; ii<s1.length; ii++) if (stringOK) stringOK = 0!=sn.compareToIgnoreCase(s1[ii]);
					//  exclude file names invalid in DOS/Windows
					if (stringOK) {
						stringOK = 0 !=sn.compareToIgnoreCase("com");
						if (stringOK) stringOK = 0 !=sn.compareToIgnoreCase("lpt");
						if (!stringOK && verboseb) IJ.error("File Name '"+ sn +"#' would be invalid in Windows(TM).");
					}
				} else {
					//  exclude names already in array (non-Windows systems)
					for (int ii=0; ii<s1.length; ii++) if (stringOK) stringOK = 0!=sn.compareTo(s1[ii]);
				}
				
				//  expand string array by the new one (if valid)
				if (stringOK) {
					String [] s2= new String[s1.length];
					for (int ii=0; ii<s1.length; ii++) s2[ii] = s1[ii];
					s1= new String[s1.length+1];
					for (int ii=0; ii<s2.length; ii++) s1[ii] = s2[ii];
					s1[s2.length]=sn;
				}
			}
			a=p;
		}
	}
	if (0<s1.length) return s1;
	return s0;
}

private ImagePlus openImage(String Pnm, String Fnm) {
	// -- ignore upper- / lowercase in file name if applicable
	try {
		if (!IJ.isWindows()) {
		// if (!System.getProperty("os.name").startsWith("Windows")) {
			File directory = new File(Pnm);
			String [] sUc = null;   // --  array of file names
			if (directory.isDirectory())  sUc = directory.list();
			boolean fileFound=false;
			if (null != sUc) {
				for (int k=0; k<sUc.length; k++) if (!fileFound)  fileFound = 0 == sUc[k].compareTo (Fnm);
				for (int k=0; k<sUc.length; k++) if (!fileFound)  {
					fileFound = sUc[k].equalsIgnoreCase(Fnm);
					if (fileFound) Fnm=sUc[k];
				}
			}
			if (!fileFound) return null;
		}
	}
	catch (SecurityException e) {}
	catch (NullPointerException e) {}
	
	// -- get the file
	File F=new File(Pnm+BSL+Fnm);
	if (null==F) return null;
	Fnm=F.getName();
	// -- bugfix for downcompatibility
	if (Fnm.endsWith("'")) Fnm=Fnm.substring(0,Fnm.length()-1);
	// -- now let's try opening the file
	ImagePlus img=null;
	try {
		TiffDecoder td = null;
		if (Pnm.endsWith(BSL)) td = new TiffDecoder(Pnm, Fnm);
        		else td = new TiffDecoder(Pnm+BSL, Fnm);
        		FileInfo[] info=null;
        		info = td.getTiffInfo();
		if (info!=null) img = new Opener().openTiffStack(info);
	} 
	catch(IOException e) {
		if (!Canceled) {
			Fnm = Fnm+ "'";
			if (verboseMode) Fnm = Fnm+ ";\n("+e+")";
			if ((""+e).startsWith("java.io.IOException: Unsupported BitsPerSample")) 
				Fnm = Fnm+"\nImageJ version 1.33b or later required. - Please update.";
			if (!silentMode) IJ.error("Failed to load '"+Pnm+BSL+Fnm);
			errString+=(", Failed to load '"+Pnm+BSL+Fnm);
		}
		Canceled=true;
		// img = CloseImageWindow(img);
		img = null;
	}
	catch(OutOfMemoryError e) {
		if (!Canceled) {
			Fnm = Fnm+ "'";
			errString+=(", Failed to load '"+Pnm+BSL+Fnm+"("+e+") Probably not enough memory available");
			if (verboseMode) Fnm = Fnm+ ";\n("+e+")";
			if (!silentMode) IJ.error("Failed to load '"+Pnm+BSL+Fnm+"\nProbably not enough memory available.");
		}
		Canceled=true;
		img = CloseImageWindow(img);
		img = null;
	}
	return img;
}

private boolean CovertTo32bitStack (ImagePlus imp) {
	String impTitle="";
	try {
		impTitle=imp.getTitle();
		int itype=imp.getType();
		if (itype==ImagePlus.GRAY32) return true;  // nothing to do.
		// -- ImageConverter can't convert color images directly. We need to convert them to 8-bit greyscale first.
		if (itype==ImagePlus.COLOR_256 || itype==ImagePlus.COLOR_RGB) {
			new StackConverter(imp).convertToGray8();
			if (verboseMode) IJ.error(impTitle + " has been converted to greyscale format.\nColor information discarded!");
		}
		new StackConverter(imp).convertToGray32();
		imp.changes = true;
		return true;
	}
	catch (IllegalArgumentException e) {
		if (!Canceled) {
			if (verboseMode) IJ.error("Failed to covert image ''"+impTitle+"'' to a 32 bit stack;\n"+e);
			else if (!silentMode) IJ.error("Failed to covert image ''"+impTitle+"'' to a 32 bit stack!");
			errString+=(", Failed to covert image ''"+impTitle+"'' to a 32 bit stack!");
		}
	}
	catch(OutOfMemoryError e) {
		if (!silentMode && !Canceled) IJ.error("Failed to covert image ''"+impTitle+"'' to a 32 bit stack:\nNot enough memory!");
		if (!Canceled) errString+=(", Failed to covert image ''"+impTitle+"'' to a 32 bit stack: not enough memory");
	}	
	catch(NullPointerException e) {
		if (!silentMode && !Canceled) IJ.error("Failed to covert image ''"+impTitle+"'' to a 32 bit stack: not found!");
		if (!Canceled) errString+=(", Failed to covert image ''"+impTitle+"'' to a 32 bit stack: not found");
	}
	Undo.reset();
	Canceled=true;
	imp = CloseImageWindow(imp);
	imp=null;
	return false;
}

private void extendStack(ImagePlus i1, int eHeight) {
	String ImgTitle="";
	try {
		ImgTitle=i1.getShortTitle();
		double v1;
		int width = i1.getWidth();
		int height = i1.getHeight();
		ImageProcessor ip1;
		ImageProcessor ip2;
		int slices1 = i1.getStackSize();
		float[] ctable1 = i1.getCalibration().getCTable();
		ImageStack stack1 = i1.getStack();
		ImageStack stack2 = new ImageStack(width, eHeight);
		for (int n=1; n<=slices1; n++) {
			ip1 = stack1.getProcessor(n);
			ip2 = new FloatProcessor(width, eHeight);
			for (int x=0; x<width; x++) {
				for (int y=0; y<height && y<eHeight; y++) {
					v1 = (double) ip1.getPixelValue(x,y);
					ip2.putPixelValue(x,y,v1);
				}
			}
			if (1==n) ip2.setCalibrationTable(ctable1);
			stack2.addSlice("", ip2);
		}
		i1.setStack(null, stack2);
		return;
	}
	catch (NullPointerException e) {}
	catch (IllegalArgumentException e) {}
	if (!silentMode && !Canceled) IJ.error("Failed to calculate image ''"+ImgTitle+"''!");
	if (!Canceled) errString+=(", Failed to calculate image ''"+ImgTitle+"''");
	IJ.showStatus("canceled.");
	Canceled=true;
}

private void blankArtifactInStack(ImagePlus img1, int fromZ, int toZ) {
	if (null == img1) return;
	String ImageName = img1.getTitle();
	if (1 > fromZ || 1 > toZ) {
		if (!silentMode && !Canceled) IJ.error("Failed to blank artifact in image ''"+ImageName+"'':\nInvalid Limits!");
		if (!Canceled) errString+=(", Failed to blank artifact in image ''"+ImageName+"'': invalid limits");
		return;
	}
	try {
		float[] ctable1 = img1.getCalibration().getCTable();
		ImageStack stack1 = img1.getStack();
		int width = stack1.getWidth();
		int height = stack1.getHeight();
		int endZ = stack1.getSize();
		if (toZ > endZ) toZ = endZ;
		if (fromZ >  toZ || (1==fromZ && toZ == endZ)) {
			if (!silentMode && !Canceled) IJ.error("Failed to blank artifact in image ''"+ImageName+"'':\nInvalid parameters!");
			if (!Canceled) errString+=(", Failed to blank artifact in image ''"+ImageName+"'': invalid parameters");
			return;
		}
		
		ImageStack stack2 = img1.createEmptyStack();
		ImageProcessor ip1;
		if (1==fromZ) {
			for (int z=1; z <= toZ; z++) {
				ip1 = stack1.getProcessor(toZ+1).duplicate();
				stack2.addSlice(stack1.getSliceLabel(z), ip1);
			}
		} else {
			for (int z=1; z < fromZ-1; z++) {
				ip1 = stack1.getProcessor(z);
				stack2.addSlice(stack1.getSliceLabel(z), ip1);
			}
			if  (toZ == endZ) {
				for (int z=fromZ; z <= toZ; z++) {
					ip1 = stack1.getProcessor(fromZ-1).duplicate();
					stack2.addSlice(stack1.getSliceLabel(z), ip1);
				}
			} else {
				ImageProcessor ipF = stack1.getProcessor(fromZ-1);
				stack2.addSlice(stack1.getSliceLabel(fromZ-1), ipF);
				ipF = stack1.getProcessor(fromZ-1);
				ImageProcessor ipT = stack1.getProcessor(toZ+1);
				double vF, vT, v1;
				for (int z=fromZ; z <= toZ; z++) {
					ip1 = ipF.duplicate();
					for (int x=0; x<width; x++) for (int y=0; y<height; y++) {
					vF = (double) ipF.getPixelValue(x,y);
					vT = (double) ipT.getPixelValue(x,y);
					v1 = (vF * (double)(toZ-z+1)  +  vT* (double)(z-fromZ+1)) / (double)(toZ-fromZ+2);
						ip1.putPixelValue(x, y, v1);
					}
					stack2.addSlice(stack1.getSliceLabel(z), ip1);
				}
			}
		}
		if  (toZ < endZ) {
			for (int z=toZ+1; z <= endZ; z++) {
				ip1 = stack1.getProcessor(z);
				stack2.addSlice(stack1.getSliceLabel(z), ip1);
			}
		}
		img1.setStack(null, stack2);
	}
	catch(OutOfMemoryError e) {
		if (!silentMode && !Canceled) IJ.error("Failed to blank artifact in image ''"+ImageName+"'':\nNot enough memory!");
		if (!Canceled) errString+=(", Failed to blank artifact in image ''"+ImageName+"'': not enough memory");
	}
	catch(NullPointerException e) {
		if (!silentMode && !Canceled) IJ.error("Failed to blank artifact in image ''"+ImageName+"'':\nNot found!");
		if (!Canceled) errString+=(", Failed to blank artifact in image ''"+ImageName+"'': not found");
	}
	catch (IllegalArgumentException e) {
		if (!silentMode && !Canceled) IJ.error("Failed to blank artifact in ''"+ImageName+"'',\nprobably canceled.");
		if (!Canceled) errString+=(", Failed to blank artifact in image ''"+ImageName+"'': probably canceled");
	}
} // void blankArtifactInStack(ImagePlus img1, int fromZ, int toZ)

private ImagePlus duplicateStack(ImagePlus img1, String Title, boolean SelectFrames) {
	ImageStack stack2 = null;
	ImagePlus img3 = null;
	String ImageName = "";
	try {
		ImageName = img1.getTitle();
		float[] ctable1 = img1.getCalibration().getCTable();
		ImageStack stack1 = img1.getStack();
		int width = stack1.getWidth();
		int height = stack1.getHeight();
		int a = SelectFrames?baselinefirst :1;
		int o = SelectFrames?responselast:stack1.getSize();
		if (0==Title.length()) Title=img1.getShortTitle();
		stack2 = img1.createEmptyStack();
		for (int i=a; i<=o ; i++) if (!SelectFrames || i<=baselinelast ||  i>=responselast ) {
			ImageProcessor ip1 = stack1.getProcessor(i);
			ImageProcessor ip2 = ip1.duplicate();
			ip2.setCalibrationTable(ctable1);
			stack2.addSlice(stack1.getSliceLabel(i), ip2);
		}
		img3 = new ImagePlus(Title, stack2);
		img3.setStack(null, stack2);
		return img3;
	}
	catch(OutOfMemoryError e) {
		if (!silentMode && !Canceled) IJ.error("Failed to duplicate image ''"+ImageName+"'':\nNot enough memory!");
		if (!Canceled) errString+=(", Failed to duplicate image ''"+ImageName+"'': not enough memory");
	}
	catch(NullPointerException e) {
		if (!silentMode && !Canceled) IJ.error("Failed to duplicate image ''"+ImageName+"'':\nNot found!");
		if (!Canceled) errString+=(", Failed to duplicate image ''"+ImageName+"'': not found");
	}
	catch (IllegalArgumentException e) {
		if (!silentMode && !Canceled) IJ.error("Failed to duplicate ''"+ImageName+"'',\nprobably canceled.");
		if (!Canceled) errString+=(", Failed to duplicate image ''"+ImageName+"'': probably canceled");
	}
	Canceled=true;
	if (null != stack2) stack2.trim();
	stack2=null;
	img3 = CloseImageWindow (img3);
	IJ.showStatus("canceled.");
	return null;
}

private ImagePlus duplicateStack(ImagePlus img1, String Title) {
	return duplicateStack(img1, Title, false);
}

ImagePlus duplicateStack(ImagePlus img1) {
	return duplicateStack(img1, img1.getShortTitle(), false);
}

private String [] winList() {
	int [] IDLst = WindowManager.getIDList();
	if (null==IDLst) return null;
	String s[]=new String[IDLst.length];
	for (int w=0;w<IDLst.length;w++) s[w]=WindowManager.getImage(IDLst [w]).getTitle();
	return s;
}

private void WinClose(String WinName) {
	int [] IDLst = WindowManager.getIDList();
	if (null==IDLst) return;
	int myID=0;
	for (int w=0;w<IDLst.length;w++) {
		ImagePlus Img=WindowManager.getImage(IDLst [w]);
		if (null!=Img) {if (0==Img.getTitle().compareTo(WinName)) myID=IDLst [w];}
	}
	if (0!=myID) CloseImageWindow(WindowManager.getImage(myID));
}

private ImagePlus CloseImageWindow(ImagePlus imp) {
	if (imp==null) return null;
	if (null!=imp.getWindow()) imp.getWindow().close();
	imp.changes = false;
	imp.unlock();
	imp.trimProcessor();
	imp.flush();
	return null;
}

private void Icalculate(ImagePlus i1, ImagePlus i2, double k1, double k2, int operation) {
	String ImgTitle1="", ImgTitle2="";
	try {
		ImgTitle1=i1.getShortTitle();
		ImgTitle2=i2.getShortTitle();
		double v1, v2=0;
		int width  = i1.getWidth();
		int height = i1.getHeight();
		ImageProcessor ip1, ip2;
		int slices1 = i1.getStackSize();
		int slices2 = i2.getStackSize();
		float[] ctable1 = i1.getCalibration().getCTable();
		float[] ctable2 = i2.getCalibration().getCTable();
		ImageStack stack1 = i1.getStack();
		ImageStack stack2 = i2.getStack();
		int currentSlice = i2.getCurrentSlice();
		for (int n=1; n<=slices2; n++) {
			ip1 = stack1.getProcessor(n<=slices1?n:slices1);
			ip2 = stack2.getProcessor(n);
			ip1.setCalibrationTable(ctable1);
			ip2.setCalibrationTable(ctable2);
			for (int x=0; x<width; x++) {
				for (int y=0; y<height; y++) {
					v1 = (double) ip1.getPixelValue(x,y);
					v2 = (double) ip2.getPixelValue(x,y);
					switch (operation) {
						case ADD: v2 += v1; break;
						case DIVIDE1: v2 = v1!=0.0?v2/v1:0.0; break;
						case DIVIDE2: v2 = v2!=0.0?v1/v2:0.0; break;
						case SUBTRACT: v2 -= v1; break;
						case NDIV2: v2=v2!=0.0?(v1-v2)/v2:0.0; break;
						case NDIV1: v2=v1!=0.0?(v2-v1)/v1:0.0; break;
						case MULTIPLY: v2 *= v1; break;
						case NDIFF1: v2=v1!=0.0?(v1>v2?v1-v2:v2-v1)/v1:0.0; break;
						case NDIFF2: v2=v2!=0.0?(v1>v2?v1-v2:v2-v1)/v2:0.0; break;
					}
					v2*=k1;
					v2+=k2;
					ip2.putPixelValue(x, y, v2);
				}
			}
			if (n==currentSlice) {
				i2.getProcessor().resetMinAndMax();
		 		i2.updateAndDraw();
			}
		}
		i2.updateAndDraw();  //  ##
		return;
	} 
	catch (NullPointerException e) {}
	catch (IllegalArgumentException e) {}
	if (!silentMode && !Canceled) IJ.error("Failed to calculate images ''"+ImgTitle1+"'', ''"+ImgTitle2+"''!");
	if (!Canceled) errString+=(", Failed to calculate images ''"+ImgTitle1+"'', ''"+ImgTitle2+"''");
	IJ.showStatus("canceled.");
	Canceled=true;
}

private void IcalculateSlow(ImagePlus i1, ImagePlus i2, double k1, double k2, int operation) {
	String ImgTitle1="", ImgTitle2="";
 	try {
		ImgTitle1=i1.getShortTitle();
		ImgTitle2=i2.getShortTitle();
		double v1, v2=0;
		int width  = i1.getWidth();
		int height = i1.getHeight();
		ImageProcessor ip1, ip2;
		int slices1 = i1.getStackSize();
		int slices2 = i2.getStackSize();
		float[] ctable1 = i1.getCalibration().getCTable();
		float[] ctable2 = i2.getCalibration().getCTable();
		ImageStack stack1 = i1.getStack();
		ImageStack stack2 = i2.getStack();
		int currentSlice = i2.getCurrentSlice();
		for (int n=1; n<=slices2; n++) {
			ip1 = stack1.getProcessor(n<=slices1?n:slices1);
			ip2 = stack2.getProcessor(n);
			ip1.setCalibrationTable(ctable1);
			ip2.setCalibrationTable(ctable2);
			for (int x=0; x<width; x++)  for (int y=0; y<height; y++) {
				v1 = (double) ip1.getPixelValue(x,y);
				v2 = (double) ip2.getPixelValue(x,y);
				if (Double.isInfinite(v1) || Double.isNaN(v1)) {
					if (null == failureIdxList) {
						failureIdxList = new int[4][1];
						failureIdxList [0][0] = x;
						failureIdxList [1][0] = y;
						failureIdxList [2][0] = n;
						failureIdxList [3][0] = 1;
					} else {
						boolean IdxNotFound=true;
						for (int i = 0; i < failureIdxList[0].length; i ++) if (IdxNotFound) {
							if (x==failureIdxList [0][i] && y== failureIdxList [1][i] && n== failureIdxList [2][i]) {
								IdxNotFound=false;
								failureIdxList [3][i]++;
							}
						}
						if (IdxNotFound) {
							int [][] newIdx = new int [4][failureIdxList[0].length];
							for (int i = 0; i < failureIdxList[0].length; i ++) {
								newIdx[0][i] = failureIdxList[0][i];
								newIdx[1][i] = failureIdxList[1][i];
								newIdx[2][i] = failureIdxList[2][i];
								newIdx[3][i] = failureIdxList[3][i];
							}
							failureIdxList = new int [4][1+newIdx[0].length];
							for (int i = 0; i < newIdx[0].length; i ++) {
								failureIdxList[0][i] = newIdx[0][i];
								failureIdxList[1][i] = newIdx[1][i];
								failureIdxList[2][i] = newIdx[2][i];
								failureIdxList[3][i] = newIdx[3][i];
							}
							failureIdxList [0][newIdx[0].length] = x;
							failureIdxList [1][newIdx[0].length] = y;
							failureIdxList [2][newIdx[0].length] = n;
							failureIdxList [3][newIdx[0].length] = 1;
						}
					}
				} else {
					switch (operation) {
						case ADD: v2 += v1; break;
						case SUBTRACT: v2 -= v1; break;
						case MULTIPLY: v2 *= v1; break;
						case DIVIDE1: v2 = v1!=0.0?v2/v1:0.0; break;
						case DIVIDE2: v2 = v2!=0.0?v1/v2:0.0; break;
						case NDIFF1: v2=v1!=0.0?(v1>v2?v1-v2:v2-v1)/v1:0.0; break;
						case NDIFF2: v2=v2!=0.0?(v1>v2?v1-v2:v2-v1)/v2:0.0; break;
						case NDIV1: v2=v1!=0.0?(v2-v1)/v1:0.0; break;
						case NDIV2: v2=v2!=0.0?(v1-v2)/v2:0.0; break;
					}
					v2*=k1;
					v2+=k2;
					ip2.putPixelValue(x, y, v2);
				}
			}
			if (n==currentSlice) {
				i2.getProcessor().resetMinAndMax();
		 		i2.updateAndDraw();
			}
		}
		i2.updateAndDraw();  //  ##
		return;
	} 
	catch (NullPointerException e) {}
	catch (IllegalArgumentException e) {}
	if (!silentMode && !Canceled) IJ.error("Failed to calculate images ''"+ImgTitle1+"'', ''"+ImgTitle2+"''!");
	if (!Canceled) errString+=(", Failed to calculate images ''"+ImgTitle1+"'', ''"+ImgTitle2+"''");
	IJ.showStatus("canceled.");
	Canceled=true;
}

private void IcalculateSingle(ImagePlus i1, double k1, double k2, int operation) {
	String ImgTitle="";
 	try {
		ImgTitle=i1.getShortTitle();
		double v1, v2=0;
		int width = i1.getWidth();
		int height = i1.getHeight();
		ImageProcessor ip1;
		int slices1 = i1.getStackSize();
		float[] ctable1 = i1.getCalibration().getCTable();
		ImageStack stack1 = i1.getStack();
		int currentSlice = i1.getCurrentSlice();
		for (int n=1; n<=slices1; n++) {
			ip1 = stack1.getProcessor(n);
			ip1.setCalibrationTable(ctable1);
			for (int x=0; x<width; x++) {
				for (int y=0; y<height; y++) {
					v1 = (double) ip1.getPixelValue(x,y);
					switch (operation) {
						case SCALEINV: v2=(v1+k2)/k1; break;
						case SCALEREV: v2=(v1+k2)*k1; break;
						case SCALE: v2=v1*k1+k2; break;
						case ABS: {v1=v1*k1+k2; v2=v1>0.0?v1:-v1;} break;
						case BYPASS: v2=v1; break;
					}
					ip1.putPixelValue(x,y,v2);
				}
			}
			if (n==currentSlice) {
				i1.getProcessor().resetMinAndMax();
		 		i1.updateAndDraw();
			}
		}
		return;
	}
	catch (NullPointerException e) {}
	catch (IllegalArgumentException e) {}
	if (!silentMode && !Canceled) IJ.error("Failed to calculate image ''"+ImgTitle+"''!");
	if (!Canceled) errString+=(", Failed to calculate image ''"+ImgTitle+"''");
	IJ.showStatus("canceled.");
	Canceled=true;
}

private void IcalculateSingleSlow(ImagePlus i1, double k1, double k2, int operation) {
	String ImgTitle="";
 	try {
		ImgTitle=i1.getShortTitle();
		double v1, v2=0.0;
		int width = i1.getWidth();
		int height = i1.getHeight();
		ImageProcessor ip1;
		int slices1 = i1.getStackSize();
		float[] ctable1 = i1.getCalibration().getCTable();
		ImageStack stack1 = i1.getStack();
		int currentSlice = i1.getCurrentSlice();
		for (int n=1; n<=slices1; n++) {
			ip1 = stack1.getProcessor(n);
			ip1.setCalibrationTable(ctable1);
			for (int x=0; x<width; x++) for (int y=0; y<height; y++) {
				v1 = (double) ip1.getPixelValue(x,y);
				if (Double.isInfinite(v1) || Double.isNaN(v1)) {
					if (null == failureIdxList) {
						failureIdxList = new int[4][1];
						failureIdxList [0][0] = x;
						failureIdxList [1][0] = y;
						failureIdxList [2][0] = n;
						failureIdxList [3][0] = 1;
					} else {
						boolean IdxNotFound=true;
						for (int i = 0; i < failureIdxList[0].length; i ++) if (IdxNotFound) {
							if (x==failureIdxList [0][i] && y== failureIdxList [1][i] && n== failureIdxList [2][i]) {
								IdxNotFound=false;
								failureIdxList [3][i]++;
							}
						}
						if (IdxNotFound) {
							int [][] newIdx = new int [4][failureIdxList[0].length];
							for (int i = 0; i < failureIdxList[0].length; i ++) {
								newIdx[0][i] = failureIdxList[0][i];
								newIdx[1][i] = failureIdxList[1][i];
								newIdx[2][i] = failureIdxList[2][i];
								newIdx[3][i] = failureIdxList[3][i];
							}
							failureIdxList = new int [4][1+newIdx[0].length];
							for (int i = 0; i < newIdx[0].length; i ++) {
								failureIdxList[0][i] = newIdx[0][i];
								failureIdxList[1][i] = newIdx[1][i];
								failureIdxList[2][i] = newIdx[2][i];
								failureIdxList[3][i] = newIdx[3][i];
							}
							failureIdxList [0][newIdx[0].length] = x;
							failureIdxList [1][newIdx[0].length] = y;
							failureIdxList [2][newIdx[0].length] = n;
							failureIdxList [3][newIdx[0].length] = 1;
						}
					}
					v2=0.0;
				} else {
					switch (operation) {
						case BYPASS: v2=v1; break;
						case SCALE: v2=v1*k1+k2; break;
						case SCALEREV: v2=(v1+k2)*k1; break;
						case SCALEINV: v2=(v1+k2)/k1; break;
						case ABS: {v1=v1*k1+k2; v2=v1>0.0?v1:-v1;} break;
					}
				}
				ip1.putPixelValue(x,y,v2);
			}
			if (n==currentSlice) {
				i1.getProcessor().resetMinAndMax();
		 		i1.updateAndDraw();
			}
		}
		return;
	}
	catch (NullPointerException e) {}
	catch (IllegalArgumentException e) {}
	if (!silentMode && !Canceled) IJ.error("Failed to calculate image ''"+ImgTitle+"''!");
	if (!Canceled) errString+=(", Failed to calculate image ''"+ImgTitle+"''");
	IJ.showStatus("canceled.");
	Canceled=true;
}

ImagePlus ZProjStack(ImagePlus imp, int min, int max, int method, boolean interpol) {
	String ImgTitle="";	
 	try {
		ImgTitle=imp.getShortTitle();
		IJ.showStatus("Projecting: "+ImgTitle);
		
		// -- workaround for Z-processor MEDIAN method bug (pt 1/2)
		int eHeight = 0;
		int height = imp.getHeight();
		if (method==ZProjector.MEDIAN_METHOD) {
			if (30 > height) {
				extendStack(imp, 30);
				eHeight = height;
				height = 30;
			}
		}
		
		ZProjector zproj = new ZProjector(imp);
		//  if (1 > min) min=1;
		//  if (max > imp.getStackSize()) max = imp.getStackSize();
		zproj.setStartSlice(min);
		zproj.setStopSlice(max);
		
		// --  workaround for SUM method bug (pt 1/2)
		if (method == ZProjector.SUM_METHOD) zproj.setMethod(ZProjector.AVG_METHOD);
		else zproj.setMethod(method);
		// --
		
		IJ.showStatus("Projecting: "+ImgTitle);
		zproj.doProjection();
	            // zproj.computeProjection(method);
            	ImagePlus projection =  zproj.getProjection();
	            ImageProcessor improc = projection.getProcessor();
		int width = projection.getWidth();
		// int height = projection.getHeight();
		
		// --  workaround for SUM method bug (pt 2/2)
		if (method == ZProjector.SUM_METHOD) {
			double v1;
			for (int y=0; y<height; y++) for (int x=0; x<width; x++) {
				v1 = (double) improc.getPixelValue(x,y);
				v1 *= (double) (max-min+1);
				improc.putPixelValue(x, y, v1);
			}
			// projection.updateAndDraw();
		} // --
		
		// -- workaround for Z-processor MEDIAN method bug (pt 2/2)
		if (method==ZProjector.MEDIAN_METHOD) {
			if (0 < eHeight) extendStack(projection, eHeight);
			height = eHeight;
		}
		
		if (!interpol) return projection;
		
		//  -- find invalid values in result and generate list of pixel positions
		double v1;
		int [][] indexList = null;
		for (int y=0; y<height; y++) for (int x=0; x<width; x++) {
			v1 = (double) improc.getPixelValue(x,y);
			if (Double.isInfinite(v1) || Double.isNaN(v1)) {
				if (null == indexList) {
					indexList = new int[2][1];
					indexList [0][0] = x;
					indexList [1][0] = y;
				} else {				
					int [][] newIdx = new int [2][indexList[0].length];
					for (int i = 0; i < indexList[0].length; i ++) {
						newIdx[0][i] = indexList[0][i];
						newIdx[1][i] = indexList[1][i];
					}
					indexList = new int [2][1+newIdx[0].length];
					for (int i = 0; i < newIdx[0].length; i ++) {
						indexList[0][i] = newIdx[0][i];
						indexList[1][i] = newIdx[1][i];
					}
					indexList [0][newIdx[0].length] = x;
					indexList [1][newIdx[0].length] = y;
				}
			}
		}
		ImageProcessor ip2;
		boolean [] valFound;
		double [] goodVal, normTo;
		double [][] goodVals;
		int [] cnt;
		//  -- project pixel by pixel in stack with ignoring frames of invalid value
		if (null != indexList) {
			switch(method) {
				case ZProjector.AVG_METHOD: // 0
					goodVal = new double [indexList[0].length];
					normTo = new double [indexList[0].length];
					for (int n=min; n <= max; n++) {
						ip2 = imp.getStack().getProcessor(n);
						for (int i = 0; i < indexList[0].length; i++)  {
							v1 = (double) ip2.getPixelValue(indexList[0][i],indexList[1][i]);
							if (!Double.isInfinite(v1) && !Double.isNaN(v1)) {
								goodVal[i] += v1;
								normTo[i] ++;
							}
						}
					}
					for (int i = 0; i < indexList[0].length; i++) {
						if (0 < normTo[i]) {
							goodVal[i] /= normTo[i];
							improc.putPixelValue(indexList[0][i],indexList[1][i], goodVal[i]);
						}
					}
				case ZProjector.MAX_METHOD: // 1
					double [] maxVal = new double [indexList[0].length];
					valFound = new boolean [indexList[0].length];
					for (int n=min; n <= max; n++) {
						ip2 = imp.getStack().getProcessor(n);
						for (int i = 0; i < indexList[0].length; i++)  {
							v1 = (double) ip2.getPixelValue(indexList[0][i],indexList[1][i]);
							if (!Double.isInfinite(v1) && !Double.isNaN(v1)) {
								if (maxVal[i] < v1) maxVal [i] = v1;
								valFound[i] = true;
							}
						}
					}
					for (int i = 0; i < indexList[0].length; i++) {
						if (valFound[i]) improc.putPixelValue(indexList[0][i],indexList[1][i], maxVal[i]);
					}
				case ZProjector.MIN_METHOD: // 2
					double [] minVal = new double [indexList[0].length];
					valFound = new boolean [indexList[0].length];
					for (int n=min; n <= max; n++) {
						ip2 = imp.getStack().getProcessor(n);
						for (int i = 0; i < indexList[0].length; i++)  {
							v1 = (double) ip2.getPixelValue(indexList[0][i],indexList[1][i]);
							if (!Double.isInfinite(v1) && !Double.isNaN(v1)) {
								if (!valFound[i] || minVal[i] > v1) minVal [i]= v1;
								valFound[i] = true;
							}
						}
					}
					for (int i = 0; i < indexList[0].length; i++) {
						if (valFound[i]) improc.putPixelValue(indexList[0][i],indexList[1][i], minVal[i]);
					}
				case ZProjector.SUM_METHOD: // 3
					goodVal = new double [indexList[0].length];
					cnt  = new int [indexList[0].length];
					normTo = new double [indexList[0].length];
					for (int n=min; n <= max; n++) {
						ip2 = imp.getStack().getProcessor(n);
						for (int i = 0; i < indexList[0].length; i++)  {
							v1 = (double) ip2.getPixelValue(indexList[0][i],indexList[1][i]);
							if (!Double.isInfinite(v1) && !Double.isNaN(v1)) {
								goodVal[i] += v1;
								normTo[i] ++;
							}
							cnt[i]++;
						}
					}
					for (int i = 0; i < indexList[0].length; i++) {
						if (0 < normTo[i]) {
							goodVal[i] /= normTo[i];
							goodVal[i] *= (double) cnt[i];
							improc.putPixelValue(indexList[0][i],indexList[1][i], goodVal[i]);
						}
					}
				case ZProjector.SD_METHOD: // 4
					goodVals = new double [indexList[0].length][max-min+1];
					cnt  = new int [indexList[0].length];
					for (int n=min; n <= max; n++) {
						ip2 = imp.getStack().getProcessor(n);
						for (int i = 0; i < indexList[0].length; i++)  {
							v1 = (double) ip2.getPixelValue(indexList[0][i],indexList[1][i]);
							if (!Double.isInfinite(v1) && !Double.isNaN(v1)) {
								goodVals[i][cnt[i]++] = v1;
							}
						}
					}
					double [] avVal = new double [indexList[0].length];
					double [] sdVal = new double [indexList[0].length];
					for (int i = 0; i < indexList[0].length; i++) {
						if (1 < cnt[i]) {
							for (int c = 0; c < cnt[i]; c++) avVal[i] += goodVals[i][c];
							avVal[i] /= (double) cnt[i];
							for (int c = 0; c < cnt[i]; c++)  sdVal[i] += (goodVals[i][c] - avVal[i]) * (goodVals[i][c] - avVal[i]);
							sdVal[i] = Math.sqrt(sdVal[i]) / (double) (cnt[i] - 1);
							improc.putPixelValue(indexList[0][i],indexList[1][i], sdVal[i]);
						}
					}
				case ZProjector.MEDIAN_METHOD: // 5
					goodVals = new double [indexList[0].length][max-min+1];
					cnt  = new int [indexList[0].length];
					for (int n=min; n <= max; n++) {
						ip2 = imp.getStack().getProcessor(n);
						for (int i = 0; i < indexList[0].length; i++)  {
							v1 = (double) ip2.getPixelValue(indexList[0][i],indexList[1][i]);
							if (!Double.isInfinite(v1) && !Double.isNaN(v1)) {
								goodVals[i][cnt[i]++] = v1;
							}
						}
					}
					int [][] valIdx = new int [indexList[0].length][max-min+1];
					for (int i = 0; i < indexList[0].length; i++) for (int cc = 0; cc < cnt[i]; cc++)  {
						int cIdx = 1;
						double smVal = Double.MAX_VALUE;
						int smIdx = -1;
						for (int c = 0; c < cnt[i]; c++)  if (0== valIdx[i][c])  if(smVal >  goodVals[i][c]) {
							smIdx=c;
							smVal = goodVals[i][c];
						}
						if (0 <= smIdx) valIdx[i][smIdx] = cIdx++;
					}
					for (int i = 0; i < indexList[0].length; i++) {
						double medianVal = 0;
						double normToVal = 0;
						if (1 == cnt[i])  {
							medianVal = goodVals[i][0];
						} else {
							for (int c = 0; c < cnt[i]; c++)  {
								if (valIdx[i][c] > cnt[i] / 4 && valIdx[i][c] <= cnt[i] * 3 / 4) {
									normToVal++;
									medianVal += goodVals[i][c];
								}
							} 	
						}
						if (0 < normToVal) {
							medianVal /= normToVal;
							improc.putPixelValue(indexList[0][i],indexList[1][i], medianVal);
						}
					}
			}  // switch
			//  -- look  for pixels in result that have still invalid value (i.e. NaN in all frames)
			indexList = null;
			for (int y=0; y<height; y++) for (int x=0; x<width; x++) {
				v1 = (double) improc.getPixelValue(x,y);
				if (Double.isInfinite(v1) || Double.isNaN(v1)) {
					if (null == indexList) {
						indexList = new int[2][1];
						indexList [0][0] = x;
						indexList [1][0] = y;
					} else {				
						int [][] newIdx = new int [2][indexList[0].length];
						for (int i = 0; i < indexList[0].length; i ++) {
							newIdx[0][i] = indexList[0][i];
							newIdx[1][i] = indexList[1][i];
						}
						indexList = new int [2][1+newIdx[0].length];
						for (int i = 0; i < newIdx[0].length; i ++) {
							indexList[0][i] = newIdx[0][i];
							indexList[1][i] = newIdx[1][i];
						}
						indexList [0][newIdx[0].length] = x;
						indexList [1][newIdx[0].length] = y;
					}
				}
			}
			if (null != indexList) {
				double sigma=1.5;
				int x0, x1, y0, y1;
				double fltVal, nrmTo, gk;
				int maxRad = (int) Math.round(2.8 * sigma);
				if (maxRad < 1) maxRad = 1;
				if (50 < maxRad) maxRad = 50;
				for (int i = 0; i < indexList[0].length; i++) {
					fltVal = 0;
					nrmTo = 0;
					for (int rad = 1; rad <= maxRad ; rad ++) {
						x0 = indexList[0][i]-rad;
						if (0<x0) x0=0;
						x1 = indexList[0][i]+rad;
						if (width>x1) x1=width;
						y0 = indexList[1][i]-rad;
						if (0<y0) y0=0;
						y1 = indexList[1][i]+rad;
						if (height >x1) x1=width;
						for (int x=x0; x<x1; x++) for (int y=x0; y<y1; y++)  if(x!=x1 || y!=y1)  {
							v1 = (double) improc.getPixelValue(x,y);
							if (!Double.isInfinite(v1) &&  !Double.isNaN(v1)) {
								gk = Math.exp(Math.abs(-0.5 * (x-x0) * (y-y0) / sigma / sigma));
								fltVal += (gk * v1);
								nrmTo += gk;
							}
						}
					}
					if (0 < nrmTo )  fltVal  /= nrmTo;
					improc.putPixelValue(indexList[0][i],indexList[1][i], fltVal);
				}
			}
			// projection.updateAndDraw();
		} // if (null != indexList)...
		
		return projection;
	
	} // try
	catch (NullPointerException e) {}
	catch (IllegalArgumentException e) {}
	
	if (!silentMode && !Canceled) IJ.error("Failed to calculate Z-Projection of image ''"+ImgTitle+"''!");
	if (!Canceled) errString+=(", Failed to calculate Z-Projection of image ''"+ImgTitle+"''");
	IJ.showStatus("canceled.");
	Canceled=true;
	return null;
}

private void preFilter (ImageProcessor improc, boolean modusMean, double sigma) {
	//  -- find invalid values in result and generate list of pixel positions
	
	if (modusMean) sigma = Math.round(sigma);
	double v1;
	int [][] indexList = null;
	int width = improc.getWidth();
	int height = improc.getHeight();
	for (int y=0; y<height; y++) for (int x=0; x<width; x++) {
		v1 = (double) improc.getPixelValue(x,y);
		if (Double.isInfinite(v1) || Double.isNaN(v1)) {
			if (null == indexList) {
				indexList = new int[2][1];
				indexList [0][0] = x;
				indexList [1][0] = y;
			} else {				
				int [][] newIdx = new int [2][indexList[0].length];
				for (int i = 0; i < indexList[0].length; i ++) {
					newIdx[0][i] = indexList[0][i];
					newIdx[1][i] = indexList[1][i];
				}
				indexList = new int [2][1+newIdx[0].length];
				for (int i = 0; i < newIdx[0].length; i ++) {
					indexList[0][i] = newIdx[0][i];
					indexList[1][i] = newIdx[1][i];
				}
				indexList [0][newIdx[0].length] = x;
				indexList [1][newIdx[0].length] = y;
			}
		}
	} // for (int y=0; y<height; y++) for (int x=0; x<width; x++)...
	if (null != indexList) {
		// double sigma=1.5;
		int x0, x1, y0, y1;
		double fltVal, nrmTo, gk;
		int maxRad;
		if (!modusMean) maxRad = (int) Math.round(2.8 * sigma);
		else maxRad = (int) Math.round(sigma);
		if (maxRad < 1) maxRad = 1;
		if (50 < maxRad) maxRad = 50;
		for (int i = 0; i < indexList[0].length; i++) {
			fltVal = 0;
			nrmTo = 0;
			for (int rad = 1; rad <= maxRad ; rad ++) {
				x0 = indexList[0][i]-rad;
				if (0<x0) x0=0;
				x1 = indexList[0][i]+rad;
				if (width>x1) x1=width;
				y0 = indexList[1][i]-rad;
				if (0<y0) y0=0;
				y1 = indexList[1][i]+rad;
				if (height >x1) x1=width;
				for (int x=x0; x<x1; x++) for (int y=x0; y<y1; y++)  if(x!=x1 || y!=y1)  {
					v1 = (double) improc.getPixelValue(x,y);
					if (!Double.isInfinite(v1) &&  !Double.isNaN(v1)) {
						if (!modusMean) gk = Math.exp(Math.abs(-0.5 * (x-x0) * (y-y0) / sigma / sigma));
						else gk = 1;
						fltVal += (gk * v1);
						nrmTo += gk;
					}
				}
			}
			if (0 < nrmTo )  fltVal  /= nrmTo;
			improc.putPixelValue(indexList[0][i],indexList[1][i], fltVal);
		}
	} // if (null != indexList)...
}


ImagePlus ZFilterT(ImagePlus imp, int Rad, boolean interpol) {
 	String ImgTitle="";
	ImageStack in_stack = null, out_stack = null;
	try {
		ImgTitle=imp.getShortTitle();
		IJ.showStatus("Projecting: "+ImgTitle);
		int Sz = imp.getStackSize();
		in_stack =imp.getStack();
		int width = in_stack.getWidth();
		int height = in_stack.getHeight();
		// Calibration cal = imp.getCalibration();
		float[] ctable = imp.getCalibration().getCTable();
		out_stack = imp.createEmptyStack();
		ZProjector zproj = new ZProjector(imp);
		
		//  -- Slices are numbered 1,2,...,n where n is the number of slices in the stack!
		for (int o=0;o<Sz;o++) {
			int min=o-Rad+1;
			if (1>min) min=1;
			int max =o+Rad+1;
			if (Sz<max) max=Sz;
			ImageProcessor improc;
			double v1;
			int [][] indexList = null;
			if (min+1>=max) {
				improc = in_stack.getProcessor(o+1);
			} else {
				zproj.setStartSlice(min);
				zproj.setStopSlice(max);
				zproj.setMethod(zproj.AVG_METHOD);
				zproj.doProjection();
		            	// ImagePlus projection =  zproj.getProjection();
	            		improc = zproj.getProjection().getProcessor();
				
				//  -- find invalid values in result and generate list of pixel positions
				// double v1;
				// int [][] indexList = null;
				if (interpol) for (int y=0; y<height; y++) for (int x=0; x<width; x++) {
					v1 = (double) improc.getPixelValue(x,y);
					if (Double.isInfinite(v1) || Double.isNaN(v1)) {
						if (null == indexList) {
							indexList = new int[2][1];
							indexList [0][0] = x;
							indexList [1][0] = y;
						} else {				
							int [][] newIdx = new int [2][indexList[0].length];
							for (int i = 0; i < indexList[0].length; i ++) {
								newIdx[0][i] = indexList[0][i];
								newIdx[1][i] = indexList[1][i];
							}
							indexList = new int [2][1+newIdx[0].length];
							for (int i = 0; i < newIdx[0].length; i ++) {
								indexList[0][i] = newIdx[0][i];
								indexList[1][i] = newIdx[1][i];
							}
							indexList [0][newIdx[0].length] = x;
							indexList [1][newIdx[0].length] = y;
						}
					}
				}
				ImageProcessor ip2;
				boolean [] valFound;
				double [] goodVal, normTo;
				double [][] goodVals;
				int [] cnt;
				//  -- project pixel by pixel in stack with ignoring frames of invalid value
				if (null != indexList) {
					goodVal = new double [indexList[0].length];
					normTo = new double [indexList[0].length];
					for (int n=min; n <= max; n++) {
						ip2 = imp.getStack().getProcessor(n);
						for (int i = 0; i < indexList[0].length; i++)  {
							v1 = (double) ip2.getPixelValue(indexList[0][i],indexList[1][i]);
							if (!Double.isInfinite(v1) && !Double.isNaN(v1)) {
								goodVal[i] += v1;
								normTo[i] ++;
							}
						}
					}
					for (int i = 0; i < indexList[0].length; i++) {
						if (0 < normTo[i]) {
							goodVal[i] /= normTo[i];
							improc.putPixelValue(indexList[0][i],indexList[1][i], goodVal[i]);
						}
					}
				}
			}
			//  -- look  for pixels in result that have still invalid value (i.e. NaN in all frames)
			indexList = null;
			if (interpol) for (int y=0; y<height; y++) for (int x=0; x<width; x++) {
				v1 = (double) improc.getPixelValue(x,y);
				if (Double.isInfinite(v1) || Double.isNaN(v1)) {
					if (null == indexList) {
						indexList = new int[2][1];
						indexList [0][0] = x;
						indexList [1][0] = y;
					} else {				
						int [][] newIdx = new int [2][indexList[0].length];
						for (int i = 0; i < indexList[0].length; i ++) {
							newIdx[0][i] = indexList[0][i];
							newIdx[1][i] = indexList[1][i];
						}
						indexList = new int [2][1+newIdx[0].length];
						for (int i = 0; i < newIdx[0].length; i ++) {
							indexList[0][i] = newIdx[0][i];
							indexList[1][i] = newIdx[1][i];
						}
						indexList [0][newIdx[0].length] = x;
						indexList [1][newIdx[0].length] = y;
					}
				}
			}
			if (null != indexList) {
				double sigma=1.5;
				int x0, x1, y0, y1;
				double fltVal, nrmTo, gk;
				int maxRad = (int) Math.round(2.8 * sigma);
				if (maxRad < 1) maxRad = 1;
				if (50 < maxRad) maxRad = 50;
				for (int i = 0; i < indexList[0].length; i++) {
					fltVal = 0;
					nrmTo = 0;
					for (int rad = 1; rad <= maxRad ; rad ++) {
						x0 = indexList[0][i]-rad;
						if (0<x0) x0=0;
						x1 = indexList[0][i]+rad;
						if (width>x1) x1=width;
						y0 = indexList[1][i]-rad;
						if (0<y0) y0=0;
						y1 = indexList[1][i]+rad;
						if (height >x1) x1=width;
						for (int x=x0; x<x1; x++) for (int y=x0; y<y1; y++)  if(x!=x1 || y!=y1)  {
							v1 = (double) improc.getPixelValue(x,y);
							if (!Double.isInfinite(v1) &&  !Double.isNaN(v1)) {
								gk = Math.exp(Math.abs(-0.5 * (x-x0) * (y-y0) / sigma / sigma));
								fltVal += (gk * v1);
								nrmTo += gk;
							}
						}
					}
					if (0 < nrmTo )  fltVal  /= nrmTo;
					improc.putPixelValue(indexList[0][i],indexList[1][i], fltVal);
				}
			} // if (null != indexList)...
			improc.setCalibrationTable(ctable);
			out_stack.addSlice(in_stack.getSliceLabel(o+1),  improc);
		} // for (int o=0;o<Sz;o++)...
		imp.setStack(null, out_stack);
		imp.changes=true;
		return imp;
	}
	catch(OutOfMemoryError e) {
		if (!silentMode && !Canceled) IJ.error("Failed Z filter image ''"+ImgTitle+"'':\nNot enough memory!");
		if (!Canceled) errString+=(", Failed Z filter image ''"+ImgTitle+"'': not enough memory");
	}
	catch(NullPointerException e) {
		if (!silentMode && !Canceled) IJ.error("Failed Z filter image ''"+ImgTitle+"'':\nNot found!");
		if (!Canceled) errString+=(", Failed Z filter image ''"+ImgTitle+"'': not found");
	}
	catch (IllegalArgumentException e) {
		if (!silentMode && !Canceled) IJ.error("Failed Z filter image ''"+ImgTitle+"'';\nprobably canceled.");
		if (!Canceled) errString+=(", Failed Z filter image ''"+ImgTitle+"'': probably canceled");
	}
	if (null != out_stack) out_stack.trim();
	out_stack=null;
	imp = CloseImageWindow(imp);
	imp=null;
	return null;
}

private double [] getZProfile(ImagePlus imp) {
 	String ImgTitle="";
	try {
		ImgTitle=imp.getShortTitle();
		ImageStack stack = imp.getStack();
		int Sz = stack.getSize(); 
		double [] qv = new double [Sz +1];
		int h=imp.getHeight();
		int w=imp.getWidth();
		Roi roi = null;
		if (h>15 && w > 15) roi=new Roi (w/4, h/4, w/2, h/2);
		else roi=new Roi (0, 0, w, h);
		Calibration cal = imp.getCalibration();
		try {
			for (int i=1; i<=Sz; i++) {
				ImageProcessor ip = stack.getProcessor(i);
				ip.setRoi(roi);
				ImageStatistics stats = ImageStatistics.getStatistics(ip, Measurements.MEAN, cal);
				qv[i] = (double)stats.mean;
			}
		}
		catch (NoSuchMethodError ex) {
			Rectangle r = null;
			if (h>15 && w > 15) r = new Rectangle (w/4, h/4, w/2, h/2);
			else r = new Rectangle (0, 0, w, h);
			for (int i=1; i<=Sz; i++) {
				ImageProcessor ip2 = stack.getProcessor(i);
				ip2.setRoi(r);
				ImageStatistics stats = ImageStatistics.getStatistics(ip2, Measurements.MEAN, cal);
				qv[i] = (double)stats.mean;
			}
		}
		double[] extrema = Tools.getMinMax(qv);
		if (Math.abs(extrema[1])==Double.MAX_VALUE) return null;
		else return qv;
	}
	catch(NullPointerException e) {}
	catch (IllegalArgumentException e) {}
	if (!silentMode && !Canceled) IJ.error("Failed get deviation of image ''"+ImgTitle+"''!");
	if (!Canceled) errString+=(", Failed get deviation of image ''"+ImgTitle+"''");
	return null;
}

private boolean checkNames(String [] sn, int xj, String fTag, boolean TestRaw) {
	for (int k=0; k<sn.length; k++) {
		sn[k]=sn[k].toLowerCase();
		fTag=fTag.toLowerCase();
		if (sn[k].endsWith(fTag+".tif")) {   //  tiff, TIFF, ... #################
			for (int i=0; i<nhl.length; i++) {
				String Lnhli=nhl[i].toLowerCase();
				if (sn[k].startsWith(Lnhli+"_")) {
					if (sn[k].startsWith(Lnhli+"_deltai")) {
						if (0==sn[k].compareTo(Lnhli+"_deltai_i"+fTag+".tif")) return true;
						if (0==sn[k].compareTo(Lnhli+"_deltain_in"+fTag+".tif")) return true;
						if (0==sn[k].compareTo(Lnhli+"_deltain_in%"+fTag+".tif")) return true;
					} else if (sn[k].startsWith(Lnhli+"_deltaf")) {
						if (0==sn[k].compareTo(Lnhli+"_deltaf_f"+fTag+".tif")) return true;
						if (0==sn[k].compareTo(Lnhli+"_deltafn_fn"+fTag+".tif")) return true;
						if (0==sn[k].compareTo(Lnhli+"_deltafn_fn%"+fTag+".tif")) return true;
					} else if (sn[k].startsWith(Lnhli+"_deltasd")) {
						if (0==sn[k].compareTo(Lnhli+"_deltasd_f"+fTag+".tif")) return true;
						if (0==sn[k].compareTo(Lnhli+"_deltasdn_fn"+fTag+".tif")) return true;
						if (0==sn[k].compareTo(Lnhli+"_deltasdn_fn%"+fTag+".tif")) return true;
						if (0==sn[k].compareTo(Lnhli+"_deltasd_i"+fTag+".tif")) return true;
						if (0==sn[k].compareTo(Lnhli+"_deltasdn_in"+fTag+".tif")) return true;
						if (0==sn[k].compareTo(Lnhli+"_deltasdn_in%"+fTag+".tif")) return true;
					} else if (sn[k].startsWith(Lnhli+"_di")) {
						if (0==sn[k].compareTo(Lnhli+"_div_%change"+fTag+".tif")) return true;
						if (0==sn[k].compareTo(Lnhli+"_divn_%change"+fTag+".tif")) return true;
						if (0==sn[k].compareTo(Lnhli+"_div_%changen"+fTag+".tif")) return true;
						if (0==sn[k].compareTo(Lnhli+"_div_%stack"+fTag+".tif")) return true;
						if (0==sn[k].compareTo(Lnhli+"_divn_%stack"+fTag+".tif")) return true;
						if (0==sn[k].compareTo(Lnhli+"_div_32bit"+fTag+".tif")) return true;
						if (0==sn[k].compareTo(Lnhli+"_divn_32bit"+fTag+".tif")) return true;
						if (0==sn[k].compareTo(Lnhli+"_diff"+fTag+".tif")) return true;
						if (0==sn[k].compareTo(Lnhli+"_diffn"+fTag+".tif")) return true;
						for (int ii=i+1;ii<nhl.length; ii++) {
							if (0==sn[k].compareTo(nhl[ii]+"_div_by_"+nhl[i]+"_32bit"+fTag+".tif")) return true;
							if (0==sn[k].compareTo(nhl[ii]+"_divn_by_"+nhl[i]+"_32bit"+fTag+".tif")) return true;
							if (0==sn[k].compareTo(nhl[ii]+"_div_by_"+nhl[i]+"_32bitn"+fTag+".tif")) return true;
						}
					} else {
						if (0==sn[k].compareTo(Lnhli+"_base"+fTag+".tif")) return true;
						if (0==sn[k].compareTo(Lnhli+"_basen"+fTag+".tif")) return true;
						if (0==sn[k].compareTo(Lnhli+"_resp"+fTag+".tif")) return true;
						if (0==sn[k].compareTo(Lnhli+"_respn"+fTag+".tif")) return true;
						for (int v=0; v<AVS.length; v++) if (0<AVS[v].length()) {
							if (0==sn[k].compareTo(Lnhli+AVS[v].toLowerCase()+fTag+".tif")) return true;
							if (0==sn[k].compareTo(Lnhli+AVS[v].toLowerCase()+"n"+fTag+".tif")) return true;
						}
					}
				} else if (TestRaw) if (sn[k].startsWith(Lnhli)) {
					//  -- check for file/window names like "hi23.tif"
					int nn=0;
					//  tiff, TIFF, ... #################
					try {nn=Integer.parseInt(sn[k].substring(Lnhli.length(),sn[k].length()-4-fTag.length()));}
					catch (NumberFormatException e) {}
					//  -- also check for file/window names like "hi23n.tif"
					if (0==nn) if (0==sn[k].substring(sn[k].length()-5-fTag.length(),sn[k].length()-4-fTag.length()).compareTo("n")) {
						try {nn=Integer.parseInt(sn[k].substring(Lnhli.length(),sn[k].length()-5-fTag.length()));}
						catch (NumberFormatException e) {}
					}
					if (0<nn) return true;
				}
			}
		} else if (sn[k].endsWith(fTag+".txt")) {
			if (0==sn[k].compareTo("log"+fTag+".txt")) return true;
			if (0==sn[k].compareTo("data quality map"+fTag+".txt")) return true;
		}
	}
	return false;
}

private void showProgress (double q) {
	if (0>=PBwait) return; 
	if (System.currentTimeMillis()-PgrTime < PBwait) return;  // -- has to be at least 30 ms
	IJ.showProgress(q);
	PgrTime = System.currentTimeMillis();
}
private String getFileDate(File f) {
	try { if (!f.isFile()) return "";}
	catch (SecurityException ex) {return "";}
	catch (NullPointerException ex) {return "";}
	long timestamp = f.lastModified();
	if (0>timestamp) return "";
	return DateTimeString(new Date(timestamp));
}
private String DateTimeString(Date when) {
	SimpleDateFormat sdf = new SimpleDateFormat( "EEE yyyy/MM/dd HH:mm:ss zz" );
	sdf.setTimeZone(TimeZone.getDefault()); // local time
	return sdf.format(when);
}
private static final int[] getScreenSize() {
	GraphicsEnvironment ge = GraphicsEnvironment.getLocalGraphicsEnvironment();
	GraphicsDevice[] gs = ge.getScreenDevices();
	int [] devDim = {640, 340};
	for (int j = 0; j < gs.length; j++) {
		GraphicsDevice gdv = gs[j];
		GraphicsConfiguration[] gc = gdv.getConfigurations();
		for (int i=0; i < gc.length; i++) {
			Rectangle gcBounds = gc[i].getBounds();
			int w = (int) gcBounds.getWidth();
			int h = (int) gcBounds.getHeight();
			if (devDim[0] < w) devDim[0] = w;
			if (devDim[1] < h) devDim[1] = h;
		}
	}
	return devDim;
}

//  -- returns the number of Checkboxes displayable on screen using main dialog (in two columns!)
private int getMaxNbChkbxs(boolean extend) {
	int panelHeight = 15 + checkboxHeight;
	int top = 42 + checkboxHeight + 2 * panelHeight;  //  = 147  (1 checkbox and 2 panels)
	int bottom = 96 + 4 * panelHeight; //  =256 (4 txt boxes and 2 panels with buttons)
	if (extend) bottom += panelHeight;  //  (1 more panel)
	// space outside the dialog: 60 px
	int canvasHeight = screenSize[1] - 60 - top - bottom;
	int NbChkbxs = (int) Math.floor ((double)((canvasHeight / checkboxHeight) * 2));
	if (0 > NbChkbxs && extend) {
		allowBlankArtifact = false;
		NbChkbxs = 2;
	} else if (4 > NbChkbxs) NbChkbxs = 4;
	return NbChkbxs;
}

private int getLineHeight() {
	int hght = 17;
	// = IJ.isJava14()&&IJ.isWindows()?19:17;  // until Java 1.5: 17 px
	try { if (IJ.isJava14()&&IJ.isWindows()) hght = 19; }
	catch (NoSuchMethodError e) {}
	return hght;
}

//  -- returns the number of lines displayable on screen in a message dialog
private int getMaxNbLines() {
	int topH = 46;
	int bottomH = 59;
	int spaceOutside = 90;  // minimal space outside the dialog: 90 px
	int canvasHeight = screenSize[1] -bottomH -topH -spaceOutside;
	int NbLines = (int) Math.floor((double)(canvasHeight / lineHeight));
	if (3 > NbLines) NbLines = 3;
	return NbLines;
}

private void ZoomGood(ImagePlus imp) {
	ImageWindow win = imp.getWindow();
	if (win==null) return;
	ImageCanvas ic = win.getCanvas();
	if (0>=zoomLevel) {
		int ImgHeight = imp.getHeight();
		int ImgWidth = imp.getWidth();
		zoomLevel = zoomLevels[zoomLevels.length-1];
		double myZoomLevel = -1;
		for (int j = 0; j < screenSize.length; j++) {
			if (0>=WinSizeOnScreen || 1<WinSizeOnScreen) myZoomLevel = screenSize[j] * 0.4 / ImgHeight;
			else myZoomLevel = (double) screenSize[j] * WinSizeOnScreen / ImgHeight;
			if (myZoomLevel > zoomLevel) myZoomLevel = zoomLevel;
		}
		zoomLevel = zoomLevels[0];
		for (int idx=0; idx< zoomLevels.length; idx++) if (myZoomLevel >= zoomLevels[idx]) zoomLevel = zoomLevels[idx];
		
		if (DoShowLog && DoFitLogWinSizeToScreen) {
			if (ImgHeight / screenSize[0] > ImgWidth / screenSize[1]) {
				if (0>=WinSizeOnScreen || 1< WinSizeOnScreen) LogWinHeight = (int)(screenSize[0]*0.4);
				else LogWinHeight = (int)(screenSize[0]*WinSizeOnScreen);
				LogWinWidth = (LogWinHeight*ImgWidth)/ImgHeight;
				if (screenSize[1]*0.9<LogWinWidth) LogWinWidth = (int)(screenSize[1]*0.9);
			} else {
				LogWinWidth = (int)(screenSize[1]*WinSizeOnScreen);
				if (0>=WinSizeOnScreen || 1< WinSizeOnScreen) LogWinWidth = (int)(screenSize[1]*0.4);
				LogWinHeight = (LogWinHeight*ImgWidth)/ImgWidth;
				if (screenSize[0]*0.9<LogWinHeight) LogWinHeight = (int)(screenSize[0]*0.9);
			}
		}
		if (DoAdjQmapWinSizeToScreen) {
			if (screenSize[0]*0.8<QmapWinHeight) QmapWinHeight=(int)(screenSize[0]*0.8);
			if (screenSize[1]*0.8<QmapWinWidth) QmapWinWidth=(int)(screenSize[1]*0.8);
		}
	}
	while(ic.getMagnification()<zoomLevel) {
		ic.zoomIn(0,0);
		if (ic.getMagnification()<=1.0) imp.repaintWindow();
	}
	while(ic.getMagnification()>zoomLevel) {
		ic.zoomOut(0,0);
		if (ic.getMagnification()<1.0) imp.repaintWindow();
	}
}

private void SetIgnore() {
	if (null==iob) return;
	for (int i=0;i<2;i++) { 
		// if (1==nhl.length || (2==nhl.length && !hasControl)) {
		if (1==nhl.length) {
			for(int j=1;j<=NbChkbxs;j++) {
				if (iob[1][nhl.length-1].length>j+IboxStartRow+i*NbChkbxs) 
					iob[1][nhl.length-1][j+IboxStartRow+i*NbChkbxs]=IgnoreBox[i][j].isSelected();
			}
		} else { 
			if (iob[1].length>i+IboxStartColumn) for(int j=1;j<=NbChkbxs;j++)
				if (iob[1][i].length>j+IboxStartRow) iob[1][i+IboxStartColumn][j+IboxStartRow]=IgnoreBox[i][j].isSelected();
		}
	}
}

private void FitLogWinSizeToScreen() {
	if (DoFitLogWinSizeToScreen) {
		if (screenSize[1] > screenSize[0]) {
			if (0>=WinSizeOnScreen || 1< WinSizeOnScreen) LogWinHeight=(int)(screenSize[0]*0.4);
			else LogWinHeight=(int)(screenSize[0]*WinSizeOnScreen);
			LogWinWidth=LogWinHeight;
		} else {
			if (0>=WinSizeOnScreen || 1< WinSizeOnScreen) LogWinWidth=(int)(screenSize[1]*0.4);
			else LogWinWidth=(int)(screenSize[1]*WinSizeOnScreen);
			LogWinHeight=LogWinWidth;
		}
	}
	if (DoAdjQmapWinSizeToScreen) {
		if (screenSize[0]*0.8<QmapWinHeight) QmapWinHeight=(int)(screenSize[0]*0.8);
		if (screenSize[1]*0.8<QmapWinWidth) QmapWinWidth=(int)(screenSize[1]*0.8);
	}
}

}


